Package: PDEnaiveBayes
Type: Package
Title: Plausible Naive Bayes Classifier Using PDE
Version: 0.4.0
Date: 2026-07-19
Authors@R: c(person("Michael", "Thrun",role=c("aut","cph","cre"),email= "m.thrun@gmx.net", comment = c(ORCID = "0000-0001-9542-5543")),person("Quirin", "Stier", email= "Quirin_Stier@gmx.de",role=c("aut","rev"), comment = c(ORCID = "0000-0002-7896-4737")),person("Tim Robin", "Neldner",role=c("ctr","ctb")))
Maintainer: Michael Thrun <m.thrun@gmx.net>
Description: Provides a nonparametric, multicore-capable plausible naive Bayes classifier based on Pareto density estimation (PDE). It addresses low-evidence cases through a plausibility correction. To enhance the interpretability of the flexible naive Bayes classifier by revealing its posterior structure and feature-wise, class-specific evidence, posterior probabilities can be visualized as class-wise line plots for one-dimensional data or color-coded Voronoi diagrams for pairwise feature projections, and class-conditional PDE likelihoods as overlaid, mirrored density profiles resembling violin plots. Methodological details are provided by Stier, Q., Hoffmann, J. and Thrun, M. C. (2026) "Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naive Bayes" <DOI:10.3390/make8010013>. For multicore computations, the implementation applies the general memory-sharing approach described by Thrun, M. C. and Märte, J. (2026) "memshare: Memory Sharing for Multicore Computation in R with an Application to Feature Selection by Mutual Information using PDE" <DOI:10.32614/RJ-2025-043>.
LazyLoad: yes
LazyData: TRUE
NeedsCompilation: yes
Imports: Rcpp (>= 1.0.8), RcppParallel (>= 5.1.4), pracma, plotly,
        utils, grDevices, stats, graphics, methods, ggplot2,
        DatabionicSwarm, memshare
Suggests: FCPS(>= 1.3.5), ABCanalysis, modeest, deldir, ScatterDensity
        (>= 0.1.1), gridExtra, parallelDist, parallel,
        DataVisualizations (>= 1.1.5), knitr
VignetteBuilder: knitr
LinkingTo: Rcpp, RcppParallel
Depends: R (>= 2.10)
License: GPL-3
Encoding: UTF-8
BugReports: https://github.com/Mthrun/PDEbayes/issues
Packaged: 2026-07-19 10:54:04 UTC; MCT
Author: Michael Thrun [aut, cph, cre] (ORCID:
    <https://orcid.org/0000-0001-9542-5543>),
  Quirin Stier [aut, rev] (ORCID:
    <https://orcid.org/0000-0002-7896-4737>),
  Tim Robin Neldner [ctr, ctb]
Repository: CRAN
Date/Publication: 2026-07-19 18:00:19 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-07-19 21:32:12 UTC; unix
Archs: PDEnaiveBayes.so.dSYM

useDynLib("PDEnaiveBayes", .registration = TRUE)

#exportPattern("^[[:alpha:]]+")
export(ApplyBayesTheorem4Likelihoods)
export(getPriors)
export(PlotBayesianDecision2D)
export(PlotLikelihoodFuns)
export(PlotLikelihoods)
export(PlotNaiveBayes)
export(PlotPosteriors)
export(predict.PDEbayes)
export(Predict_naiveBayes)
export(Train_naiveBayes)
export(Train_naiveBayes_multicore)

S3method(predict, PDEbayes)
#exportClasses(PDEbayes)     # if you’re also defining a new class
#exportMethods(predict)     # for S4-style exports, if needed

import(methods)
import(memshare)
import("ggplot2")
import(pracma)

importFrom(Rcpp, evalCpp)
importFrom(RcppParallel,RcppParallelLibs)

importFrom("graphics", "plot", "abline", "lines", "par", "points")
importFrom("stats", "dnorm", "sd", "approx", "density", "dist", "fft", "median", "quantile", "runif")
importFrom("grDevices", "col2rgb", "heat.colors", "colorRampPalette")
importFrom("utils", "tail", "packageVersion", "stack")
importFrom("plotly", subplot, layout, add_trace, plot_ly)
importFrom("stats", "setNames")
importFrom("utils", "head")

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/PDEnaiveBayes)](https://cran.r-project.org/package=PDEnaiveBayes)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/grand-total/PDEnaiveBayes?color=blue)](https://r-pkg.org/pkg/PDEnaiveBayes)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/last-month/PDEnaiveBayes?color=green)](https://r-pkg.org/pkg/PDEnaiveBayes)

# PDEnaiveBayes

### Table of Contents  
[1. Introduction](#introduction)  
[2. Installation](#installation)  
[3. Additional Resources](#additional)  
[4. References](#references)  

## 1. Introduction <a name="introduction"/>

A nonparametric, multicore-capable plausible naive Bayes classifier based on the Pareto density estimation (PDE) featuring a plausible approach to a pitfall in the Bayesian theorem covering low evidence cases [Stier et al., 2026].

The **PDEnaiveBayes** package provides a nonparametric approach to Naive Bayes classification, relying on Pareto Density Estimation (PDE) for likelihood modeling. It offers functionalities for:

- Training standard and multicore PDE-based Naive Bayes classifiers
- Estimating likelihoods and priors nonparametrically
- Applying Bayes’ theorem in a numerically stable manner
- Plotting likelihoods, posteriors, and decision boundaries

### Training and Prediction

#### Train_naiveBayes

Train a Pareto-density-based Naive Bayes classifier.

```r
data(Hepta)
Data <- Hepta$Data
Cls  <- Hepta$Cls
model <- Train_naiveBayes(Data, Cls, Gaussian = FALSE)
table(Cls, model$ClsTrain)
```

#### Train_naiveBayes_multicore

Multicore version of the PDE-based Naive Bayes training.

```r
if (requireNamespace("FCPS")) {
  data(Hepta)
  Data <- Hepta$Data
  Cls  <- Hepta$Cls
  model_mc <- Train_naiveBayes_multicore(
    cl       = NULL,
    Data     = Data,
    Cls      = Cls,
    Gaussian = FALSE,
    Predict  = TRUE
  )
  table(Cls, model_mc$ClsTrain)
}
```

#### Predict_naiveBayes

Predict class memberships for new data using a trained naive Bayes model.

```r
if (requireNamespace("FCPS")) {
  V    <- FCPS::ClusterChallenge("Hepta", 1000)
  Data <- V$Hepta
  Cls  <- V$Cls

  ind       <- 1:length(Cls)
  ind_train <- sample(ind, 800)
  ind_test  <- setdiff(ind, ind_train)

  model <- Train_naiveBayes(Data[ind_train, ], Cls[ind_train], Gaussian = FALSE)
  res   <- Predict_naiveBayes(Data[ind_test, ], Model = model)

  table(Cls[ind_test], res$ClsTest)
}
```

#### predict.PDEbayes

S3 prediction method for PDE-based naive Bayes models.

```r
if (requireNamespace("FCPS")) {
  V    <- FCPS::ClusterChallenge("Hepta", 1000)
  Data <- V$Hepta
  Cls  <- V$Cls

  ind       <- 1:length(Cls)
  ind_train <- sample(ind, 800)
  ind_test  <- setdiff(ind, ind_train)

  model   <- Train_naiveBayes(Data[ind_train, ], Cls[ind_train], Gaussian = FALSE)
  ClsTest <- predict.PDEbayes(object = model, newdata = Data[ind_test, ])
  table(Cls[ind_test], ClsTest)
}
```

### Probability and Density Utilities

#### getPriors

Compute class priors from class proportions.

```r
if (requireNamespace("FCPS")) {
  data(Hepta)
  Cls    <- Hepta$Cls
  Priors <- getPriors(Cls)
  print(Priors)
}
```

### Visualization Tools

The following functions visualize either the learned likelihoods, the resulting posterior probabilities, or a 2D Bayesian decision surface. They are primarily useful for inspection and model interpretability.

#### PlotNaiveBayes

Visualize the PDE-based naive Bayes model for all or selected features.

```r
Data <- as.matrix(iris[, 1:4])
Cls  <- as.numeric(iris[, 5])

model <- Train_naiveBayes(Data = Data, Cls = Cls, Plausible = FALSE)
FeatureNames <- colnames(Data)

PlotNaiveBayes(Model = model$Model, FeatureNames = FeatureNames)
```

![Alt text](inst/extdata/images/ClassCOnditionalLikelihoodsPerFeatureMDPlot_Iris.png)


#### PlotLikelihoods

Plot likelihoods for feature-class combinations.

```r
Data <- as.matrix(iris[, 1:4])
Cls  <- as.numeric(iris[, 5])

model <- Train_naiveBayes(Data = Data, Cls = Cls, Plausible = FALSE)

PlotLikelihoods(
  Likelihoods = model$Model$ListOfLikelihoods,
  Data        = Data
)
```

#### PlotLikelihoodFuns

Plot likelihood functions as obtained from PDE or Gaussian estimation.

```r
Data <- as.matrix(iris[, 1:4])
Cls  <- as.numeric(iris[, 5])

model <- Train_naiveBayes(Data = Data, Cls = Cls, Plausible = FALSE)

PlotLikelihoodFuns(
  LikelihoodFuns = model$Model$PDFs_funs,
  Data           = Data
)
```

![Alt text](inst/extdata/images/ClassConditionalLikelihoodsPerFeature_Iris.png)


#### PlotPosteriors

Plot posterior class probabilities for a selected class.

```r
Data <- as.matrix(iris[, 1:4])
Cls  <- as.numeric(iris[, 5])

model <- Train_naiveBayes(Data = Data, Cls = Cls, Plausible = FALSE)

PlotPosteriors(
  Data       = Data,
  Posteriors = model$Posteriors,
  Class      = 1
)
```

#### PlotBayesianDecision2D

Display a 2D Bayesian decision surface.

```r
Data <- as.matrix(iris[, 1:4])
Cls  <- as.numeric(iris[, 5])

model <- Train_naiveBayes(Data = Data, Cls = Cls, Plausible = FALSE)

PlotBayesianDecision2D(
  X          = Data[, 1],
  Y          = Data[, 2],
  Posteriors = model$Posteriors,
  Class      = 1
)
```
![Alt text](inst/extdata/images/2DSliceVoronoiTessallationDecisionBoundary_Iris.png)


### Tutorial Examples

The tutorial with several examples can be found on in the vignette on CRAN.

## 2. Installation <a name="installation"/>

#### Installation using CRAN (once published)
```r
install.packages("PDEnaiveBayes", dependencies = TRUE)
```

#### Installation using GitHub
Please note, that dependencies have to be installed manually.

```r
remotes::install_github("Mthrun/PDEnaiveBayes")
```

#### Installation using R Studio
Please note, that dependencies have to be installed manually.

*Tools -> Install Packages -> Repository (CRAN) -> PDEnaiveBayes*


## 3. Additional Resources <a name="additional"/>

- View package on [CRAN](https://CRAN.R-project.org/package=PDEnaiveBayes)


### Manual

The full manual for users or developers is available here:
[Package documentation](https://CRAN.R-project.org/package=PDEnaiveBayes/PDEnaiveBayes.pdf)

## 4. References <a name="references"/>

[Ultsch, 2005]	Ultsch, A.: Pareto Density Estimation: A Density Estimation for Knowledge Discovery, in Baier, D.; Wernecke, K.-D. (Eds.), Innovations in Classification, Data Science, and Information Systems (Proc. 27th Annual GfKl Conference), Springer, Berlin, pp. 91–100, https://doi.org/10.1007/3-540-26981-9_12, 2005.

[John/Langley, 1995]	John, G. H. / Langley, P.: Estimating Continuous Distributions in Bayesian Classifiers, in Proceedings of the Eleventh Conference on Uncertainty in Artificial Intelligence (UAI–95), Morgan Kaufmann, San Mateo, pp. 338–345, 1995.

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 

## ----setup, include=FALSE-----------------------------------------------------
# setupKnitr()
# knitr::opts_chunk$set(echo = TRUE,
#                       fig.align = "center",
#                       warning = FALSE,
#                       webgl = TRUE,
#                       fig.width = 8, 
#                       fig.height = 8,
#                       fig.keep = "all",
#                       fig.ext = "jpeg"
#                       )

## ----fig.width=5, fig.height=5------------------------------------------------
library(PDEnaiveBayes)

data(Hepta)
Data = Hepta$Data
Cls = Hepta$Cls

## ----fig.width=5, fig.height=5------------------------------------------------

model = Train_naiveBayes(Data, 
                          Cls, 
                          Gaussian = FALSE, 
                          Plausible = FALSE)
table(Cls, model$ClsTrain)


## ----fig.width=5, fig.height=5, eval = FALSE----------------------------------
# 
# if (requireNamespace("parallel")) {
# 
#   library(parallel)
# 
#   train_index = sample(1:nrow(Data), 0.8 * nrow(Data))
# 
#   train_data = Data[train_index, ]
#   test_data = Data[-train_index, ]
#   train_cls = Cls[train_index]
#   test_cls = Cls[-train_index]
# 
#   num_cores = detectCores() - 1
#   cl = makeCluster(num_cores)
#   if (requireNamespace("FCPS")) {
#     data(Hepta)
#     Data = Hepta$Data
#     Cls  = Hepta$Cls
#     model_mc = Train_naiveBayes_multicore(
#       cl       = cl,
#       Data     = train_data,
#       Cls      = train_cls,
#       Predict  = TRUE
#     )
#     table(train_cls, model_mc$ClsTrain)
#   }
# }

## ----fig.width=5, fig.height=5, eval=FALSE------------------------------------
# preds = Predict_naiveBayes(test_data, model_mc)
# table(test_cls, preds$Cls)

## ----fig.width=5, fig.height=5------------------------------------------------
PlotLikelihoodFuns(
  LikelihoodFuns = model$Model$PDFs_funs,
  Data        = Hepta$Data
)

## ----fig.width=5, fig.height=5, eval = FALSE----------------------------------
# # CRAN does not allow interactive plotly plots, hence the plot is given as a static image
# PlotNaiveBayes(Model = model$Model, FeatureNames = colnames(Hepta$Data))

## ----fig.width=5, fig.height=5------------------------------------------------
PlotBayesianDecision2D(
  X          = Hepta$Data[, 1],
  Y          = Hepta$Data[, 2],
  Posteriors = model$Posteriors,
  Class      = 2,
  xlab = "X1",
  ylab = "X2"
)

## ----fig.width=5, fig.height=5------------------------------------------------
PlotPosteriors(
  Data       = Data,
  Posteriors = model$Posteriors,
  Class      = 2)

## ----fig.width=5, fig.height=5------------------------------------------------

x1 = rnorm(1000,0,2)
x2 = rnorm(1000,5,5.5)
x = c(x1, x2)

f1 = dnorm(x, 0, 2)
f2 = dnorm(x, 5, 5.5)
p1 = f1 / (f1 + f2)
comp = rbinom(2000, 1, p1) + 1

model_1d = Train_naiveBayes(as.matrix(x), 
                          comp, 
                          Gaussian = FALSE, 
                          Plausible = TRUE,
                          PlotIt = TRUE)

KernelPlausibleLikelihoods=function(PDFs_funs,c_Kernels_list,Data,Cls,PlausibleCenters=NULL,threshold=1e-12){
  
  #umweg ueber data
  if(is.null(PlausibleCenters)){
    if(isTRUE(requireNamespace("FCPS",quietly = T))){
      CenterPerClass=FCPS::ClusterApply(DataOrDistances = Data,FUN = CenterEstimate,Cls = Cls,Simple = T)
      PlausibleCenters=t(CenterPerClass)
    }else{
      stop("KernelPlausibleLikelihoods: For Plausible=T FCPS package needs to be installed")
    }
  }
  c_Kernels_list_stand=c_Kernels_list
  N=max(vapply(c_Kernels_list,nrow,integer(1)))
  d=length(c_Kernels_list)
  class_l=ncol(c_Kernels_list[[1]])
  for(i in 1:d){
    x=c_Kernels_list[[i]]
    vec=seq(from=min(x,na.rm=T),to=max(x,na.rm=T),length.out=N)
    xy=vec
    for(k in 2:class_l){
      xy=cbind(xy,vec)
    }
    c_Kernels_list_stand[[i]]=xy
  }
  Likelihoods_PrePlausible=interpolatePDF4Kernels(KernelList = c_Kernels_list_stand,PDFfunList =PDFs_funs )
  PDFs_funs_PrePlausible=PDFs_funs
  KernelData=c_Kernels_list_stand[[1]][,1,drop=F]
  if(d>1){
    for(f in 2:d){
      KernelData = cbind(KernelData,c_Kernels_list_stand[[f]][,1])
    }
  }
  #ListOfLikelihoodsV=KernelPlausibleLikelihoods(ListOfLikelihoods = Likelihoods_PrePlausible,KernelsList=c_Kernels_list,PlausibleCenters=PlausibleCenters,threshold=Threshold)
  ListOfLikelihoodsV=PlausibleLikelihoods(Likelihoods = Likelihoods_PrePlausible,Data = KernelData,Cls = Cls,threshold = threshold,PlausibleCenters = PlausibleCenters)
  Epsilon=ListOfLikelihoodsV$Epsilon
  ArrayOfLikelihoods=ListOfLikelihoodsV$Likelihoods
  #PlotLikelihoods(ArrayOfLikelihoods,KernelData)
  PDFs_funs=GetLikelihoodFunction(c_Kernels_list_stand,likelihoodsArray2List(ArrayOfLikelihoods))
  
  names(PDFs_funs)=colnames(Data)
  
  return(list(PDFs_funs=PDFs_funs,PlausibleCenters=PlausibleCenters,Epsilon=Epsilon,PlausibleLikelihoods=ArrayOfLikelihoods,Likelihoods_PrePlausible=Likelihoods_PrePlausible))
}  
#Diese Variante geht nicht, weil fixed_kmeans auf den kernel nicht gut funktioniert, wieso weis ich nicht  
  # #Likelihoods=likelihoodsArray2List(ListOfLikelihoods)#die array variante schein fehlerhaft zu sein
  # 
  # Likelihoods=ListOfLikelihoods
  # #Kernels=likelihoodsArray2List(KernelsList)#die array variante schein fehlerhaft zu sein
  # 
  # Kernels=listOfLikelihoods2Array(KernelsList)
  # #__________________________________________________________________________________________
  # #  Make Likelihoods plausible  ----
  # #__________________________________________________________________________________________
  # #Cases that have a probability of belonging to a particular class below 
  # #the threshold 𝜀 calculated above can either be left “unclassified”, 
  # #which is called “reasonable Bayes” classification, or their classification 
  # #can be based on the distance to the class center of neighborhood classified cases, 
  # #which is called “plausible Bayes” classification.
  # 
  # if(!is.list(Likelihoods)){
  #     V=dim(Likelihoods)
  #     N=V[1]
  #     class_l=V[2]
  #     d=V[3]
  #     BC_limit_vec=rep(NaN,d)
  #     #The threshold 𝜀 below which the evidence is considered as low is estimated 
  #     #for each pdf of the posteriors using a computed ABC analysis. Subset “C” of
  #     #this item categorization contains the nonprofitable values, i.e., the lowest 
  #     #probabilities (“the trivial many”). 
  #     for( f in 1:d){
  #       #skip variable if any centers lie very near to each other
  #       #nearer than 5%
  #       Kernels_cur=Kernels[,1,f]#as.vector(Kernels[,,f])#second dimension is per class, we take all classes as this dim is irrelevant
  #       #vermutlich sinnvoller nur mode weitesten links mit weitesten rechts zu vergleichen
  #       # diffs=unlist(sapply(1:(class_l-1),function(lag,x) diff(x,lag),PlausibleCenters[f,]))
  #       diffs=abs(max(PlausibleCenters[f,])-min(PlausibleCenters[f,]))
  #       check=min(abs(diffs))<diff(quantile4LargeVectors(Kernels_cur,c(0.5,0.6)),lag=1)
  #       if(isTRUE(check)){
  #         BC_limit_vec[f]=0
  #         next;
  #       }
  # 
  #       Prop_cur=apply(Likelihoods[,,f,drop=F],1,function(x) sum(x[is.finite(x)]))
  #    
  #       if(isTRUE(requireNamespace("ABCanalysis",quietly=TRUE))){
  #         ## identify critical points below threshold ----
  #         #The threshold 𝜀 for a class assignment probability that is considered 
  #         #too low for the assignment decision is defined as the BC limit calculated 
  #         #in the ABC analysis. In this way, the “trivial many” probabilities are 
  #         #removed from the Bayesian class assignment decision
  #         
  #           V=ABCanalysis::ABCanalysis(Prop_cur)
  #           BClimit=V$BCLimit
  # 
  #         
  #         #gibts erst ab 1.2.2
  #         # ABCcls=ABCanalysis::ABCclassification(exp(ClassDecisionLogProb[,i]))
  #         # UnReasonableInd=which(ABCcls>2)#klasse drei is plausible oder reasonable
  #         BC_limit_vec[f]=BClimit
  #         UnReasonableInd <- which(Prop_cur < class_l*BClimit)
  #       }else{#stupid fallback
  #         warning("Please install ABCanalysis")
  #         UnReasonableInd <- which(Prop_cur < 0.1)
  #         
  #       }#end check if package installed
  #       #ab hier ist noch ein bug drin, da falsch korrigiert wird
  #       #cases that are unreasonable are made plausible
  #       #alternative would be to return class assignment of zero (unknown)
  #       #if any liklihood has and unreasonable row for a specific case
  #       #this would be too strong
  #       if(length(UnReasonableInd)>0){
  #         #Kernels_cur=seq(from=min(Kernels_cur),to=max(Kernels_cur),length.out=N)#sort(sample(unique(Kernels_cur),N))#size has to be the same as Likelihoods and in ascending rank
  #         #funktioniert nur, da PlausibleCenters in reihenfolge der klassen ist!
  #         Cls=fixed_kmeans(Kernels_cur,Centers = PlausibleCenters[f,]) #klasse k=1:length(PlausibleCenters)
  #         #likelihoods sind in der reihenfolge der datenpunkte
  #         #funktioniert nur, da likelihood spalten in reihenfolge der klassen sind!
  #         LikelihoodCls=max.col(Likelihoods[,,f,drop=F],ties.method = "first")#apply(LL, 1, which.max)#klasse k=1:nrow(LikelihoodCls)
  #         #identify ALL rows that are not plausible at the moment EVEN if above threshold
  #         diff_ind_x_all=which(LikelihoodCls!=Cls)
  #         #select only rows that are below threshold
  #         diff_ind_x=intersect(UnReasonableInd,diff_ind_x_all)
  #         # #epsilon change makes sure that the correct ones are now really maximal, i.e.,
  #         # #the likelihood that the class with the closest distance is assigned to x is maximal and infuences in the right direction
  #         # #the Posteriors
  #         eps    <- threshold#^(1/4)#set epsilon
  #         if(length(diff_ind_x)>0){#switch them to plaubile ones
  #           for(p in 1:length(diff_ind_x)){
  #             #which the inplausubile lilihood has a current maximum
  #             colcur_max=LikelihoodCls[diff_ind_x[p]]
  #             #the current maximum value of the implausible class 
  #             Max_cur=Likelihoods[diff_ind_x[p],colcur_max,f,drop=F]
  #             #what is the class likelihood that should have be maximal
  #             col_should_max=Cls[diff_ind_x[p]]
  #             #lower the maximum value of the inplausible class with eps
  #             Likelihoods[diff_ind_x[p],colcur_max,f]=Max_cur-eps
  #             #replace with the value that is currently maximal and add eps
  #             Likelihoods[diff_ind_x[p],col_should_max,f]=Max_cur+eps
  #             #do not change likelihoods for other classes
  #           }
  #         }
  #       }#end length(UnReasonableInd)>0)
  #     }#end for( f in 1:d)
  # }else{# Likelihoods is a list - (old legacy code)
  #   ListOfLikelihoods=Likelihoods
  #   ListOfLikelihoods_plausible=ListOfLikelihoods
  #   d=length(ListOfLikelihoods)#fuer jedes features gibt es ein listenelement
  #   N=nrow(ListOfLikelihoods[[1]])#durch interpolateDistributionOnData ist jedes listenelement der gleichen laenge gleich laenge der daten
  #   #If the “plausible Bayes” calculation results
  #   #in an undefined value, the class with the closest distance is assigned to x
  #   #: class(x) = argmin({d(x,m1), …, d(x,m𝑐)}).
  #   class_l=ncol(ListOfLikelihoods[[1]])
  #   #print(class_l)
  #   BC_limit_vec=rep(NaN,d)
  #   for( f in 1:d){
  #     #skip variable if any centers lie very near to each other
  #     #nearer than 5%
  #     LL=ListOfLikelihoods[[f]]
  #     Kernels_cur=Kernels[[f]][,1]#as.vector(Kernels[[f]])
  # 
  #     diffs=abs(max(PlausibleCenters[f,])-min(PlausibleCenters[f,]))
  #     check=min(abs(diffs))<diff(quantile(Kernels_cur,c(0.5,0.6)),lag=1)
  #     if(isTRUE(check)){
  #       BC_limit_vec[f]=0
  #       next;
  #     }
  #     #Kernels_cur=sort(sample(Kernels_cur,nrow(LL)))#but they have to be in the righ ranks to fit Likelihoods!
  # 
  #     #funktioniert nur, da PlausibleCenters in reihenfolge der klassen ist!
  #     Cls=fixed_kmeans(Kernels_cur,Centers = PlausibleCenters[f,]) #klasse k=1:length(PlausibleCenters)
  #     #likelihoods sind in der reihenfolge der datenpunkte
  #     #funktioniert nur, da likelihood spalten in reihenfolge der klassen sind!
  #     LikelihoodCls=max.col(LL,ties.method = "first")#apply(LL, 1, which.max)#klasse k=1:nrow(LikelihoodCls)
  #     #select wich likeliehoods are not plausbile
  #     #for debugging
  #     # x=Data[,f]
  #     # ind=order(x,decreasing = F)
  #     # plot(x[ind],LL[ind,1],type="l",col=DataVisualizations::DefaultColorSequence[1],main="Bayesian",lwd=2)
  #     # for(i in 2:ncol(LL)){
  #     #   points(x[ind],LL[ind,i],type="l",col=DataVisualizations::DefaultColorSequence[i],lwd=2)
  #     # }
  #     # DataVisualizations::DefaultColorSequence[1:3]
  #     # points(x,rep(0.2,length(x)),col=DataVisualizations::DefaultColorSequence[Cls])
  #     #
  #     #kein plan wieso das falsch ist
  #     # diff_ind_x <- which(LikelihoodCls != Cls)
  #     # if(length(diff_ind_x) > 0){
  #     #   col_from <- LikelihoodCls[diff_ind_x]
  #     #   col_to   <- Cls[diff_ind_x]
  #     #   eps    <- threshold^(1/4)
  #     #   Max_cur <- LL[diff_ind_x,col_from]
  #     #   LL[diff_ind_x,col_to] <- Max_cur + eps
  #     #   LL[diff_ind_x,col_from] <- Max_cur - eps
  #     # }
  #     
  #     diff_ind_x=which(LikelihoodCls!=Cls)#identify the rows that are not plausible at the moment
  #     # #epsilon change makes sure that the correct ones are now really maximal, i.e.,
  #     # #the likelihood that the class with the closest distance is assigned to x is maximal and infuences in the right direction
  #     # #the Posteriors
  #     eps    <- threshold#^(1/4)#set epsilon
  #     if(length(diff_ind_x)>0){#switch them to plaubile ones
  #       for(p in 1:length(diff_ind_x)){
  #         #which the inplausubile lilihood has a current maximum
  #         colcur_max=LikelihoodCls[diff_ind_x[p]]
  #         #the current maximum value of the implausible class 
  #         Max_cur=LL[diff_ind_x[p],colcur_max]
  #         #what is the class likelihood that should have be maximal
  #         col_should_max=Cls[diff_ind_x[p]]
  #         #lower the maximum value of the inplausible class with eps
  #         LL[diff_ind_x[p],colcur_max]=Max_cur-eps
  #         #replace with the value that is currently maximal and add eps
  #         LL[diff_ind_x[p],col_should_max]=Max_cur+eps
  #         #do not change likelihoods for other classes
  #       }
  #     }
  #     
  #     # #store them
  #     ListOfLikelihoods_plausible[[f]]=LL
  #     #fore debugging
  #     # plot(x[ind],LL[ind,1],type="l",col=DataVisualizations::DefaultColorSequence[1],main="Bayesian",lwd=2)
  #     # for(i in 2:ncol(LL)){
  #     #   points(x[ind],LL[ind,i],type="l",col=DataVisualizations::DefaultColorSequence[i],lwd=2)
  #     # }
  #     # DataVisualizations::DefaultColorSequence[1:3]
  #     # points(x,rep(0.2,length(x)),col=DataVisualizations::DefaultColorSequence[Cls])
  #   }
  #   
  #   #The threshold 𝜀 below which the evidence is considered as low is estimated 
  #   #for each pdf of the posteriors using a computed ABC analysis. Subset “C” of
  #   #this item categorization contains the nonprofitable values, i.e., the lowest 
  #   #probabilities (“the trivial many”). 
  #   for( f in 1:d){
  #     LL=ListOfLikelihoods[[f]]
  #     Prop_cur=apply(LL,1,function(x) sum(x[is.finite(x)]))
  #     
  #     
  #     if(isTRUE(requireNamespace("ABCanalysis",quietly=TRUE))){
  #       ## identify critical points below threshold ----
  #       #The threshold 𝜀 for a class assignment probability that is considered 
  #       #too low for the assignment decision is defined as the BC limit calculated 
  #       #in the ABC analysis. In this way, the “trivial many” probabilities are 
  #       #removed from the Bayesian class assignment decision
  # 
  #         V=ABCanalysis::ABCanalysis(Prop_cur)
  #         BClimit=V$BCLimit
  #  
  #       #gibts erst ab 1.2.2
  #       # ABCcls=ABCanalysis::ABCclassification(exp(ClassDecisionLogProb[,i]))
  #       # UnReasonableInd=which(ABCcls>2)#klasse drei is plausible oder reasonable
  #       if(is.nan(BC_limit_vec[f])){
  #         BC_limit_vec[f]=BClimit
  #         UnReasonableInd <- which(Prop_cur < class_l*BClimit)
  #       }else{
  #         #centers sind zu nahe beieiander
  #         UnReasonableInd=NULL
  #       }
  # 
  #     }else{#stupid fallback
  #       warning("Please install ABCanalysis")
  #       UnReasonableInd <- which(Prop_cur < 0.1)
  #       
  #     }#end check if package installed
  #     #cases that are unreasonable are made plausible
  #     #alternative would be to return class assignment of zero (unknown)
  #     #if any liklihood has and unreasonable row for a specific case
  #     #this would be too strong
  #     if(length(UnReasonableInd)>0){
  #       #Cases that have a probability of belonging to a particular class below 
  #       #the threshold 𝜀 ... have to be in plausible categorization
  #       #reset likelihoods on these UnReasonableInd for prediction
  #       LL=ListOfLikelihoods[[f]]
  #       LL2=ListOfLikelihoods_plausible[[f]]
  #       LL[UnReasonableInd,]=LL2[UnReasonableInd,]
  #       ListOfLikelihoods[[f]]=LL
  #       # print(f)
  #       # print(LL[UnReasonableInd,])
  #       # print(LL2[UnReasonableInd,])
  #     }#end length(UnReasonableInd)>0)
  #   }#end for( f in 1:d)
  #   Likelihoods=listOfLikelihoods2Array(ListOfLikelihoods)
  #   
  # }#end if ! is list Likelihoods 
  #   return(list(Likelihoods=Likelihoods,Epsilon=BC_limit_vec))
#}

Train_naiveBayes_multicore=function(cl=NULL,Data,Cls,Plausible=TRUE,Predict=FALSE,Priors, UseMemshare=FALSE,...){
  
  if(!is.matrix(Data)){
    warning("Train_naiveBayes_multicore: Data is not matrix, calling as.matrix()")
    Data=as.matrix(Data)
  }
  if(!is.numeric(Cls)){
    warning("Train_naiveBayes: Cls is not numneric, trying to transform.")
    if(isTRUE(requireNamespace("FCPS",quietly = T))){
      V=FCPS::ClusterCreateClassification(Cls)
      Cls=V$Cls
      message(paste0(V$ClusterNames,collapse = " "))
      message("transformed to")
      message(paste0(names(V$ClusterNames),collapse = " "))
    }else{
      warning("Train_naiveBayes: Please install FCPS package for stable Cls transformation.")
      Cls=as.numeric(Cls)
    }
  }
  extra_args <- list(...)
  
  if (
    length(extra_args) > 0L &&
    (is.null(names(extra_args)) || any(!nzchar(names(extra_args))))
  ) {
    stop(
      "Train_naiveBayes_multicore: all arguments in ... must be named."
    )
  }
  if (
    "ParetoRadiusPerFeauture" %in% names(extra_args) &&
    length(extra_args$ParetoRadiusPerFeauture) > 1L
  ) {
    stop(
      "Train_naiveBayes_multicore: a feature-specific ",
      "ParetoRadiusPerFeauture is currently unsupported."
    )
  }
  N  <- nrow(Data)
  nc <- length(Cls)
  
  if (N != nc) {
    warning(
      "Train_naiveBayes_multicore: length of Cls unequal to ",
      "nrow(Data); shortening input."
    )
    
    n_keep <- min(N, nc)
    
    if (n_keep == 0L) {
      stop("Train_naiveBayes_multicore: no observations available.")
    }
    
    ind  <- seq_len(n_keep)
    Data <- Data[ind, , drop = FALSE]
    Cls  <- Cls[ind]
  }
  
  bool_fin <- is.finite(Cls)
  
  if (!all(bool_fin)) {
    warning(
      "Train_naiveBayes_multicore: deleting observations ",
      "with non-finite class labels."
    )
    
    Cls  <- Cls[bool_fin]
    Data <- Data[bool_fin, , drop = FALSE]
  }
  
  if (length(Cls) == 0L) {
    stop("Train_naiveBayes_multicore: no finite class labels remain.")
  }
  
  if (length(unique(Cls)) < 2L) {
    stop("Train_naiveBayes_multicore: at least two classes are required.")
  }
  
  if(missing(Priors))
    Priors=getPriors(Cls)
  
  feature_names <- colnames(Data)
  
  if (is.null(feature_names)) {
    feature_names <- paste0("X", seq_len(ncol(Data)))
    colnames(Data) <- feature_names
  }
  
  if(is.null(cl)){#for debugging purposes
    message("Train_naiveBayes_multicore: is called with a single core without memshare. 'cl=NULL' is only for debugging purposes. ")
    List=apply(Data,MARGIN = 2,function(x,Cls,Priors,Plausible,...){
      col=as.matrix(x)
      return(Train_naiveBayes(Data = col,Cls = Cls,Priors=Priors,Plausible=Plausible,Predict = FALSE,...)$Model)
    },Cls,Priors,Plausible,...)
  }else if (UseMemshare) {
    #memshare accepts only double in Vars
    if(isTRUE(Plausible)){
      Plausible_mem=1
    }else{
      Plausible_mem=0
    }
    stamp <- gsub(
      "[^0-9]",
      "",
      format(Sys.time(),"M%OS6")
    )
    mem_namespace <- sprintf(
      "PB_%x_%s",
      as.integer(Sys.getpid()),
      stamp
    )
    on.exit({
      memshare::memshare_gc(namespace = mem_namespace,cluster = cl)
    })
    
    List=memshare::memApply(X = Data, MARGIN = 2, FUN = function(x,Cls,Priors,Plausible,...) 
      PDEnaiveBayes::Train_naiveBayes(Data = x,Cls = Cls,Priors=Priors,Plausible=Plausible,Predict = FALSE,...)$Model, 
      NAMESPACE = mem_namespace, CLUSTER = cl,VARS = list(Cls=Cls,Priors=Priors,Plausible=Plausible_mem,...))


  } else {
    List=parallel::parApply(cl = cl,X = Data,MARGIN = 2,FUN = function(x,Cls,Priors,Plausible,...){
      col=as.matrix(x)
      m=PDEnaiveBayes::Train_naiveBayes(Data = col,Cls = Cls,Priors=Priors,Plausible=Plausible,Predict = FALSE,...)$Model
      return(m)
    },Cls,Priors,Plausible,...)
  }
  
  if (!is.list(List) || length(List) != ncol(Data)) {
    stop("Train_naiveBayes_multicore: A training backend did not return the correct model, use 'cl=NULL' for debugging.")
  }
  
  if(is.null(names(List))){
    #minor bur memshare: returned list ist without names sometimes
    names(List)=feature_names
  }
  valid_models <- vapply(
    List,
    function(x) is.list(x) && !is.null(names(x)),
    logical(1)
  )
  
  if (!all(valid_models)) {
    stop(
      "Train_naiveBayes_multicore: Feature training failed for: ",
      paste(feature_names[!valid_models], collapse = ", ")
    )
  }
  
  model_component_names <- names(List[[1L]])

  same_structure <- vapply(
    List,
    function(x) identical(names(x), model_component_names),
    logical(1)
  )
  
  if (!all(same_structure)) {
    stop("Train_naiveBayes_multicore: Per-feature models do not have identical components, use 'cl=NULL' for debugging.")
  }
  
  model=new.env()
  for(i in seq_along(model_component_names)){
    nestedlist_cur=lapply(List, "[[",i)
    if(length(nestedlist_cur[[1]][[1]])>1){
      flattened_list_cur <- lapply(nestedlist_cur, function(x) x[[1]])
      names(flattened_list_cur) <- names(nestedlist_cur)
    }else{
      flattened_list_cur=nestedlist_cur
    }
    assign(model_component_names[i],flattened_list_cur,envir = model)
  }
  model=as.list(model)
  
  Gaussian <- if ("Gaussian" %in% names(extra_args)) {
    isTRUE(extra_args$Gaussian)
  } else {
    FALSE
  }
  model$Plausible=Plausible && !Gaussian


  radii <- lapply(
    List,
    `[[`,
    "ParetoRadiusPerFeauture"
  )
  
  scalar_radii <- vapply(
    radii,
    function(x) {
      is.atomic(x) &&
        is.null(dim(x)) &&
        length(x) == 1L
    },
    logical(1)
  )
  
  matrix_radii <- vapply(
    radii,
    function(x) {
      is.matrix(x) && ncol(x) == 1L
    },
    logical(1)
  )
  
  if (all(scalar_radii)) {
    
    model$ParetoRadiusPerFeauture <- setNames(
      vapply(
        radii,
        function(x) as.numeric(x)[1L],
        numeric(1)
      ),
      feature_names
    )
    
  } else if (all(matrix_radii)) {
    
    model$ParetoRadiusPerFeauture <- do.call(
      cbind,
      radii
    )
    
    colnames(model$ParetoRadiusPerFeauture) <-
      feature_names
    
  } else {
    
    stop(
      "Train_naiveBayes_multicore: inconsistent ",
      "Pareto-radius structures between features."
    )
  }
  
  if(isTRUE(model$Plausible)){
    if(is.list(model$PlausibleCenters)){
      model$PlausibleCenters=do.call(rbind,model$PlausibleCenters)
      rownames(model$PlausibleCenters)=names(List)
    }
  }
  
  model$Priors <- List[[1L]]$Priors
  
  if ("Thetas" %in% names(model)) {
    theta_is_null <- vapply(
      model$Thetas,
      is.null,
      logical(1)
    )
    
    if (all(theta_is_null)) {
      model$Thetas <- NULL
    } else if (any(theta_is_null)) {
      stop(
        "Train_naiveBayes_multicore: Gaussian theta matrices ",
        "are missing for some features."
      )
    } else {
      names(model$Thetas) <- feature_names
    }
  }
  
  class(model) <- "PDEbayes"
  
  if(isTRUE(Predict)){
    res=Predict_naiveBayes(Data=Data,Model=model,  cl = cl,...)
    return(list(
      Model=model,
      ClsTrain=res$Cls,
      Posteriors=res$Posteriors
    ))
  }else{
    return(list(
      Model=model,
      ClsTrain=NULL,
      Posteriors=NULL
    ))
  }
}

listOfLikelihoods2Array=function(ListOfLikelihoods){
  #to do try:simplify2array
  if(is.list(ListOfLikelihoods)){
  d=length(ListOfLikelihoods)
  class_l=max(unique(sapply(ListOfLikelihoods, ncol)))
  N=max(unique(sapply(ListOfLikelihoods, nrow)))
  
  ArrayOfLikelihoods=array(NaN,c(N,class_l,d))
  for(i in 1:d){
    LL=ListOfLikelihoods[[i]]
    ArrayOfLikelihoods[1:nrow(LL),1:ncol(LL),i]=LL
  }
    return(ArrayOfLikelihoods)
  }else{
    return(ListOfLikelihoods)
  }
}

PlotPosteriors=function(Data,Posteriors,Class=1,CellColorsOrPallette,Showpoints=TRUE,NoBins,ShowLegend=TRUE){
# PlotPosteriors(Data,Posteriors)
#Input
#Data                         numerical data matrix [1:n,1:k]
# Posteriors[1:n, 1:l]    Numeric matrices with posterior probabilities
#                         according to the bayesian Theorem.
#OPTIONAL
# Class             scalar defining which class to look at
# CellColorsOrPallette function defining the color palette or a character vector of colors
# Showpoints         TRUE, Punkte werden angezeigt
# NoBins             number of bins for class posteriori
# ShowLegend         TRUE, show one posterior legend for all pairwise plots
  
#OUTPUT
# ggplot2 object
#Note: gridExtra is required only for arranging multidimensional plots
#author MCT 05/2025
  
  if(!is.matrix(Posteriors)){
    Posteriors=as.matrix(Posteriors)
  }
  if(!is.matrix(Data)){
    xlab = deparse1(substitute(Data)) #has to come before vectorization in next line
    Data=as.vector(Data)
    ind=order(Data)
    # plot(Data[ind],Posteriors[ind,1],col=DatabionicSwarm::DefaultColorSequence[3],type="l",lwd=2,xlab=xlab,ylab="Posteriors")
    # for(i in 2:ncol(Posteriors)){
    #   points(Data[ind],Posteriors[ind,i],col=DatabionicSwarm::DefaultColorSequence[3+i],type="l",lwd=2)
    # }
    X = Data[ind]
    P = Posteriors[ind, , drop = FALSE]
    if (is.null(colnames(P))) colnames(P) = paste0("Class ", seq_len(ncol(P)))
    
    # Make a "long" data.frame using base R only
    long = stack(as.data.frame(P))                  # columns -> rows; gives 'values' and 'ind'
    long$x    = rep(X, times = ncol(P))             # replicate x for each class
    long$ord  = rep(seq_along(X), times = ncol(P))  # order index to preserve original sequence
    long      = long[order(long$ord), ]
    names(long)[names(long) == "values"] = "posterior"
    names(long)[names(long) == "ind"]    = "class"
    
    # Colors: indices 3, 4, 5, ... as in your base plot
    if(requireNamespace("DataVisualizations",quietly = TRUE)){
      pal = DataVisualizations::DefaultColorSequence[3 + 0:(ncol(P) - 1L)]
    }else{
      pal = rep(NA_character_,ncol(P))
    }
    if(length(pal)!=ncol(P) || anyNA(pal)){
      pal = grDevices::hcl.colors(ncol(P),palette = "Dark 3")
    }
    
    # ggplot2 issues of finding variables by names in datastructures
    # CMD check requires the variable to be known (initiated)
    # Naming the variables in a list/data.frame/... is not enough (for CMD check)
    x = NULL
    y = NULL
    posterior = NULL
    class = NULL
    
    p=ggplot(long, aes(x = x, y = posterior, color = class, group = class)) +
      geom_path(linewidth = 1) +
      scale_color_manual(values = pal, name = "Class") +
      labs(x = xlab, y = "Posterior") +
      coord_cartesian(ylim = c(0,1)) +
      theme_minimal()
    
    
    return(p)
  }else{
    d=ncol(Data)
    if(d<=1){
      warning("PlotPosteriors: Data has to have more than one dimension. Cannot plot.")
      return(invisible(NULL))
    }
    if(!requireNamespace("gridExtra",quietly = TRUE)){
      warning("PlotPosteriors: Please install package gridExtra")
      return(NULL)
    }
    feature_names=colnames(Data)
    if(is.null(feature_names)){
      feature_names=paste0("V",seq_len(d))
    }
    ggobjList=list()
    LegendMapping=NULL
    k=1
    for(x in 1:d){
      if(x<d)
        for(y in (x+1):d){
          Arguments=list(X = Data[,x], Y = Data[,y],
                         Posteriors = Posteriors,
                         Class = Class,
                         Showpoints = Showpoints, PlotIt = FALSE,
                         xlab=feature_names[x],
                         ylab=feature_names[y])
          if(!missing(NoBins)){
            Arguments$NoBins=NoBins
          }
          if(!missing(CellColorsOrPallette)){
            Arguments$CellColorsOrPallette=CellColorsOrPallette
          }
          Result=tryCatch(
            do.call(PlotBayesianDecision2D,Arguments),
            error=function(e){
              warning(paste0("PlotPosteriors: Could not create plot for ",
                             feature_names[x]," and ",feature_names[y],": ",
                             conditionMessage(e)),call. = FALSE)
              NULL
            }
          )
          if(isTRUE(ShowLegend) && is.null(LegendMapping) &&
             is.list(Result) && !is.null(Result$Mapping) &&
             "Color" %in% colnames(Result$Mapping) &&
             length(Result$Mapping$Color)>=2L){
            LegendMapping=Result$Mapping
          }
          if(is.list(Result) && !is.null(Result$GGobj)){
            obj=Result$GGobj
          }else{
            obj=NULL
          }
          if(!is.null(obj)){
            ggobjList[[k]]=obj
          }else{
            ggobjList[[k]]=ggplot2::ggplot()
          }

          k=k+1
        }
    }
    numPlots = length(ggobjList)
    ColNo=max(1L,ceiling(sqrt(numPlots)))

    LegendGrob=NULL
    if(isTRUE(ShowLegend) && !is.null(LegendMapping)){
      PDEnaiveBayes.legend.posterior=NULL
      LegendPlot=ggplot2::ggplot(
        data.frame(PDEnaiveBayes.legend.posterior=c(0,1)),
        ggplot2::aes(x=1,
                     y=PDEnaiveBayes.legend.posterior,
                     fill=PDEnaiveBayes.legend.posterior)
      ) +
        ggplot2::geom_tile() +
        ggplot2::scale_fill_gradientn(
          colours=as.character(LegendMapping$Color),
          limits=c(0,1),
          breaks=c(0,0.25,0.5,0.75,1),
          name="Posterior"
        ) +
        ggplot2::theme_void() +
        ggplot2::theme(legend.position="right")

      LegendTable=ggplot2::ggplotGrob(LegendPlot)
      LegendIndex=which(grepl("^guide-box",LegendTable$layout$name))
      LegendIndex=LegendIndex[
        !vapply(LegendTable$grobs[LegendIndex],
                inherits,logical(1),"zeroGrob")
      ]
      if(length(LegendIndex)>0L){
        LegendGrob=LegendTable$grobs[[LegendIndex[1L]]]
      }
    }

    ArrangeArguments=c(ggobjList,list(ncol=ColNo))
    if(!is.null(LegendGrob)){
      ArrangeArguments$right=LegendGrob
    }
    p=do.call(gridExtra::grid.arrange,ArrangeArguments)
    
    return(p)
  }
}

PlotLikelihoods=function(Likelihoods,Data,PlausibleLikelihoods=NULL,Epsilon=NULL,PlausibleCenters=NULL,PlotCutOff=4,xlim){
# PlotLikelihoods(Likelihoods,Data)
#Input
#Likelihoods            List of d numeric matrices, one per feature, each matrix with 1:k columns containing 
#                             the distribution of class 1:k per feature, i.e., the Likelihood per class
#Data                         numerical data matrix [1:n,1:k]
#OPTIONAL
#PlausibleLikelihoods     List of d numeric matrices of plausible Likelihood, one per feature, each matrix with [1:n,1:k] with k columns containing 
#                             the distribution of class 1:k per feature 
#Epsilon                      scalar defining epsiolon fo plausible likelihoods
#PlausibleCenters             [d:k] plausible centers used to compute plausible likelihoods
#PlotCutOff                   scalar defining the how many feature starting from 1 should ne plottet or numerical vector defining the index of features to be plottet
#                             in second case should not be too many otherwise plot yields an error
#OUTPUT
# native plots in grid

#author MCT 05/2025
#reset par after plotting
if(!is.matrix(Data)){
  Data=as.matrix(Data)
}
if(is.list(Likelihoods)){
  Likelihoods=listOfLikelihoods2Array(Likelihoods)
}
if(is.list(PlausibleLikelihoods)){
  PlausibleLikelihoods=listOfLikelihoods2Array(PlausibleLikelihoods)
}
def.par <- par(no.readonly = TRUE)
on.exit(par(def.par))
if(is.list(Likelihoods)){
  #number of dimensions
  dinit = length(Likelihoods)
  class_l=ncol(Likelihoods[[1]])
  # Set layout: 2 rows per item in Likelihoods
}else{
  V=dim(Likelihoods)
  N=V[1]
  class_l=V[2]
  dinit=V[3]
}


if(!is.null(PlotCutOff)){
  if(is.numeric(PlotCutOff)){
    if(length(PlotCutOff)==1){
      d=min(c(dinit,PlotCutOff),na.rm = T)
      rangePlots=1:d
    }else{
      rangePlots=PlotCutOff
      rangePlots[rangePlots>dinit]=dinit
      rangePlots=rangePlots[rangePlots>0]
      d=max(rangePlots,na.rm = T)-min(rangePlots,na.rm = T)+1
    }
  }else{
    rangePlots=1:dinit
    d=dinit
  }
}else{
  rangePlots=1:dinit
  d=dinit
}

if(is.null(PlausibleLikelihoods)){
  par(mfrow = c(d, 1))  # d rows, 1 columns
}else{
  par(mfrow = c(d, 2))  # d rows, 2 columns
}

if(isTRUE(requireNamespace("DataVisualizations",quietly = T))){
  Colors=DataVisualizations::DefaultColorSequence[1:class_l]
}else{
  warning("PlotLikelihoods: Please install DataVisualizations package for color setting.")
  Colors=rep(c("black","blue","green","gold","magenta","red","grey","orange"),class_l)
  Colors=Colors[1:class_l]
}

#plot all dimensions

for(f in rangePlots){
  x=Data[,f]
  ind=order(x,decreasing = F)
  fnam=colnames(Data)[f]
  if(is.list(Likelihoods)){
    LL=Likelihoods[[f]]
  }else{
    LL=Likelihoods[,,f]
  }

  maxY1=apply(LL,2,function(x) max(x[is.finite(x)]))
  
  if(!is.null(PlausibleLikelihoods)){
    if(is.list(Likelihoods)){
      LL2=PlausibleLikelihoods[[f]]
    }else{
      LL2=PlausibleLikelihoods[,,f]
    }
    maxY2=apply(LL2,2,function(x) max(x[is.finite(x)]))
    ylim=c(0,max(c(maxY1,maxY2)))
  }else{
    ylim=c(0,c(max(maxY1)))
  }
  
  if(missing(xlim)){
    plot(x[ind],LL[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],main="Bayesian",
         ylim=ylim,lwd=2,ylab="Class Likelihoods")
  }else{
    plot(x[ind],LL[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],main="Bayesian",
         ylim=ylim,lwd=2,ylab="Class Likelihoods",xlim=xlim)
  }
  for(i in 2:class_l){
    points(x[ind],LL[ind,i],type="l",col=Colors[i],ylim=ylim,lwd=2)
  }
  if(!is.null(PlausibleLikelihoods)){
    if(missing(xlim)){
      plot(x[ind],LL2[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],ylab="Class Likelihoods",
           main=paste0("Plausible Bayes with approx. Limit ",round(Epsilon[f]/class_l,3)),ylim=ylim,lwd=2)
    }else{
      plot(x[ind],LL2[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],ylab="Class Likelihoods",
           main=paste0("Plausible Bayes with approx. Limit ",round(Epsilon[f]/class_l,3)),ylim=ylim,lwd=2,xlim=xlim)
    }

    if(!is.null(PlausibleCenters)){
      PlausibleCenters_cur=PlausibleCenters[f,]
      abline(v=PlausibleCenters_cur[1],col=Colors[1])
    }
    for(i in 2:class_l){
      points(x[ind],LL2[ind,i],type="l",col=Colors[i],ylim=ylim,lwd=2)
      if(!is.null(PlausibleCenters)){
        abline(v=PlausibleCenters_cur[i],col=Colors[i])
      }
    }
  }#end in case of plausible
}#end for rangePlots
}

predict.PDEbayes= function(object, newdata, type = c("class", "response","prob"), ...) {
  if (missing(newdata)) {
    stop("Please supply `newdata` to predict()", call. = FALSE)
  }
  type <- match.arg(type)

  res=Predict_naiveBayes(Data=newdata,Model=object,...)
  
  switch (type,
          prob = {
            out=res$Posteriors
          },
          class={
            out=factor(res$Cls, levels = sort(unique(res$Cls)))
          },
          response={
            out=res$Cls
          }
  )
  return(out)
}

getPriors=function(Cls){
  # V=getPriors(Cls)
  # 
  # INPUT
  # Cls[1:n]    Numeric vector with class labels 
  # 
  # OUTPUT
  # Priors[1:l]    Numeric vector with prior probability for each class.
  # 
  # temp=table(Cls)
  # Priors=as.vector(temp/length(Cls))
  # names(Priors)=paste0("C",names(temp))
  
  V = rle(sort(Cls, method = "radix", decreasing = FALSE))
  countPerCluster = V$lengths
  uniqueClusters = V$values
  Priors=countPerCluster/sum(countPerCluster)
  names(Priors)=uniqueClusters
  
  return(Priors)
}

CenterEstimate=function(Feature,na.rm=TRUE,smallNo=100,...){
  #estimates the center of likelihood based on statistical estimators cited below
  #mode=CenterEstimate(Feature)
  #INPUT 
  #Feature        a feature [1:N]
  #OPTIONAL
  # na.rm    TRUE: filters for finite values
  # smallNo   no of points considered for a sample to be small,default 100, afaik undefined in [Ekblom,1972]
  # ...     further parameters for package modeest
  #OUTPUT
  # Mode  scalar value of Mode
  #Author MCT 05/2025
  if(isTRUE(na.rm)){
    Feature=Feature[is.finite(Feature)]
  }  
  if(!requireNamespace("modeest",quietly = T)){
    warning("CenterEstimate: Please install package modeest for approriate Plausible Center estimation, returning mean")
    if(length(Feature)>0)
      return(mean(Feature))
  }
  if(length(Feature)>=smallNo){
    #Bickel, D. R., & Fruehwirth, R.: On a fast, robust estimator of the mode: Comparisons to other robust estimators with applications. Computational Statistics & Data Analysis, 50(12), 3500-3530,2006.
    return(modeest::hsm(x = Feature,...))
  }else if(length(Feature)<smallNo&length(Feature)>3){
    #Ekblom, H.: A Monte Carlo investigation of mode estimators in small samples. Journal of the Royal Statistical Society: Series C (Applied Statistics), 21(2), 177-184,1972
    return(modeest::shorth(x = Feature,...))
  }else if(length(Feature)==3){
    #Bickel, D. R., & Fruehwirth, R.: On a fast, robust estimator of the mode: Comparisons to other robust estimators with applications. Computational Statistics & Data Analysis, 50(12), 3500-3530,2006.
    return(median(Feature))
  }else if(length(Feature)==2){
    return(mean(Feature))
  }else if(length(Feature)==1){
    return(Feature)
  }else if(length(Feature)==0){
    warning("CenterEstimate: returning zero as feature contains no information.")
    return(0)
  }else{
    warning("CenterEstimate: returning zero because something went wrong.")
    return(0)
  }
}

PlausibleLikelihoods=function(Likelihoods,Data,Cls,threshold=1e-4,c_PDFS_List=NULL,PlausibleCenters=NULL){
   #__________________________________________________________________________________________
  #  Make Likelihoods plausible  ----
  #__________________________________________________________________________________________
  #Cases that have a probability of belonging to a particular class below 
  #the threshold 𝜀 calculated above can either be left “unclassified”, 
  #which is called “reasonable Bayes” classification, or their classification 
  #can be based on the distance to the class center of neighborhood classified cases, 
  #which is called “plausible Bayes” classification.
  
  if(is.null(PlausibleCenters)){
    # In the one-dimensional case, a “plausible Bayes” classification is computed using
    #the distances from x to the modes (=maxima) or mean values {m1, ⋯, m𝑐} of the pdfs
    #of the class likelihoods. 
    if(!requireNamespace("FCPS",quietly = T)){
      stop("PlausibleLikelihoods: Please install package FCPS for Plausible Center estimation")
    }
    CenterPerClass=FCPS::ClusterApply(DataOrDistances = Data,FUN = CenterEstimate,Cls = Cls,Simple = T)
    
    #CenterPerClass=FCPS::ClusterApply(Data,modeest::mlv,Cls,Simple = T)
    PlausibleCenters=t(CenterPerClass)
    #cl=sort(unique(Cls),decreasing = F)
    #colnames(PlausibleCenters)=paste0("Class",cl)
  }

  if(!is.list(Likelihoods)){
      V=dim(Likelihoods)
      N=V[1]
      class_l=V[2]
      d=V[3]
      BC_limit_vec=rep(NaN,d)
      #The threshold 𝜀 below which the evidence is considered as low is estimated 
      #for each pdf of the posteriors using a computed ABC analysis. Subset “C” of
      #this item categorization contains the nonprofitable values, i.e., the lowest 
      #probabilities (“the trivial many”). 
      for( f in 1:d){
        if(N>1)
          LL=Likelihoods[,,f]#drop=F belaesst ein array
        else
          LL=as.matrix(Likelihoods[,,f,drop=F])
        #skip variable if any centers lie very near to each other
        #nearer than10%
        #vermutlich sinnvoller nur mode weitesten links mit weitesten rechts zu vergleichen
        # diffs=unlist(sapply(1:(class_l-1),function(lag,x) diff(x,lag),PlausibleCenters[f,]))
        #diffs=abs(max(PlausibleCenters[f,])-min(PlausibleCenters[f,]))
        pc_len=length(PlausibleCenters[f,])
        diffs <- as.matrix(dist(PlausibleCenters[f,],method="manhattan"))
        threshold_diff=diff(quantile4LargeVectors(Data[,f],c(0.5,0.6)),lag=1)
        
        check_ind=which(diffs<threshold_diff & upper.tri(diffs),arr.ind = T)#identify any diff only once
        if(nrow(check_ind)==(pc_len*(pc_len-1))/2){#all diffs are smaller
          BC_limit_vec[f]=0
          next;
        }
        #approximation posterior, priors are irrelevant here
        Prop_cur=apply(LL,1,function(x) prod(pmin(pmax(x[is.finite(x)],threshold),1-threshold)))
        # print(f)
        # print(Prop_cur)
        if(isTRUE(requireNamespace("ABCanalysis",quietly=TRUE))){
          ## identify critical points below threshold ----
          #The threshold 𝜀 for a class assignment probability that is considered 
          #too low for the assignment decision is defined as the BC limit calculated 
          #in the ABC analysis. In this way, the “trivial many” probabilities are 
          #removed from the Bayesian class assignment decision
          
          #dieser teil haengt davon ab, wie lang der vector und mit welchen werten ist
          #das unterscheidet sich in trainings und test daten
          #alu's paper definiert nicht, ueber welche daten die abc analyse getaetigt werden soll
          if(is.null(c_PDFS_List)){#wir haben nur trainingsdaten
            V=ABCanalysis::ABCanalysis(Prop_cur)
            BClimit=V$BCLimit
          }else{#es gibt predicton und training
            LL_train=c_PDFS_List[[f]]
            x_train=apply(LL_train,1,function(x) prod(pmin(pmax(x[is.finite(x)],threshold),1-threshold)))
            insert=c(Prop_cur,x_train)#use test and train for abc analysis                                        
            V=ABCanalysis::ABCanalysis(insert)
            BClimit=V$BCLimit
          }
          
          #gibts erst ab 1.2.2
          # ABCcls=ABCanalysis::ABCclassification(exp(ClassDecisionLogProb[,i]))
          # UnReasonableInd=which(ABCcls>2)#klasse drei is plausible oder reasonable
          BC_limit_vec[f]=BClimit
          UnReasonableInd <- which(Prop_cur < BClimit)
        }else{#stupid fallback
          warning("Please install ABCanalysis")
          UnReasonableInd <- which(Prop_cur < 0.1)
          
        }#end check if package installed
        #ab hier ist noch ein bug drin, da falsch korrigiert wird
        #cases that are unreasonable are made plausible
        #alternative would be to return class assignment of zero (unknown)
        #if any liklihood has and unreasonable row for a specific case
        #this would be too strong
        if(length(UnReasonableInd)>0){
          #funktioniert nur, da PlausibleCenters in reihenfolge der klassen ist!
          Cls=fixed_kmeans(Feature = Data[,f],Centers = PlausibleCenters[f,]) #klasse k=1:length(PlausibleCenters)
          #likelihoods sind in der reihenfolge der datenpunkte
          #funktioniert nur, da likelihood spalten in reihenfolge der klassen sind!
          LikelihoodCls=max.col(LL,ties.method = "first")#apply(LL, 1, which.max)#klasse k=1:nrow(LikelihoodCls)
          #identify ALL rows that are not plausible at the moment EVEN if above threshold
          diff_ind_x_all=which(LikelihoodCls!=Cls)
          #clear cols that should not make changes
          #check_ind has entries defining col1->col2 and reverese i.e., col2->col1 type of change
          if(nrow(check_ind)>0&length(diff_ind_x_all)>0){
            diff_ind_x_all_pre=diff_ind_x_all
            mat=cbind(LikelihoodCls,Cls)[diff_ind_x_all,,drop=FALSE]
            #defines proposed plausible changes of likelihoods
            #in wich the difference in centers is to small
            #i.e., centers were not approximated correctly
            for(h in 1:nrow(check_ind)){
              boolvec=apply(mat,1,function(x) sum(x==check_ind[h,])==2)|apply(mat,1,function(x) sum(x==rev(check_ind[h,]))==2)
              #View(mat[!boolvec,])
              #only plausible changes remain that not fillfill above condition
              diff_ind_x_all_cur=diff_ind_x_all_pre[!boolvec]
              diff_ind_x_all=intersect(diff_ind_x_all,diff_ind_x_all_cur)
            }
            
          }
          #select only rows that are below threshold
          diff_ind_x=intersect(UnReasonableInd,diff_ind_x_all)
          # #epsilon change makes sure that the correct ones are now really maximal, i.e.,
          # #the likelihood that the class with the closest distance is assigned to x is maximal and infuences in the right direction
          # #the Posteriors
          eps    <- min(c(LL[LL>threshold],threshold^(1/4)))#set epsilon
          if(length(diff_ind_x)>0){#switch them to plaubile ones
            for(p in 1:length(diff_ind_x)){
              #which the inplausubile lilihood has a current maximum
              colcur_max=LikelihoodCls[diff_ind_x[p]]
              #the current maximum value of the implausible class 
              Max_cur=Likelihoods[diff_ind_x[p],colcur_max,f,drop=F]
              #what is the class likelihood that should have be maximal
              col_should_max=Cls[diff_ind_x[p]]
              #lower the maximum value of the inplausible class with eps
              Likelihoods[diff_ind_x[p],colcur_max,f]=max(Max_cur-eps,0)
              #replace with the value that is currently maximal and add eps
              Likelihoods[diff_ind_x[p],col_should_max,f]=Max_cur+eps
              #do not change likelihoods for other classes
            }
          }
        }#end length(UnReasonableInd)>0)
      }#end for( f in 1:d)
  }else{# Likelihoods is a list - (old legacy code)
    warning("Plausible Likelihoods: legacy code is used!")
    ListOfLikelihoods=Likelihoods
    ListOfLikelihoods_plausible=ListOfLikelihoods
    d=length(ListOfLikelihoods)#fuer jedes features gibt es ein listenelement
    N=nrow(ListOfLikelihoods[[1]])#durch interpolateDistributionOnData ist jedes listenelement der gleichen laenge gleich laenge der daten
    #If the “plausible Bayes” calculation results
    #in an undefined value, the class with the closest distance is assigned to x
    #: class(x) = argmin({d(x,m1), …, d(x,m𝑐)}).
    class_l=ncol(ListOfLikelihoods[[1]])
    BC_limit_vec=rep(NaN,d)
    for( f in 1:d){
      #skip variable if any centers lie very near to each other
      #nearer than 5%
      diffs=abs(max(PlausibleCenters[f,])-min(PlausibleCenters[f,]))
      check=min(abs(diffs))<diff(quantile(Data[,f],c(0.5,0.6)),lag=1)
      if(isTRUE(check)){
        BC_limit_vec[f]=0
        next;
      }
      LL=ListOfLikelihoods[[f]]
      #funktioniert nur, da PlausibleCenters in reihenfolge der klassen ist!
      Cls=fixed_kmeans(Data[,f],Centers = PlausibleCenters[f,]) #klasse k=1:length(PlausibleCenters)
      #likelihoods sind in der reihenfolge der datenpunkte
      #funktioniert nur, da likelihood spalten in reihenfolge der klassen sind!
      LikelihoodCls=max.col(LL,ties.method = "first")#apply(LL, 1, which.max)#klasse k=1:nrow(LikelihoodCls)
      #select wich likeliehoods are not plausbile
      #for debugging
      # x=Data[,f]
      # ind=order(x,decreasing = F)
      # plot(x[ind],LL[ind,1],type="l",col=DataVisualizations::DefaultColorSequence[1],main="Bayesian",lwd=2)
      # for(i in 2:ncol(LL)){
      #   points(x[ind],LL[ind,i],type="l",col=DataVisualizations::DefaultColorSequence[i],lwd=2)
      # }
      # DataVisualizations::DefaultColorSequence[1:3]
      # points(x,rep(0.2,length(x)),col=DataVisualizations::DefaultColorSequence[Cls])
      #
      #kein plan wieso das falsch ist
      # diff_ind_x <- which(LikelihoodCls != Cls)
      # if(length(diff_ind_x) > 0){
      #   col_from <- LikelihoodCls[diff_ind_x]
      #   col_to   <- Cls[diff_ind_x]
      #   eps    <- threshold^(1/4)
      #   Max_cur <- LL[diff_ind_x,col_from]
      #   LL[diff_ind_x,col_to] <- Max_cur + eps
      #   LL[diff_ind_x,col_from] <- Max_cur - eps
      # }
      
      diff_ind_x=which(LikelihoodCls!=Cls)#identify the rows that are not plausible at the moment
      # #epsilon change makes sure that the correct ones are now really maximal, i.e.,
      # #the likelihood that the class with the closest distance is assigned to x is maximal and infuences in the right direction
      # #the Posteriors
      eps    <- threshold^(1/4)#set epsilon
      if(length(diff_ind_x)>0){#switch them to plaubile ones
        for(p in 1:length(diff_ind_x)){
          #which the inplausubile lilihood has a current maximum
          colcur_max=LikelihoodCls[diff_ind_x[p]]
          #the current maximum value of the implausible class 
          Max_cur=LL[diff_ind_x[p],colcur_max]
          #what is the class likelihood that should have be maximal
          col_should_max=Cls[diff_ind_x[p]]
          #lower the maximum value of the inplausible class with eps
          LL[diff_ind_x[p],colcur_max]=max(Max_cur-eps,0)
          #replace with the value that is currently maximal and add eps
          LL[diff_ind_x[p],col_should_max]=Max_cur+eps
          #do not change likelihoods for other classes
        }
      }
      
      # #store them
      ListOfLikelihoods_plausible[[f]]=LL
      #fore debugging
      # plot(x[ind],LL[ind,1],type="l",col=DataVisualizations::DefaultColorSequence[1],main="Bayesian",lwd=2)
      # for(i in 2:ncol(LL)){
      #   points(x[ind],LL[ind,i],type="l",col=DataVisualizations::DefaultColorSequence[i],lwd=2)
      # }
      # DataVisualizations::DefaultColorSequence[1:3]
      # points(x,rep(0.2,length(x)),col=DataVisualizations::DefaultColorSequence[Cls])
    }
    
    #The threshold 𝜀 below which the evidence is considered as low is estimated 
    #for each pdf of the posteriors using a computed ABC analysis. Subset “C” of
    #this item categorization contains the nonprofitable values, i.e., the lowest 
    #probabilities (“the trivial many”). 
    for( f in 1:d){
      LL=ListOfLikelihoods[[f]]
      #approximation posterior, priors are irrelevant here
      Prop_cur=apply(LL,1,function(x) prod(pmin(pmax(x[is.finite(x)],threshold),1-threshold)))
      
      
      if(isTRUE(requireNamespace("ABCanalysis",quietly=TRUE))){
        ## identify critical points below threshold ----
        #The threshold 𝜀 for a class assignment probability that is considered 
        #too low for the assignment decision is defined as the BC limit calculated 
        #in the ABC analysis. In this way, the “trivial many” probabilities are 
        #removed from the Bayesian class assignment decision
        
        #dieser teil haengt davon ab, wie lang der vector und mit welchen werten ist
        #das unterscheidet sich in trainings und test daten
        #alu's paper definiert nicht, ueber welche daten die abc analyse getaetigt werden soll
        if(is.null(c_PDFS_List)){#wir haben nur trainingsdaten
          V=ABCanalysis::ABCanalysis(Prop_cur)
          BClimit=V$BCLimit
        }else{#es gibt predicton und training
          LL_train=c_PDFS_List[[f]]
          x_train=apply(LL_train,1,function(x) prod(pmin(pmax(x[is.finite(x)],threshold),1-threshold)))
          insert=c(Prop_cur,x_train)#use test and train for abc analysis                                        
          # if(length(Prop_cur)<3){
          #   x_train=exp(LogPropMatTrain[,i])
          #   insert=c(x_train,Prop_cur)
          # }else{
          #   insert=Prop_cur
          # }
          V=ABCanalysis::ABCanalysis(insert)
          BClimit=V$BCLimit
        }
        #gibts erst ab 1.2.2
        # ABCcls=ABCanalysis::ABCclassification(exp(ClassDecisionLogProb[,i]))
        # UnReasonableInd=which(ABCcls>2)#klasse drei is plausible oder reasonable
        if(is.nan(BC_limit_vec[f])){
          BC_limit_vec[f]=BClimit
          UnReasonableInd <- which(Prop_cur < BClimit)
        }else{
          #centers sind zu nahe beieiander
          UnReasonableInd=NULL
        }

      }else{#stupid fallback
        warning("Please install ABCanalysis")
        UnReasonableInd <- which(Prop_cur < 0.1)
        
      }#end check if package installed
      #cases that are unreasonable are made plausible
      #alternative would be to return class assignment of zero (unknown)
      #if any liklihood has and unreasonable row for a specific case
      #this would be too strong
      if(length(UnReasonableInd)>0){
        #Cases that have a probability of belonging to a particular class below 
        #the threshold 𝜀 ... have to be in plausible categorization
        #reset likelihoods on these UnReasonableInd for prediction
        LL=ListOfLikelihoods[[f]]
        LL2=ListOfLikelihoods_plausible[[f]]
        LL[UnReasonableInd,]=LL2[UnReasonableInd,]
        ListOfLikelihoods[[f]]=LL
        # print(f)
        # print(LL[UnReasonableInd,])
        # print(LL2[UnReasonableInd,])
      }#end length(UnReasonableInd)>0)
    }#end for( f in 1:d)
    Likelihoods=listOfLikelihoods2Array(ListOfLikelihoods)
    
  }#end if ! is list Likelihoods 
    return(list(Likelihoods=Likelihoods,PlausibleCenters=PlausibleCenters,Epsilon=BC_limit_vec))
}

fixed_kmeans <- function(Feature, Centers) {
  
  ## distance matrix: rows = observations, cols = centres
  dists <- abs(outer(Feature, Centers, FUN = "-"))
  
  ## nearest centre index for each observation
  cluster <- max.col(-dists, ties.method = "first")
  
  #d dimensional case
  # # Data: matrix or data.frame of points (rows = samples, columns = features)
  # # Centers: matrix or data.frame of fixed cluster centers (rows = centers, columns = features)
  # 
  # if (!is.matrix(Data)) Data <- as.matrix(Data)
  # if (!is.matrix(Centers)) Centers <- as.matrix(Centers)
  # 
  # # Sanity checks
  # if (ncol(Data) != ncol(Centers)) stop("Data and centers must have the same number of columns (features).")
  # 
  # # Compute distances from each point to each center
  # dists <- as.matrix(dist(rbind(Data, Centers)))
  # dists <- dists[1:nrow(Data), (nrow(Data)+1):(nrow(Data)+nrow(Centers))]
  # 
  # # Assign each point to the closest center
  # Cls <- apply(dists, 1, which.min)
  
  return(cluster)
}

interpolatePDF4Data=function(Data,PDFfunList,cl){
  if(!is.matrix(Data)){
    Data=as.matrix(Data)
  }
  cl_length=length(PDFfunList[[1]])
  N=nrow(Data)
  d=ncol(Data)
  Array=array(NaN,c(N,cl_length,d))

  # if(missing(cl)){
   for(i in 1:d){
    for(cc in 1:cl_length){
        x=Data[,i]
        bb=!is.finite(x)
        if(sum(bb)==0){
          cpdf=PDFfunList[[i]][[cc]](x)
        }else{
          cpdf=rep(0,length(x))
          cpdf[!bb]=PDFfunList[[i]][[cc]](x[!bb])
        }
        cpdf[cpdf<0]=0 #auch kleine schwankungen unter null nicht erlaubt
        #plot(x,cpdf)
        Array[1:length(cpdf),cc,i]=cpdf
      }
    }
  # }else{
  # 
  #   Matrix=array(NaN,c(N,cl_length))
  #   #das lohnt sich nur bei sehr groessen N und D
  #   parallel::clusterExport(cl, c("Data","PDFfunList","cl_length","Matrix"))
  #   
  #   List=parallel::parLapply(cl,1:d,function(i,Matrix){
  #     for(cc in 1:cl_length){
  #       x=Data[,i]
  #       bb=!is.finite(x)
  #       if(sum(bb)==0){
  #         cpdf=PDFfunList[[i]][[cc]](x)
  #       }else{
  #         cpdf=rep(0,length(x))
  #         cpdf[!bb]=PDFfunList[[i]][[cc]](x[!bb])
  #       }
  #       cpdf[cpdf<0]=0 #auch kleine schwankungen unter null nicht erlaubt
  #       #plot(x,cpdf)
  #       Matrix[1:length(cpdf),cc]=cpdf
  #     }
  #     return(Matrix)
  #   },Matrix)
  #   
  #   for (i in seq_along(List)) {
  #     Array[ , , i] <- List[[i]]
  #   }
  # }

  return(Array)
}

interpolatePDF4Kernels=function(KernelList,PDFfunList){
  
  matList <- Map(
    function(funs_i, kern_i) {
      # split into a list of column‐vector, each column repesents a class likelihood
      xs <- asplit(kern_i, 2)
      # apply each f to its x, mapply will give you a N×class_length matrix
      mapply(function(f, x) f(x),
             funs_i, xs,
             SIMPLIFY = TRUE, USE.NAMES = FALSE)
    },
    PDFfunList,
    KernelList
  )
  arr=simplify2array(matList)
  arr[arr<0]=0 #auch kleine schwankungen unter null nicht erlaubt
  # 2) stack those matrices along the 3rd dimension which is 1:d
  return(arr)
  
}

likelihoodsArray2List=function(Likelihoods,ClassLength=NULL){

  #to do try: asplit
  if(is.array(Likelihoods)){
    V=dim(Likelihoods)
    if(is.null(ClassLength)){
      #we have an 3d array
      N=V[1]
      class_l=V[2]
      d=V[3]
    }else{
      #we have an 2 d array, i.e. the training data to define the kernels
      N=V[1]
      class_l=ClassLength
      #nochmal spaeter dreuber nachdenken
	  #if(V[2]%%class_l!=0)
      #  stop("likelihoodsArray2List: number of columns is not divisible by ClassLength.")
      #d=V[2]/class_l
      Likelihoods=array(Likelihoods,dim = c(N,class_l,d))
    }

    ListOfLikelihoods=list()
    for(i in 1:d){
      cur=Likelihoods[,,i]
      if(N>1)
        ListOfLikelihoods[[i]]=cur
      else
        ListOfLikelihoods[[i]]=matrix(cur,nrow=1)
    }
    return(ListOfLikelihoods)
  }else{
    return(Likelihoods)
  }

}

Train_naiveBayes=function(Data,Cls,Predict=TRUE,Priors,...){
  # V=Train_naiveBayes(Data,Cls,Gaussian=TRUE)
  # 
  # INPUT
  # Data[1:n,1:d]    Numeric matrix for data with n observations and d features.
  # Cls[1:n]         Numeric vector with k class labels
  # 
  # OPTIONAL
  # Predict           TRUE: yields ClsTrain and Posteriors, FALSE: yields only Model and Thetas
  #                   Note: Only if Predict TRUE, EvalPlausible can be true
  # Priors            numerical vector [1:k] defining the prior probabilities of k classes, if missing, estimated from Cls
  #                   
  # ... 
  # 
  # OUTPUT
  # Model                   Model that stores naive bayes if Gaussian=F according to the bayesian Theorem.     
  #                         is a list of
                            # Priors[1:l]             Numeric vector with prior probability for each class.
                            # Thetas
                            # PDFs_funs
                            # c_Kernels_list
                            # PlausibleCenters
  # ClsTrain[1:n]           if Predict=T: Numeric vector with training cls, otherwise NULL
  # Posteriors[1:n, 1:l]    if Predict=T: Numeric matrices with posterior probabilities, otherwise NULL
  
  # dots = list("Plausible" = TRUE, "Gaussian" = FALSE, "Type" = 1, "Threshold" = 1e-12, "PlotIt" = FALSE)
  ## memshare addition double to boolean
  if(Predict==1) Predict=TRUE
  if(Predict==0) Predict=FALSE
  
  #DAU catching ----
  if(!is.matrix(Data)){
    warning("Train_naiveBayes: Data is not matrix, calling as.matrix()")
    #otherwise in defineOrEstimateDistribution Error in xtfrm.data.frame(x) : cannot xtfrm data frames when sort is called on a columnd of df
    Data=as.matrix(Data)
  }
  
  if(!is.numeric(Cls)){
    warning("Train_naiveBayes: Cls is not numneric, trying to transform.")
    if(isTRUE(requireNamespace("FCPS",quietly = T))){
      V=FCPS::ClusterCreateClassification(Cls)
      Cls=V$Cls
      message(paste0(V$ClusterNames,collapse = " "))
      message("transformed to")
      message(paste0(names(V$ClusterNames),collapse = " "))
    }else{
      warning("Train_naiveBayes: Please install FCPS package for stable Cls transformation.")
      Cls=as.numeric(Cls)
    }
  }
  
  N=nrow(Data)
  nc=length(Cls)
  if(N!=nc){
    warning("Train_naiveBayes: length of Cls unequal nrow of data, please check input for correct learning. Shortening Input.")
    Data=Data[1:min(c(N,nc)),,drop=FALSE]
    Cls=Cls[1:min(c(N,nc))]
  }
  
  bool_fin=is.finite(Cls)
  if(sum(bool_fin)!=length(Cls)){
    warning("Train_naiveBayes: Not all elements in Cls are finite, deleting not finite ones.")
    Cls=Cls[bool_fin]
    Data=Data[bool_fin,,drop=FALSE]
  }


  #Argument catching ----
  dots=list(...)
  
  has_plausible <- "Plausible" %in% names(dots)
  has_Gaussian <- "Gaussian" %in% names(dots)
  has_Threshold <- "Threshold" %in% names(dots)
  has_PlotIt <- "PlotIt" %in% names(dots)
  has_PlotCutOff<- "PlotCutOff" %in% names(dots)
  has_globalPR <- "GlobalPR" %in% names(dots)
  
  if(has_globalPR){
    GlobalPR = dots$GlobalPR
    if(GlobalPR==1) GlobalPR=TRUE
    if(GlobalPR==0) GlobalPR=FALSE
  }else{
    GlobalPR = T
  }
  
  if (has_plausible) {
    Plausible= dots$Plausible #kann true oder false sein
    #for memshare
    if(Plausible==1) Plausible=TRUE
    if(Plausible==0) Plausible=FALSE
    
    EvalPlausible=FALSE
  }else{
    #wenn nicht eingestellt wird beides evaluiert
    Plausible=FALSE
    EvalPlausible=TRUE
  }
  if(isFALSE(Predict))
    EvalPlausible=FALSE
  if (has_Gaussian) {
    Gaussian= dots$Gaussian
    #for memshare transform pars
    if(Gaussian==0) Gaussian=FALSE
    if(Gaussian==1) Gaussian=TRUE
  }else{
    Gaussian=FALSE
  }
  if (has_Threshold) {
    Threshold= dots$Threshold 
  }else{
    Threshold=1e-4
  }
  if (has_PlotIt) {
    PlotIt= dots$PlotIt 
  }else{
    PlotIt=FALSE
  }
  if (has_PlotCutOff) {
    PlotCutOff= dots$PlotCutOff
  }else{
    PlotCutOff=min(c(ncol(Data),4))
  }


  
  #Getting Priors and Defining Classes internally ----
  unique_classes=sort(unique(Cls),decreasing = F)#class ordering externally can be arbitrary
  if(missing(Priors)){
    Priors=getPriors(Cls)
  }else{
   
    if(length(Priors)>length(unique_classes)){
      warning("Train_naiveBayes: Input parameter Priors is longer than k classes.")
      Priors=Priors[1:length(unique_classes)]
    }
    if(length(Priors)<length(unique_classes)){
      warning("Train_naiveBayes: Input parameter Priors is smaller than k classes, fallback to getPriors().")
      
      Priors=getPriors(Cls)
    }
    if(is.null(names(Priors))){
      names(Priors)=unique_classes
      warning("Train_naiveBayes: Input parameter Priors is not named vector, assuming priors are in order 1:k.")
    }else{
      prior_labels=suppressWarnings(as.numeric(names(Priors)))
      ii=match(unique_classes,prior_labels)
      if(anyNA(ii)){
        warning("Train_naiveBayes: Given Priors have to be named numeric vector in order of sort(unique(Cls))")
      }else{
        Priors=Priors[ii]
      }
    }
  }
  

  
  prior_labels=suppressWarnings(as.numeric(names(Priors)))
  if(anyNA(prior_labels) || length(prior_labels)!=length(unique_classes) || any(unique_classes!=prior_labels)){
    warning("Train_naiveBayes: Given Priors have to be named numeric vector in order of sort(unique(Cls))")
  }

  if(isTRUE(Plausible)&any(1:length(unique_classes)!=unique_classes)){
      if(isTRUE(requireNamespace("FCPS",quietly = T))){
        #make sure that internally Cls is ordered 1:k
        if(utils::packageVersion("FCPS")>"1.3.4"){
          Cls=FCPS::ClusterRedefine(Cls,NewLabels = 1:length(unique_classes),OldLabels =unique_classes,Silent = T)
        }else{
          warning("Train_naiveBayes: Please update FCPS for better stability of algorithm.")
        }
        #otherwise unsure if in PlausibleLikelihoods() FCPS::ClusterApply has to correct ordering
      }else{
        warning("Train_naiveBayes: Please install FCPS package for better stability of algorithm.")
      }
  }

  
  #Start Learning Process----
  c_2List=GetLikelihoods(Data = Data,Cls = Cls,...)
  Model=c_2List
  Model$Priors=Priors
  c_Kernels_list=c_2List$c_Kernels_list
  Thetas=c_2List$Thetas
  PDFs_funs=c_2List$PDFs_funs
  ParetoRadiusPerFeauture=c_2List$ParetoRadiusPerFeauture
  if(isFALSE(Gaussian)){
    #this makes sure that now the ListOfLikelihoods has the length of data
    ArrayOfLikelihoodsOnData=interpolatePDF4Data(Data,PDFs_funs)
    #return(ArrayOfLikelihoods)
    if(isTRUE(Plausible)|isTRUE(EvalPlausible)){
      PDFs_funs_PrePlausible=PDFs_funs
      V=KernelPlausibleLikelihoods(PDFs_funs = PDFs_funs,c_Kernels_list = c_Kernels_list,Data = Data,Cls = Cls,threshold=Threshold)
      PDFs_funs=V$PDFs_funs
      #PlausibleLikelihoods=V$PlausibleLikelihoods
      #Likelihoods_PrePlausible=V$Likelihoods_PrePlausible
      Epsilon=V$Epsilon
      PlausibleCenters=V$PlausibleCenters
      ArrayOfLikelihoodsOnData=interpolatePDF4Data(Data,PDFs_funs)
      Model$PDFs_funs=PDFs_funs
      #PlotLikelihoods(Likelihoods = Likelihoods_PrePlausible,Data,PlausibleLikelihoods = ArrayOfLikelihoodsOnData)
    }else{
      Epsilon=NULL
      PlausibleCenters=NULL
    }
    Model$PlausibleCenters=PlausibleCenters
  }else{
    #override
    EvalPlausible=FALSE
    Plausible=FALSE
    ArrayOfLikelihoodsOnData=listOfLikelihoods2Array(c_2List$ListOfLikelihoods)
  }

  if(isTRUE(PlotIt)){
    if(isTRUE(Gaussian)){
      PlotLikelihoods(ArrayOfLikelihoodsOnData,Data = Data,PlotCutOff=PlotCutOff)
    }else if(is.null(Epsilon)){
      PlotLikelihoodFuns(PDFs_funs,Data = Data,PlotCutOff=PlotCutOff)
    }else{
      PlotLikelihoodFuns(PDFs_funs_PrePlausible,Data = Data,PlausibleLikelihoodFuns = PDFs_funs,Epsilon=Epsilon,PlausibleCenters=PlausibleCenters,PlotCutOff=PlotCutOff)
    }
  }
  class(Model)="PDEbayes"
  if(isTRUE(Predict)){
  PosteriorsV1=ApplyBayesTheorem4Likelihoods(Likelihoods = ArrayOfLikelihoodsOnData,
                                             Priors = Priors, threshold = Threshold)
  Posteriors=PosteriorsV1$Posteriors
  cls_train_col=PosteriorsV1$Cls
  
  if(isTRUE(EvalPlausible)){
  #Decide if plausible is better ----
    Likelihoods_PrePlausibleOnData=interpolatePDF4Data(Data,PDFs_funs_PrePlausible)
    PosteriorsV_2=ApplyBayesTheorem4Likelihoods(Likelihoods = Likelihoods_PrePlausibleOnData,
                                              Priors = Priors, threshold = Threshold)
    Posteriors2=PosteriorsV_2$Posteriors
    cls_train_col2=PosteriorsV_2$Cls
    
    #decide based on shannon information
    #measures how diverse and balanced clusters are on average in each solution
    #skewed clusters get a lower entropy (i.e. all points in one cluster)
    #even if one cluster is small and several bigger, the average should be still
    # a good measure
    shan=FCPS::ClusterShannonInfo(cbind(cls_train_col,cls_train_col2))$MeanInfo
    if(shan[1]<shan[2]){
      cls_train_col=cls_train_col2
      Posteriors=Posteriors2
      Plausible=FALSE
      Model$PDFs_funs=PDFs_funs_PrePlausible
    }else{
      Plausible=TRUE
      Model$PDFs_funs=PDFs_funs
    }
  }
  Model$Plausible=Plausible
  
  if(isTRUE(requireNamespace("FCPS",quietly = T))){
    if(utils::packageVersion("FCPS")>"1.3.4"){
      cls_train=FCPS::ClusterRedefine(cls_train_col,NewLabels = unique_classes,OldLabels =1:ncol(Posteriors),Silent = T)
    }else{
      warning("please update FCPS")
      cls_train=cls_train_col
    }
  }else{
    warning("please install FCPS")
    cls_train=cls_train_col
  }

  if(!is.null(rownames(Data))){
    names(cls_train)=rownames(Data)
  }

  return(list(
    Model=Model,
    ClsTrain=cls_train,
    Posteriors=Posteriors
    ))
  }else{
    return(list(
      Model=Model,
      ClsTrain=NULL,
      Posteriors=NULL
    ))
  }
}

GetLikelihoods=function(Data,Cls,...){
  # V=GetLikelihoods(Data,Cls,Gaussian=TRUE)
  # 
  # INPUT
  # Data[1:n,1:d]    Numeric matrix for data with n observations and d features.
  # Cls              Integer Vector with class lables
  # 
  # OPTIONAL
  # Gaussian     (Optional: Default=TRUE). Assume gaussian distribution.         
  # 
  # OUTPUT
  # c_Kernels_list       List of d numeric matrices, one per feature, each matrix with 1:k columns containing 
  #						 the kernels of class 1:k
  # ListOfLikelihoods    List of d numeric matrices, one per feature, each matrix with 1:k columns containing 
  #                      the distribution of class 1:k per feature, i.e., the Likelihood per class 
  # Thetas: If Gaussian=TRUE:  List of d numeric matrices, one per feauture, each matrix with 1:k rows containing 
  #                      the mean in the first column and the standard deviation in teh seconf columd of class 1:k

  dots=list(...)
  has_plausible <- "Plausible" %in% names(dots)
  has_Gaussian <- "Gaussian" %in% names(dots)
  has_globalPR <- "GlobalPR" %in% names(dots)
  
  if(has_globalPR){
    GlobalPR = dots$GlobalPR
  }else{
    GlobalPR = T
  }
  
  if (has_Gaussian) {
    Gaussian= dots$Gaussian
  }else{
    Gaussian=F
  }

  unique_classes=sort(unique(Cls),decreasing = F)
  d=ncol(Data)
  
  #bottleneck des classifiers
  if(isFALSE(Gaussian)){
    has_ParetoRadiusVector<- "ParetoRadiusPerFeauture" %in% names(dots)
    if (has_ParetoRadiusVector) {
      ParVec= dots$ParetoRadiusPerFeauture 
    } else{
      has_cl<- "cl" %in% names(dots)
      
      if(isTRUE(GlobalPR)){
        if(has_cl & isTRUE(requireNamespace("parallel",quietly = T))){
          
          ParVec=parallel::parApply(cl = dots$cl,X = Data,MARGIN = 2, FUN = function(Feature){
            par <- if (packageVersion("DataVisualizations") >= "1.4.0") DataVisualizations::ParetoRadius_fast(Feature) else DataVisualizations::ParetoRadius(Feature)
            return(par)
          })
          
        }else{
          ParVec=apply(Data,MARGIN = 2, function(Feature){
            par <- if (packageVersion("DataVisualizations") >= "1.4.0") DataVisualizations::ParetoRadius_fast(Feature) else DataVisualizations::ParetoRadius(Feature)
            return(par)
          })
        }
      }else{
        if(has_cl & isTRUE(requireNamespace("parallel",quietly = T))){
          
          ParVec = rbind()
          
          for(i in 1:length(unique_classes)){
            Idx         = which(Cls == unique_classes[i])
            ClassParVec = parallel::parApply(cl = dots$cl,X = Data[Idx, ,drop=FALSE], MARGIN = 2, FUN = function(Feature){
              par <- if (packageVersion("DataVisualizations") >= "1.4.0") DataVisualizations::ParetoRadius_fast(Feature) else DataVisualizations::ParetoRadius(Feature)
              return(par)
            })
            ParVec = rbind(ParVec, ClassParVec)
          }
          
        }else{
          
          ParVec = rbind()
          
          for(i in 1:length(unique_classes)){
            Idx         = which(Cls == unique_classes[i])
            ClassParVec = apply(Data[Idx, ,drop=FALSE], MARGIN = 2, function(Feature){
              par <- if (packageVersion("DataVisualizations") >= "1.4.0") DataVisualizations::ParetoRadius_fast(Feature) else DataVisualizations::ParetoRadius(Feature)
              return(par)
            })
            ParVec = rbind(ParVec, ClassParVec)
          }
          row.names(ParVec) = unique_classes
          #colnames(ParVec) = paste0("F", 1:dim(Data)[2])
        }
      }
    }
  }else{
    #will be ignored for gaussian bayes
    ParVec=rep(0,d)
  }

  #3D array anstatt liste geht nicht, da anzahl an rows sich hier noch aendert. erst in interpolate DistributionOnData koennte man ein 3D array wiedergeben
  # alternatrive ware max no rows abschaetzen und hinten mit NaN auffuellen s.DataVisualizations::CombineRows(), aber das ist vermutlich weniger effizient
  PDFs=list()
  PDFs_funs=list()
  Kernels=list()
  Thetas=list()
  for(i in 1:d){
    c_pdf=list()
    c_kernel=list()
    c_theta=list()
    funs=list()
    for(cc in 1:length(unique_classes)){    #alles ist in reihenfolge der klassen anzuordnen!
      Classind=which(Cls==unique_classes[cc])
      if(length(Classind)>0){
        # Global ParetoRadius = ParVec[i]
        # Class-dependend ParetoRadius = ParVec[cc, i]
        
        if(is.vector(ParVec)){
          c_pdf_list = defineOrEstimateDistribution(Feature = Data[, i, drop = F],
                                                    ClassInd = Classind,
                                                    ParetoRadius = ParVec[i],
                                                    ...)
        }else{
          c_pdf_list = defineOrEstimateDistribution(Feature = Data[, i, drop = F],
                                                    ClassInd = Classind,
                                                    ParetoRadius = ParVec[cc, i],
                                                    ...)
        }
        
        c_pdf[[cc]]=c_pdf_list$PDF
        c_kernel[[cc]]=c_pdf_list$Kernels
        funs[[cc]]=c_pdf_list$PDF_fun
        if(is.vector(ParVec)){
          if(cc==1)
            ParVec[i]=c_pdf_list$ParetoRadius
        }else{
          ParVec[cc,i]=c_pdf_list$ParetoRadius
        }
        if(isTRUE(Gaussian))
          c_theta[[cc]]=c_pdf_list$Theta
      }else{
        c_pdf[[cc]]=c(0,0)
        c_kernel=c(min(Data[,i],na.rm = T),max(Data[,i],na.rm = T))
        
        if(isTRUE(Gaussian))
          c_theta[[cc]]=c(mean(Data[,i],na.rm = T),sd(Data[,i],na.rm = T))
      }
    }#end for each class
    PDFs_funs[[i]]=funs
    if(isTRUE(Gaussian)){
      ThetaPerClass=do.call(rbind,c_theta)
      Thetas[[i]]=ThetaPerClass
    }else{
      ThetaPerClass=NULL
      Thetas=NULL
    }
    
    nr=max(sapply(c_pdf, length))
    mat_pdf=matrix(0,nrow = nr,ncol = length(unique_classes))
    mat_kernels=matrix(NaN,nrow = nr,ncol = length(unique_classes))

    colnames(mat_pdf)=paste0("C",unique_classes)
    colnames(mat_kernels)=paste0("C",unique_classes)

    
    for(cc in 1:length(unique_classes)){
      cur_pdf=c_pdf[[cc]]
      cur_kernel=c_kernel[[cc]]
      
      mat_pdf[1:length(cur_pdf),cc]=cur_pdf
      mat_kernels[1:length(cur_kernel),cc]=cur_kernel

    }#end for each class
    
    PDFs[[i]]=mat_pdf
    Kernels[[i]]=mat_kernels
  }#end for each feature
  
   names(PDFs_funs)=colnames(Data)
   if(is.vector(ParVec))
     names(ParVec)=colnames(Data)
   else
     colnames(ParVec)=colnames(Data)
   if(!is.null(Thetas))
     names(Thetas)=colnames(Data)
   names(PDFs)=colnames(Data)
   names(Kernels)=colnames(Data)
  return(list(c_Kernels_list=Kernels,ListOfLikelihoods=PDFs,PDFs_funs=PDFs_funs,ParetoRadiusPerFeauture=ParVec,Thetas=Thetas))
}

fitParameters=function(Feature,ClassInd,Robust=FALSE,na.rm=TRUE,SD_Threshold=0.0001){
  # V=fitParameters(Feature,ClassInd,Robust=FALSE,na.rm=TRUE,SD_Threshold=0.0001)
  # 
  # INPUT
  # Feature[1:n]    Numeric Vector
  # ClassInd        Integer Vector with class indices
  # 
  # OPTIONAL
  # Robust       (Optional: Default=FALSE). Robust computation if set to TRUE.
  # na.rm        (Optional: Default=TRUE). Remove na.
  # SD_Threshold    (Optional: Default=0.00001). Threshold value for std.
  # 
  # OUTPUT
  # Parameters[1:2]    Numeric vector with Mean and Std.
  # 
  Feature_Class=Feature[ClassInd]
   if(!requireNamespace("DataVisualizations")){
   Robust=FALSE
   warning("fitParameters: requires internal package dbt.Statistic to work which is not installed, setting Robust=FALSE")
   }
  if(isTRUE(Robust)){
 
    me=DataVisualizations::Meanrobust(Feature_Class,na.rm = na.rm)
    std=DataVisualizations::Stdrobust(Feature_Class,na.rm = na.rm)

  }else{
    me=mean(Feature_Class,na.rm=na.rm)
    std=sd(Feature_Class,na.rm=na.rm)
  }

  if(!is.finite(std) || std<SD_Threshold)
    std=SD_Threshold

  return(c(Mean=me,Std=std))
}

Predict_naiveBayes=function(Data,Model,...){
  # V=Predict_naiveBayes(Data,Priors,c_2List_Train,Thetas)
  # 
  # INPUT
  # Data[1:n,1:d]         Numeric matrix for Data with n observations and d features.
  # Priors[1:l]               Numeric vector with prior probability for each class.
  # c_2List_Train
  # Thetas                    List of what?
  # 
  # OUTPUT
  # ClsTest[1:n]         Numeric vector with predicted class labels
  # Posteriors[1:n,1:l]  Numeric matrix with posterior probability according to
  #                      the bayesian theorem.
  # ListOfLikelihoods    List of m numeric matrices with l columns containing 
  #                      the distribution of class i in 1:l
  #
  if(!is.matrix(Data)){
    warning("Predict_naiveBayes: Data is not mattrix, calling as.matrix()")
   
    Data=as.matrix(Data)
  }
  
  #Getting Arguments ----
  dots=list(...)
  
  has_cl <- "cl" %in% names(dots)
  if (has_cl) {
    cl= dots$cl#multicore
  }else{#single core
    cl=NULL
  }
  
  
  has_plausible <- "Plausible" %in% names(dots)
  has_Threshold <- "Threshold" %in% names(dots)
  
  has_PlotIt <- "PlotIt" %in% names(dots)
  has_PlotCutOff<- "PlotCutOff" %in% names(dots)
  
  if (has_PlotIt) {
    PlotIt= dots$PlotIt 
  }else{
    PlotIt=FALSE
  }
  if (has_PlotCutOff) {
    PlotCutOff= dots$PlotCutOff 
  }else{
    PlotCutOff=min(c(4,ncol(Data)))
  }
  
  if (has_plausible) {

    Plausible= dots$Plausible 
  }else{
    Plausible=F
    PlausibleCenters=NULL
  }

  if (has_Threshold) {
    Threshold= dots$Threshold 
  }else{
    Threshold=.Machine$double.eps*1000
  }
  
  if(missing(Model)){
    Model=dots #somewhere here the input has to be stored
  }else{
    has_Model <- "Model" %in% names(Model)
    if (has_Model) {
      Model= Model$Model 
    }
  }
  
  has_Priors <- "Priors" %in% names(Model)
  has_Thetas <- "Thetas" %in% names(Model)
  has_PlausibleCenters <- "PlausibleCenters" %in% names(Model)
  if (has_PlausibleCenters) {
    
    PlausibleCenters= Model$PlausibleCenters 
  }else{
    PlausibleCenters=NULL
    if(isTRUE(Plausible)){
      warning("Predict_naiveBayes: Plausible=TRUE but list element PlausibleCenters is missing in c_2List_Train meaning that model was not train plausible, please correct. Setting Plausible to FALSE.")
      Plausible=FALSE
    }
  }
  if (has_Thetas) {
    Thetas= Model$Thetas 
  }else{
    Thetas=NULL
  }
  if (has_Priors) {
   Priors= Model$Priors 
  }else{
    stop("Priors are missing in input arguments.")
  }
  #legacy code for back compatibility
  has_c_2List_Train <- "c_2List_Train" %in% names(Model)
    if (has_c_2List_Train) {
  
        c_2List_Train= Model$c_2List_Train 
      
        has_Plausible2 <- "Plausible" %in% names(c_2List_Train)
        if (has_Plausible2) {
          #overide user setting
          Plausible= c_2List_Train$Plausible 
        }
        
        has_PlausibleCenters <- "PlausibleCenters" %in% names(c_2List_Train)
      if (has_PlausibleCenters) {
    
        PlausibleCenters= c_2List_Train$PlausibleCenters 
      }else{
        PlausibleCenters=NULL
        if(isTRUE(Plausible)){
          warning("Predict_naiveBayes: Plausible=TRUE but list element PlausibleCenters is missing in c_2List_Train meaning that model was not train plausible, please correct. Setting Plausible to FALSE.")
          Plausible=FALSE
        }
      }
    }#missing has_c_2List_Train is caught below


  if(!is.null(Thetas)){
      # Likelihood generation for simple Gaussian model if ThetaPerClass given----
      d=ncol(Data)
      Header=colnames(Data)
      d2=length(Thetas)
      if(d!=d2 || (!is.null(Header) && !is.null(names(Thetas)) && !identical(Header,names(Thetas)))){
        warning("Predict_naiveBayes: columns or column order of Data do not match Thetas. Trying to match by colname(Data)...")
        Data=Data[,match(x = names(Thetas),table = Header,nomatch = 0),drop=FALSE]
      }
      if(ncol(Data)!=d2)
        stop("Predict_naiveBayes: number of data columns does not equal length of Thetas. Please correct the input.")
      class_len=length(Priors)
      ListOfLikelihoods=list()
      for(m in 1:ncol(Data)){
        Feature=Data[,m]
        Likelihood=list()
        ThetaPerClass=Thetas[[m]]
        for(cc in 1:class_len){
          me = ThetaPerClass[cc,1]
          std = ThetaPerClass[cc,2]
          Likelihood[[cc]]=dnorm(Feature,me,std)
        }
        ListOfLikelihoods[[m]]=do.call(cbind,Likelihood)
      }
      PDFs=ListOfLikelihoods
      Epsilon=NULL
  }else{ #default case of ListOfLikelihoods given, then ThetaPerClass are ignored
    if(!has_c_2List_Train)
      c_2List_Train=Model
    c_Kernels_list=c_2List_Train$c_Kernels_list
    PlausibleCenters=c_2List_Train$PlausibleCenters
    PDFs_funs=c_2List_Train$PDFs_funs
    #Error Catching ----
    d=ncol(Data)
    Header=colnames(Data)
    d2=length(PDFs_funs)
    d3=length(c_Kernels_list)
    if(d!=d2 || (!is.null(Header) && !is.null(names(PDFs_funs)) && !identical(Header,names(PDFs_funs)))){
      warning("Predict_naiveBayes: columns or column order of Data do not match PDFs_funs. Trying to match by colname(Data)...")
      Data=Data[,match(x = names(PDFs_funs),table = Header,nomatch = 0),drop=FALSE]
    }
    d=ncol(Data)
    
    if(d!=d2){
      stop("Predict_naiveBayes: number of data columns still does not equal length of the list PDFs_funs. Please correct the input.")
    }
    if(d!=d3){
      stop("Predict_naiveBayes: columns of data does not equal length of c_Kernels_list. Please correct the input.")
    }
    # case_length1=unique(sapply(ListOfLikelihoods, nrow))
    # case_length2=unique(sapply(c_Kernels_list, nrow))
    # if(length(case_length2)!=1){
    #   warning("Predict_naiveBayes: rows of matrices in c_Kernels_list vary, please check input data")
    # }
    # if(any(case_length1!=case_length2)){
    #   warning("Predict_naiveBayes: rows of matrices in ListOfLikelihoods do not equal rows of matrices in c_Kernels_list, please check input data.")
    # }

    #class_length1=unique(sapply(ListOfLikelihoods, ncol))
     class_length1=unique(sapply(PDFs_funs, length))
     class_length2=unique(sapply(c_Kernels_list, ncol))
    if(length(class_length1)!=1){
      warning("Predict_naiveBayes: columns of matrices in ListOfLikelihoods vary, please check input data")
    }
    if(length(class_length2)!=1){
      warning("Predict_naiveBayes: columns of matrices in c_Kernels_list vary, please check input data")
    }
    class_length=length(Priors)
    if(class_length!=class_length1){
      warning("Predict_naiveBayes: length of of Priors does not equal length of ListOfLikelihoods. Please correct the input")
    }
    
    #Prediction ----
    #PDFs=interpolateDistributionOnData(c_Kernels_list = c_Kernels_list,c_PDFS_List = ListOfLikelihoods,Data = Data,na.rm=TRUE)
    #PDFs=PDFsV$ListOfLikelihoods
    #PDFs  =PDFsV$ArrayOfLikelihoods

    if(is.null(cl)){#for debugging purposes
      PDFs=interpolatePDF4Data(Data,PDFs_funs)
    }else{
      PDFs=interpolatePDF4Data(Data = Data,PDFs_funs,cl)
    }
    

    # if(isTRUE(Plausible)){
    #   #ich muss die korrektur nochmal tun, dass ich PlausibleLikelihoods korrektur der trainingsdaten
    #   #nicht in der ausgabe mitnehme, da ich die PDE dichte schaetzung im original fuer
    #   #eine korrekte interpolateDistributionOnData auf neuen daten benoetige
    #   #ausgabe von dem resultut von PlausibleLikelihoods mit korrekten kernels in Train:naiveBayes
    #   #fuert zur inkorrekten interpolation, warum auch immer
    #   Likelihoods4Plot=PDFs
    #   V2=PlausibleLikelihoods(Likelihoods=PDFs,Data=Data,Cls=NULL,threshold=Threshold,c_PDFS_List=NULL,PlausibleCenters=PlausibleCenters)
    #   Epsilon=V2$Epsilon
    #   PDFs=V2$Likelihoods
    # }else{
      Epsilon=NULL
    # }
  }

  PosteriorsV=ApplyBayesTheorem4Likelihoods(Likelihoods = PDFs,Priors = Priors,threshold = Threshold)
  Posteriors=PosteriorsV$Posteriors
  cls_test_col=PosteriorsV$Cls
  #DefineDecisionBoundaries: TakeOptimalDecision with MAP
  #cls_test_col=max.col(Posteriors)#apply(Posteriors,1,which.max) #breaks ties randomly
  cls_train_lab=as.numeric(gsub("C","",names(Priors)))
  
  #ordering in input classes, which do not necessarly are 1:k
  if(isTRUE(requireNamespace("FCPS",quietly = T))){
    if(utils::packageVersion("FCPS")>"1.3.4")
      cls_test=FCPS::ClusterRedefine(cls_test_col,NewLabels = cls_train_lab,OldLabels =1:ncol(Posteriors),Silent = T)
    else{
      warning("Predict_naiveBayes: Please update FCPS")
      cls_test=cls_test_col
    }
  }else{
    warning("Predict_naiveBayes: Please install FCPS")
    cls_test=cls_test_col
  }
  if(!is.null(rownames(Data))){
    names(cls_test)=rownames(Data)
  }

  # Plotting ----
  if(isTRUE(PlotIt)){
    if(is.null(Epsilon)){
      PlotLikelihoods(PDFs,Data = Data,PlotCutOff=PlotCutOff)
    }else{
      PlotLikelihoods(PDFs, Data = Data, PlausibleLikelihoods = PDFs,
                      Epsilon = Epsilon, PlausibleCenters = PlausibleCenters,
                      PlotCutOff = PlotCutOff)
    }
  }
  
  return(list(Cls=cls_test,Posteriors=Posteriors,DataLikelihoodsPerClass=PDFs,ClsTest=cls_test)) #ClsTest for back compatibility, will be deleted in final version
  
}

PlotBayesianDecision2D=function(X, Y, Posteriors, Class = 1, NoBins,
                                CellColorsOrPallette, Showpoints = TRUE,
                                xlim, ylim, xlab, ylab, main, PlotIt = TRUE){
  #ggobj=PlotBayesianDecision2D(X,Y,Posteriors)
  # Zeichnet der Voronoizellen eingefaerbt mit den Posteriors
  #yellow ist sicher klasse 1 und rot sicher klasse 2
  #orange toene sind dazwischen
  #bei multiplen klassen ist es geringer posterior vs hoher posterior
  # ohne Border correction, so dass offene ungefaebten Randzellen zu sehen sind
  # INPUT
  # [X(1:n),Y(1:n)]      Punktkoordinaten
  # Posteriors(1:n,1:Class) matrix of posteriors
  #           
  # OPTIONAL
  # Class             scalar defining which class to look at
  # NoBins            number of bins for class posteriori
  # CellColorsOrPallette either a function defining the color palette of a character vector of colors of length NoBins
  # Showpoints         TRUE, Punkte werden angezeigt
  #xlim,yim
  #PlotIt             TRUE: prints GGPLOT2 object, FALSE: not shown plot
  # OUTPUT
  # GGobj             OBJECT OF GGPLOT2
  # Mapping           [1:m,1:3] matrix that maps colors to Posterior to range of Posteriors values (center of bin=Kernels) for which a color is defined
  #author mct 05/2025
  #1.editor mct 07/2026: bugfix in posterior color mapping, better unique point computation for d>2 cia median.
  if(length(X)!=length(Y)){
    stop("PlotBayesianDecision2D: length of X does not equal length of Y.")
  }
  if(!is.matrix(Posteriors)){
    warning("PlotBayesianDecision2D: Posteriors is not a matrix, calling as.matrix(Posteriors)")
    Posteriors=as.matrix(Posteriors)
  }
  D=Posteriors[,Class]
  
  if(length(X)!=length(D)){
    stop("PlotBayesianDecision2D: length of X does not equal number of rows of Posteriors.")
  }
  if(sum(abs(X-Y))==0){
    warning("PlotBayesianDecision2D: X and Y are identical, cannot polt.")
    return(invisible(NULL))
  }
  if (missing(xlab)) 
    xlab = deparse1(substitute(X))
  if (missing(ylab)) 
    ylab = deparse1(substitute(Y))
  
  # Validate and constrain posterior probabilities
  if (any(!is.finite(D))) {
    stop("PlotBayesianDecision2D: Posteriors contain non-finite values.")
  }
  
  if (any(D < 0 | D > 1)) {
    warning("Posterior values outside [0, 1] are being clamped.")
    D <- pmin(1, pmax(0, D))
  }
  
  if(missing(NoBins)){
    optNrOfBins=DataVisualizations::OptimalNoBins(D)
    optNrOfBins = min(c(100,optNrOfBins)) #
  }else{
    optNrOfBins=NoBins
  }
  # Use length.out rather than by to guarantee exactly NoBins + 1 edges
  edges <- seq(
    from = 0,
    to = 1,
    length.out = optNrOfBins + 1L
  )
  
  Cls <- findInterval(
    D,
    vec = edges,
    all.inside = TRUE
  )
  
  # Important: construct the palette for the complete probability range
  NumberColors <- optNrOfBins
  
  if(!requireNamespace("deldir",quietly = T)){
    stop("PlotBayesianDecision2D: Please install package deldir")
  }
  
  if(!requireNamespace("deldir",quietly = T)){
    stop("PlotBayesianDecision2D: Please install package deldir")
  }

  up = DatabionicSwarm::UniquePoints(cbind(X, Y))
  # Aggregate posteriors of observations sharing the same two-dimensional coordinates.
  Dunique = vapply(up$UniqueInd, function(i) {
    stats::median(D[up$Uniq2DatapointsInd == i])
  }, numeric(1))
  # CellColorPalette = colorRampPalette(c("white", "pink", "yellow", "orange", "red", "firebrick", "darkred"))
  
  if (missing(CellColorsOrPallette)) {
    anchors <- c(
      "#FFFFFF", "#FFF7BC", "#F7E225", "#FDAE61", "#F46D43", "#D73027", "#7F0000" )
    
    Cols <- grDevices::colorRampPalette(anchors, space = "Lab" )(NumberColors)
    
  } else if (is.function(CellColorsOrPallette)) {
    Cols <- CellColorsOrPallette(NumberColors)
    
  } else {
    supplied <- as.character(CellColorsOrPallette)
    
    if (length(supplied) == NumberColors) {
      # Interpret as one explicitly supplied color per bin
      Cols <- supplied
    } else if (length(supplied) >= 2L) {
      # Interpret a shorter vector as palette anchors
      Cols <- grDevices::colorRampPalette(supplied,space = "Lab" )(NumberColors)
    } else {
      stop("At least two palette colors are required.",.Call=TRUE)
    }
  }
  if (length(Cols) != NumberColors) {
    stop("Palette function did not return the requested number of colors.",.Call=TRUE)
  }
  #if(isTRUE(Simple)){
  Kernels <- head(edges, -1L) + diff(edges) / 2
  
  Output <- data.frame(
    Color    = Cols,
    Kernels  = Kernels,
    BinNo    = seq_len(NumberColors),
    Lower    = head(edges, -1L),
    Upper    = tail(edges, -1L),
    Observed = seq_len(NumberColors) %in% unique(Cls),
    stringsAsFactors = FALSE
  )  
  if(missing(xlim)){
    xlim=c(min(X),max(X))
  }
  if(missing(ylim)){
    ylim=c(min(Y),max(Y))
  }
  
  if(length(up$UniqueInd)<3){
    warning("PlotBayesianDecision2D: Not enough unique points in data. Doing nothing.")
    return(invisible(NULL))
  }

  out = try({
    DelTri = deldir::deldir(X[up$UniqueInd],Y[up$UniqueInd],z = Dunique)
    VoronoiCells = deldir::tile.list(DelTri)
  })

  if (inherits(out, "try-error")) {
    return(invisible(list(GGobj=NULL,Mapping=Output)))
  }

#  labels <- factor(CellColor)
  
  # ggplot2 issues of finding variables by names in datastructures
  # CMD check requires the variable to be known (initiated)
  # Naming the variables in a list/data.frame/... is not enough (for CMD check)
  x = NULL
  y = NULL
  PDEnaiveBayes.Voronoi.x     = NULL
  PDEnaiveBayes.Voronoi.y     = NULL
  PDEnaiveBayes.Voronoi.x2    = NULL
  PDEnaiveBayes.Voronoi.y2    = NULL
  PDEnaiveBayes.Voronoi.group = NULL
  PDEnaiveBayes.Voronoi.fill  = NULL
  
  voronoi_df = do.call(rbind, lapply(seq_along(VoronoiCells), function(i) {
    cell = VoronoiCells[[i]]
    cell_posterior = if (!is.null(cell$z)) cell$z else Dunique[cell$ptNum]
    cell_class = findInterval(cell_posterior, vec = edges, all.inside = TRUE)
    data.frame(PDEnaiveBayes.Voronoi.x = cell$x,
               PDEnaiveBayes.Voronoi.y = cell$y,
               PDEnaiveBayes.Voronoi.group = i, #color_label = labels[i],
               PDEnaiveBayes.Voronoi.fill = Cols[cell_class]  # Fill color associated with this cell's posterior
               )
  }))
  
  #globalVariables(c("x", "y", "group", "fill"))
  
  ggobj=ggplot2::ggplot(voronoi_df, ggplot2::aes(x = PDEnaiveBayes.Voronoi.x,
                                                 y = PDEnaiveBayes.Voronoi.y,
                                                 group = PDEnaiveBayes.Voronoi.group,
                                                 fill = PDEnaiveBayes.Voronoi.fill)) +
    ggplot2::geom_polygon(color = "black", show.legend = FALSE) +
    ggplot2::scale_fill_identity() + # +  scale_fill_manual(values = levels(labels), name = "Region Color")
    ggplot2::coord_equal(xlim = xlim, ylim = ylim) +
    ggplot2::theme_minimal() + 
    ggplot2::xlab(xlab) +
    ggplot2::ylab(ylab)
  
  if (Showpoints){
    ggobj = ggobj + ggplot2::geom_point(data = data.frame(PDEnaiveBayes.Voronoi.x2 = X[up$UniqueInd],
                                                          PDEnaiveBayes.Voronoi.y2 = Y[up$UniqueInd]),
                                        ggplot2::aes(PDEnaiveBayes.Voronoi.x2,
                                                            PDEnaiveBayes.Voronoi.y2),
                                        inherit.aes = FALSE)
  }
  if(!missing(main)){
    ggobj=ggobj+ggplot2::ggtitle(main)
  }
  if(isTRUE(PlotIt)){
    print(ggobj)
  }
  return(invisible(list(GGobj=ggobj,Mapping=Output)))
  
}
#i see no sense in this
#globalVariables(c("PDEnaiveBayes.Voronoi.x", "PDEnaiveBayes.Voronoi.y",
#                  "PDEnaiveBayes.Voronoi.group", "PDEnaiveBayes.Voronoi.fill",
#                  "PDEnaiveBayes.Voronoi.x2", "PDEnaiveBayes.Voronoi.y2"))

ApplyBayesTheorem4Likelihoods=function(Likelihoods,Priors,threshold=.Machine$double.eps*1000){
  # V=ApplyBayesTheorem4Likelihoods(ListOfLikelihoods,Priors,threshold=0.00001)
  # 
  # INPUT
  # ListOfLikelihoods    List of m numeric matrices with l columns containing 
  #                      the distribution of class i in 1:l
  # Priors[1:l]          Numeric vector with prior probability for each class.
  # 
  # OPTIONAL
  # threshold           (Optional: Default=0.00001).
  # NormalizeWithGMM    (Optional: Default=FALSE).
  # 
  # OUTPUT
  # Posteriors[1:n, 1:l]    Numeric matrices with posterior probabilities
  #                         according to the bayesian Theorem.
  # 
  
  #__________________________________________________________________________________________  
  # Step 1: # Clip extremely small probabilities account for non finite
  #__________________________________________________________________________________________
  #beachte reihenfoge a,b,c) im vorgehen
  
  if(!is.list(Likelihoods)){
    V=dim(Likelihoods)
    N=V[1]
    class_l=V[2]
    d=V[3]
    Likelihoods[is.finite(Likelihoods) & Likelihoods>=1]=1-threshold #
    Likelihoods <- pmax(Likelihoods, threshold)#sehr wichtig fuer performanz
    #dann erst b)
    Likelihoods[!is.finite(Likelihoods)]=1
    #Likelihoods[Likelihoods==0]=1 #schwerer fehler
    #compute log probabilities in one simple step
    LogPropMat <- apply(log(Likelihoods), c(1, 2), sum)
    
    #print(LogPropMat)
  }else{#Likelihoods is a list (old legacy code)
    class_l=length(Priors)
    N=nrow(Likelihoods[[1]])#durch interpolateDistributionOnData ist jedes listenelement der gleichen laenge gleich laenge der daten
    d=length(Likelihoods)
    ListOfLikelihoods=Likelihoods
        #variante dauert  1369.24 
    # ListOfLikelihoods=lapply(ListOfLikelihoods,FUN = function(xmat,threshold){#iterate features
    #   xmat=apply(xmat, 2, function(xcol,threshold){#iterate classes
    #     #a)       #stelle sicher das kein log(1)=0 oder log(x)>0 fuer gueltig gemessene dichten exisitert, auch im peak
    #     xcol=pmin(xcol, 1 - threshold,na.rm = F) # distribution_current_Class[distribution_current_Class>=1]=1-threshold#b)
    #    # b)#stelle sicher das kein inf addiert wird
    #     xcol=pmax(xcol, threshold,na.rm = F)  # distribution_current_Class[distribution_current_Class<threshold]=threshold
    #     return(xcol)
    #  },threshold)
    #   #ist anscheinend nicht das gleiche wie das apply
    #   # xmat=Rfast::colPmin(x = xmat, y =  1 - threshold)
    #   # xmat=Rfast::colPmax(x = xmat, y = threshold)
    #    #c) erst danach! : #stelle sicher, das addition immer geht aber nicht beachtet wird, log(1)=0
    #    xmat[!is.finite(xmat)]=1
    #    return(xmat)
    #  },threshold)
    #_________________________________________________________________________
    # Step 2 compute Log (joint) Probabilities for every case ----
    #__________________________________________________________________________________________
    # LogPropMat=matrix(0,nrow =N ,ncol = class_l)
    # 
    # 
    # Loop over features (dimensions), summing log-likelihoods for each class
    # for (f in 1:d) {
    #   LogPropMat = LogPropMat + log(ListOfLikelihoods[[f]])
    # }
    
    LogPropMat=matrix(NaN,nrow =N ,ncol = class_l)
    #LogPropMat_plausible=matrix(NaN,nrow = nrow(ListOfLikelihoods[[1]]),ncol = class_l)
    for(cc in 1:class_l){ #variante dauert 601.02, d.h. nur 4 prozent der oberen zeit
      probability=log(Priors[cc])*0
      # if(!is.null(PlausibleCenters)){
      # probability_plausible=log(Priors[cc])*0
      # }
      for(f in 1:d){
        distribution_current_Class=ListOfLikelihoods[[f]][,cc]
        # #bayes theorem ohne normierung, da die fuer MAP ein konstanter faktor
        # #muss sieben zahlen pro case geben, falls sieben klassen gegeben
        #das ist falsch, nur area=1 reicht aus!
       # distribution_current_Class[distribution_current_Class>=1]=1-threshold #stelle sicher das kein log(1)=0 fuer gueltig gemessene dichten, auch im peak
        # #unwissende stellen setzen wir null, damit die reihenfolge in probability sich nicht aendert
        distribution_current_Class[is.finite(distribution_current_Class) & distribution_current_Class>=1]=1-threshold
        distribution_current_Class[!is.finite(distribution_current_Class)]=1#stelle sicher, das addition immer geht aber nicht beachtet wird, log(1)=0
        distribution_current_Class[distribution_current_Class<threshold]=threshold#stelle sicher das kein inf addiert wird
        
        #achtung zahlen kleiner 1 sind negativ, kleinen groesser 1 positiv
        #da die distribution_current_Class kleiner 1 sind, bis auf peaks
        #ist das hier noch keine wahrscheinlichkeit im eigentlichen sinne
        #dieser schritt ist die multipliktion der wahrscheinlichkeiten, d.h. features sind unabhaengig von einander
        probability=probability+log(distribution_current_Class)#minimal kommt hier raus d*threshold
        # if(!is.null(PlausibleCenters)){
        #   distribution_current_Class_plausible=ListOfLikelihoods_plausible[[f]][,cc]
        #   distribution_current_Class_plausible[distribution_current_Class_plausible>=1]=1-threshold #stelle sicher das kein log(1)=0 bei guetigen dichten
        #   #ungemessene dichten/unbekannte werden null
        #   distribution_current_Class_plausible[!is.finite(distribution_current_Class_plausible)]=1
        #   distribution_current_Class_plausible[distribution_current_Class_plausible<threshold]=threshold
        #
        #   probability_plausible=probability_plausible+log(distribution_current_Class_plausible)
        # }
      }#end for all dims
      LogPropMat[1:length(probability),cc]=probability
      # if(!is.null(PlausibleCenters)){
      #   LogPropMat_plausible[1:length(probability_plausible),cc]=probability_plausible
      # }
    }#end for all classes
    
  }
  
  #transform to probabilities
  PropMat=exp(LogPropMat)
  #Compute Evidence
  NormalizationFactor = PropMat %*% Priors
  

  ##Numerical correction of Evidence ----
  #minimum nach threshold oben ist d*threshold
  ZeroInd <- which(NormalizationFactor < threshold*d)
  
  if (length(ZeroInd) > 0) {
    NormalizationFactor[ZeroInd] = threshold*d
  }
  
  # if(!is.null(PlausibleCenters)){
  #   PropMat_plausible=exp(LogPropMat_plausible)
  #   NormalizationFactor_plausible = PropMat_plausible %*% Priors
  # 
  #   ZeroInd <- which(NormalizationFactor_plausible <2*d*threshold)
  #   
  #   if (length(ZeroInd) > 0) {
  #     NormalizationFactor_plausible[ZeroInd] = 2*d*threshold#10^(-7)
  #   }
  # }#end if(!is.null(PlausibleCenters))
  
  Posteriors <- PropMat*0
  ClassDecisionLogProb <- PropMat*0 #ignores normalization and remains in log for better numerical stability
 #____________________________________________________________________________
 # Step 3: Add class priors ----
 #____________________________________________________________________________

 # For every observation i and class c:
 #
 # ClassDecisionLogProb[i,c] =
 #   log p(x_i | C_c) + log P(C_c)
 #
 # This is the logarithm of the unnormalized posterior numerator.
  ClassDecisionLogProb = sweep(LogPropMat, 2, log(Priors), FUN = "+") 

#____________________________________________________________________________
 # Step 4: Compute the evidence safely on the logarithmic scale ----
 #____________________________________________________________________________
 # The evidence for observation i is
 #
 # p(x_i) =
 #   sum_c p(x_i | C_c) * P(C_c).
 #
 # A direct implementation such as
 #
 #   exp(ClassDecisionLogProb)
 #
 # can underflow to zero when many small feature likelihoods are multiplied.
 # The information is still present in ClassDecisionLogProb, so the evidence
 # is calculated with the log-sum-exp identity:
 #
 # log(sum_c(exp(a_c))) =
 #   m + log(sum_c(exp(a_c-m))),
 #
 # where m=max_c(a_c).
 #
 # Subtracting the row maximum ensures that the largest exponent is exp(0)=1.
 # Consequently, the normalization remains numerically stable even when the
 # original log-probabilities are very negative.
  max_log_prob=apply(ClassDecisionLogProb,1,max)
  log_NormalizationFactor=max_log_prob+
    log(rowSums(exp(sweep(ClassDecisionLogProb,1,max_log_prob,FUN = "-"))))

  #____________________________________________________________________________
 # Step 5: Compute normalized posterior probabilities ----
 #____________________________________________________________________________

 # On the logarithmic scale, division by the evidence becomes subtraction:
 #
 # log P(C_c | x_i) =
 #   log[p(x_i | C_c)P(C_c)] - log p(x_i).

  log_Posteriors = sweep(ClassDecisionLogProb, 1, log_NormalizationFactor, FUN = "-")
  Posteriors=exp(log_Posteriors)
  
  # for (i in c(1:class_l)) {
  #   Posteriors[, i] <- exp(LogPropMat[, i]  + log(Priors[i])-log(NormalizationFactor))
  #   ClassDecisionLogProb[, i] <- LogPropMat[, i] + log(Priors[i])
  #     
  # }#end for (i in c(1:class_l)) {

 #____________________________________________________________________________
 # Step 6: Maximum-a-posteriori class assignment ----
 #____________________________________________________________________________

 # Normalization is not required for the class decision because the evidence
 # is identical for all classes of one observation. The exponential function
 # is also strictly increasing, so the maximum logarithmic score gives the
 # same class as the maximum posterior probability.

  # if x1 > x2, then exp(x1) > exp(x2)
  #max.col resolves ties randomly
  Cls=max.col(ClassDecisionLogProb)#we do not need normalization for defining class
  
  # if(!is.null(PlausibleCenters)){
  #   #renormalisierung der posterioirs
  #   #dreht mir die matrix
  #   #Posteriors=apply(Posteriors,1,function(x) x/sum(x) )
  #   for(i in 1:nrow(Posteriors)){
  #     Posteriors[i,]=Posteriors[i,]/sum(Posteriors[i,])
  #   }
  # }
  return(list(Posteriors=Posteriors,Cls=Cls))
}

defineOrEstimateDistribution=function(Feature,ClassInd,Gaussian=FALSE,
                                      ParetoRadius=NULL,
                                      InternalPlotIt=FALSE,
                                      SD_Threshold=0.001,
                                      ...){
  # INPUT
  # Feature[1:n]    Numeric Vector
  # ClassInd        Integer Vector with class indices
  # 
  # OPTIONAL
  # Gaussian     (Optional: Default=TRUE). Assume gaussian distribution.         
  # Robust       (Optional: Default=FALSE). Robust computation if set to TRUE.
  # PlotItInternal       (Optional: Default=FALSE).
  # na.rm        (Optional: Default=TRUE). Remove na.
  # SD_Threshold    (Optional: Default=0.00001). Threshold value for std.
  # Type          1: MCT PDE, 2: QS PDE
  # OUTPUT
  # Kernels[1:m]    Numeric vector with domain of a 1D pdf
  # PDF[1:m]        Numeric vector of pdf
  # Theta           Numeric vector with parameters of gaussian - NULL if no gaussian used.
  # 
  dots=list(...)
  has_Robust <- "Robust" %in% names(dots)
  if (has_Robust) {
    Robust= dots$Robust
    if(Robust==1) Robust=TRUE
    if(Robust==0) Robust=FALSE
  }else{
    Robust=FALSE
  }
  has_Type <- "Type" %in% names(dots)
  if (has_Type) {
    Type= dots$Type
  }else{
    Type=1
  }
  
  has_narm <- "na.rm" %in% names(dots)
  if (has_narm) {
    na.rm= dots$na.rm
  }else{
    na.rm=TRUE
  }
  Fun=NULL
  QuantityThreshold = 30
  UniqueValuesThreshold = 12
  MinMax = c(min(Feature,na.rm = T), max(Feature,na.rm = T))

  if(sum(is.finite(MinMax))!=2){
	#uniqueFeature=1
    if(!is.finite(MinMax[1])) MinMax[1]=0#dummy value
    if(!is.finite(MinMax[2])) MinMax[2]=1#dummy value
  }else{
    #uniqueFeature=unique(Feature)
	#uniqueFeature=uniqueFeature[is.finite(uniqueFeature)]#no missing or inf
  }

  #kernels_p=sort(uniqueFeature) #automatische setzung der kernels (kernels_p=NULL) verschlechtert das ergebnis deutlich
  #n=min(c(length(uniqueFeature),200))
  # breaks = pretty(c(min(Feature,na.rm = T), max(Feature,na.rm = T)), n = n, min.n = 1)
  # nB = length(breaks)
  # kernels_p = 0.5 * (breaks[-1L] + breaks[-nB])
  #kernels_p=seq(from=min(Feature,na.rm = T),to= max(Feature,na.rm = T),length.out=n)
  if(isTRUE(Gaussian)){#define distibution
    Theta=fitParameters(Feature,ClassInd=ClassInd,Robust=Robust,na.rm=na.rm,SD_Threshold=SD_Threshold)
    pdf=dnorm(Feature,Theta[1],Theta[2])#already likelihood on given data
    Kernels=Feature
  }else{#estimate Distribution
    #Performanz haeng stark davon ab wie hier die kernels gewaehlt werden
    if(isFALSE(requireNamespace("DataVisualizations"))){
      Type=2
      N_UniqueFC=20
      warning("Please install DataVisualizations. setting Type of PDE to 2")
    }
    Theta=NULL
    Feature_Class=Feature[ClassInd]
    if(isTRUE(na.rm))
      Feature_Class=Feature_Class[is.finite(Feature_Class)]
    N_UniqueFC=length(unique(Feature_Class))
    if(is.null(ParetoRadius)){
	  if(!requireNamespace("DataVisualizations",quietly = T)){
		warning("Please install package DataVisualizations.")
		return(NULL)
	  }else{
		if(packageVersion("DataVisualizations")<"1.1.5"){
		  warning("Please update package DataVisualizations to version 1.1.5 or higher.")
		  return(NULL)
		}
	  }
      ParetoRadius=if(packageVersion("DataVisualizations")>="1.4.0")DataVisualizations::ParetoRadius_fast(Feature) else DataVisualizations::ParetoRadius(Feature)

      
      #bayes is sensitiv gegenueber radius ->volles feature nehmen ist besser als uber class feature
    #in spezialaellen ist die interne berechnung nicht durchfuerbar
    }#ind dem fall wird NULL zurueckgegeben
    if(!is.finite(ParetoRadius)){
      #in dem fall koennen wir keine dichte schätzen
      N_UniqueFC=0#fuert zu dem fall von densityEstimation4smallNoCases()
	  ParetoRadius=0
    }
    
   if(N_UniqueFC>UniqueValuesThreshold & length(Feature_Class)>QuantityThreshold){#density estimation by PDE is possible, see MDplot paper
      #if(Type==1 | N_UniqueFC<UniqueValuesThreshold){#Type=2 scheint dort nicht alle daten bei dichte messung zu beruecksichten
    if(Type==1 ){
      #KernelVec has to be equally spaced
      KernelVec=seq(from=MinMax[1]-ParetoRadius,to=MinMax[2]+ParetoRadius,length.out=512)
	    if(!requireNamespace("DataVisualizations",quietly = T)){
			warning("Please install package DataVisualizations.")
			return(NULL)
		  }else{
			if(packageVersion("DataVisualizations")<"1.1.5"){
			  warning("Please update package DataVisualizations to version 1.1.5 or higher.")
			  return(NULL)
			}
		  }
      V=DataVisualizations::ParetoDensityEstimation(Feature_Class, Silent = T,kernels = KernelVec,
                                                      Compute = "cpp", paretoRadius = ParetoRadius)
        paretoRadius=V$paretoRadius
        Kernels=V$kernels#[-c(1,length(V$kernels))]
        paretoDensity=V$paretoDensity#[-c(1,length(V$kernels))]
        # FFT-based smoothing step 
        # assume Kernels is equally spaced
        #but there are always numerical instabilities, hence mean
        dx = mean(diff(Kernels))
        m  = length(Kernels)
        #midpoint
        mid = ceiling(m/2)
        #computes the physical distances of each grid point from the center of your kernel grid
        rel = (seq_len(m) - mid) * dx
        # Gaussian smoothing kernel: use paretoRadius as bandwidth
        kFFT = dnorm(rel / paretoRadius) / paretoRadius * dx
        #rectangle kernel
        # kFFT = ifelse(abs(rel) <= paretoRadius,
        #               dx / (2 * paretoRadius),
        #               0)
        # Zero-pad to length L
        #choose L as the next power of two at least m + m - 1 so 
        # that when we multiply FFTs we perform a linear convolution
        L = 2^ceiling(log2(2*m - 1))
        #Zero-padding both vectors to length L prevents wrap-around artifacts
        #from the FFT and leverages FFT speedups for power-of-two lengths
        w_pad = c(paretoDensity, rep(0, L - m))# original (normalized) Pareto density vector extended with zeros up to length L
        k_pad = c(kFFT, rep(0, L - m))#discrete smoothing kernel vector (kFFT) similarly zero-padded to length L.
        # Convolution via FFT
        #adding the kernel to the same length as w_pad is required so their FFTs can be multiplied element-wise to compute the convolution.
        conv = fft(fft(w_pad) * fft(k_pad), inverse = TRUE) / L
        #similar to conv2 <- convolve(paretoDensity, kFFT, type = "open") #but with controlling the FFT length and power-of-two padding
        # Extract smoothed density
        #ignore zero padding which is symmetrically distributed to both sides
        pdf = Re(conv[((L-m)/2):(ceiling((2*m+L-m)/2)-1)])#Re(conv[1:(2*m)])
        # #centered non-causal moving average
        # n=round(0.05*length(Kernels),0)
        # pdf= stats::filter(V$paretoDensity, rep(1/n, n), sides = 2)
        # #left side moving average, causal
        # #pdf=TSAT::MovingAverage(V$paretoDensity,25)
        # #kein smoothing  an den enden
        # pdf[1:n]=V$paretoDensity[1:n]
        # pdf[(length(pdf)-n):length(pdf)]=V$paretoDensity[(length(pdf)-n):length(pdf)]
        # # pdf=V$paretoDensity
    
        # if(length(Feature_Class)>3&length(Feature_Class)<1000){
        #   Feature_Class2=sampleByPDE(Kernels,pdf,1000)
        #     V2=DataVisualizations::ParetoDensityEstimation(c(Feature_Class,Feature_Class2),MinAnzKernels = 800,KernelRange = MinMax,Silent = T,Compute = "cpp_exp",paretoRadius = ParetoRadius)
        #     paretoRadius=V2$paretoRadius
        #     pdf=V2$paretoDensity
        #     Kernels=V2$kernels
        # }
    }else if(Type==2){##type=2, Rs density estimation
        V=density(Feature_Class,na.rm=T)
        pdf=V$y
        Kernels=V$x
    }else if(Type==3){
      V=density(Feature_Class,na.rm=T,bw=ParetoRadius)
      pdf=V$y
      Kernels=V$x
    }else{
      stop("defineOrEstimateDistribution: Select type Type=1 or Type=2 or Type=3")
    }
  
        #stats::spline scheint nicht so gut zu sein

         #pdf=pdf/quantile(pdf,c(0.999)) #sonst sind feature mit dichte >1 hoeher gewichtet
         #quantile ignoriert peaks
        # c_pdfV=stats::spline(Kernels, pdf, xout =smoothX, ties = "ordered",method = "fmm")
        # pdf=c_pdfV$y
        # Kernels=c_pdfV$x
      # }else{# if(Type==2){
      #   V=   ParetoDensityEstimation2(Feature_Class,Kernel = kernels_p,paretoRadius = ParetoRadius)
      #   pdf=V$Density  
      #   Kernels=V$Kernel
      #   # if(length(Feature_Class)>3&length(Feature_Class)<1000){
      #   #   Feature_Class2=sampleByPDE(Kernels,pdf,1000)
      #   #   V2=ParetoDensityEstimation2(c(Feature_Class,Feature_Class2),Kernel = kernels_p,paretoRadius = ParetoRadius)
      #   #   pdf=V2$Density  
      #   #   Kernels=V2$Kernel
      #   # }
      # }#end density estimation of type 2
      }else{#fall back to hist-based density estimation
        V=densityEstimation4smallNoCases(Feature,ClassInd)
        pdf=V$Density  
        Kernels=V$Kernels
      }#end if QuantityThreshold and UniqueValuesThreshold is appropriate
    
    ##no error catching if something went wrong
    # if(length(Kernels)!=length(pdf)){
    #   warning("defineOrEstimateDistribution: Unable to estimate PDE correctly, please check input data, fallback to densityEstimation4smallNoCases()")
    #   V=densityEstimation4smallNoCases(Feature,ClassInd)
    #   pdf=V$Density  
    #   Kernels=V$Kernels
    # }
    if(sum(is.finite(pdf))>1&sum(is.finite(Kernels))>1){#at least two values are required
      #ist ein fehler, weil ich moeglicherweise die eine klasse dann niedriger als
      # die andere klasse gewichte
      # if(sum(pdf>1,na.rm = T)/length(ClassInd)>0.05){#more than five percent of pdf values is above one
      #   #we have peaks in density which will result in bad liklihoods in log
      #   pdf=pdf/max(pdf,na.rm=TRUE)#now pdf is not normalized to area of one
      # }
    }else{
      warning("defineOrEstimateDistribution: Unable to estimate PDE correctly, please check input data, fallback to densityEstimation4smallNoCases()")
      V=densityEstimation4smallNoCases(Feature,ClassInd)
      pdf=V$Density  
      Kernels=V$Kernels
    }

  }# end if(isTRUE(Gaussian))
  if(isFALSE(Gaussian)){
    if(N_UniqueFC>UniqueValuesThreshold & length(Feature_Class)>QuantityThreshold){
      if(sum(is.finite(pdf))>2 &sum(is.finite(Kernels))>2){
        smoothX  = seq(from=min(Kernels,na.rm = T), to=max(Kernels,na.rm = T),length.out=1000)
        Fun=stats::splinefun(x = Kernels, y = pdf,method="monoH.FC", ties = "ordered")
        pdf=Fun(smoothX)
        Kernels=smoothX
      }else{

        Kernels= c(MinMax[1],MinMax[2])
        pdf=c(0,0)
        Fun=stats::approxfun(x = Kernels, y = pdf,rule = 2, ties = "ordered")
        pdf=Fun(Kernels)
      }
    }else{
      if(sum(is.finite(pdf))>2 &sum(is.finite(Kernels))>2){
        smoothX  = seq(from=min(Kernels,na.rm = T), to=max(Kernels,na.rm = T),length.out=1000)
        Fun=stats::approxfun(x = Kernels, y = pdf,rule = 2, ties = "ordered")
        pdf=Fun(smoothX)
        Kernels=smoothX
      }else{
        Kernels= c(MinMax[1],MinMax[2])
        pdf=c(0,0)
        Fun=stats::approxfun(x = Kernels, y = pdf,rule = 2, ties = "ordered")
        pdf=Fun(Kernels)
      }
    }
    # V=get_pdf_fun(Kernels,pdf,length.out=1000,N_UniqueFC=N_UniqueFC,UniqueValuesThreshold=UniqueValuesThreshold,N_class=length(ClassInd),QuantityThreshold=QuantityThreshold)
    # Fun=V$PDFfun
    # pdf=V$pdf
    # Kernels=V$Kernels
  }
  if(isTRUE(InternalPlotIt)){
   plot(Kernels,pdf,xlab="If Feature where in Class",ylab="PDF")
  }
  return(list(Kernels=Kernels,PDF=pdf,Theta=Theta,PDF_fun=Fun,ParetoRadius=ParetoRadius))
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

quantile4LargeVectors <- function(x, probs) {
    .Call(`_PDEnaiveBayes_quantile4LargeVectors`, x, probs)
}

PlotNaiveBayes = function(Model,FeatureNames,ClassNames,DatasetName = "Data",nrows = 1,FeatureOrderOrSubset,NumFeaturesPerRow = 4,Colors, IndividualFigures = FALSE){
  
  Kernels     = Model$c_Kernels_list
  Likelihoods = Model$ListOfLikelihoods
  Priors      = Model$Priors
  
  if(missing(Colors)){
    Colors = c("blue",         "gold", "firebrick",  "limegreen",
               "dodgerblue", "orange",   "magenta",      "green",
               "cyan",          "red",      "plum",  "darkgreen",
               "turquoise",  "bisque",  "lavender", "lightgreen")
    
    Colors = c("black", #a
               "gold", #b
               "firebrick", #c
               "plum", #d
               "blue", "steelblue", # e1 e2
               "cyan", #f
               "purple", "magenta", #g1 g2
               "green", "limegreen",  "darkgreen", # h1 h2 h3
               "bisque",  # i
               "turquoise",  "orange", "lightgreen") # Remainder
  }
  
  if(length(Kernels) == 0L){
    stop("PlotNaiveBayes: Model$c_Kernels_list contains no features.")
  }
  
  NumClasses  = dim(Kernels[[1]])[2]
  NumFeatures = length(Kernels)
  
  if(missing(FeatureNames)){
    FeatureNames = paste0("X", seq_len(NumFeatures))
  }
  
  if(length(FeatureNames) != NumFeatures){
    stop(
      "PlotNaiveBayes:`FeatureNames` must have one name for every model feature: ",
      NumFeatures, " names are required."
    )
  }
  
  if(missing(FeatureOrderOrSubset)){
    FeatureOrderOrSubset = seq_len(NumFeatures)
  }
  
  if(!is.numeric(FeatureOrderOrSubset) || anyNA(FeatureOrderOrSubset) || any(FeatureOrderOrSubset %% 1 != 0)){
    stop("PlotNaiveBayes:`FeatureOrderOrSubset` must contain integer feature indices.")
  }
  
  FeatureOrderOrSubset = as.integer(FeatureOrderOrSubset)
  
  if(length(FeatureOrderOrSubset) == 0L){
    stop("PlotNaiveBayes:`FeatureOrderOrSubset` must select at least one feature.")
  }
  
  if(any(FeatureOrderOrSubset < 1L | FeatureOrderOrSubset > NumFeatures)){
    stop(
      "PlotNaiveBayes: Every value in `FeatureOrderOrSubset` must be between 1 and ",
      NumFeatures, "."
    )
  }
  
  if(anyDuplicated(FeatureOrderOrSubset)){
    stop("PlotNaiveBayes:`FeatureOrderOrSubset` must not contain duplicate indices.")
  }
  
  if(missing(ClassNames)){
    ClassNames = paste0("C", seq_len(NumClasses))
  }
  
  if(length(ClassNames) != NumClasses){
    stop(
      "PlotNaiveBayes:`ClassNames` must have one name for every class: ",
      NumClasses, " names are required."
    )
  }
  
  if(length(Priors) != NumClasses){
    stop("PlotNaiveBayes: The number of priors does not match the number of classes.")
  }
  
  if(length(Colors) < NumClasses){
    stop(
      "PlotNaiveBayes: At least ", NumClasses,
      " colors are required, but only ", length(Colors),
      " were supplied."
    )
  }
  
  if(length(NumFeaturesPerRow) != 1L || is.na(NumFeaturesPerRow) || NumFeaturesPerRow < 1L || NumFeaturesPerRow %% 1 != 0) {
    stop("PlotNaiveBayes:`NumFeaturesPerRow` must be a positive integer.")
  }
  
  NumFeaturesPerRow = as.integer(NumFeaturesPerRow)
  
  # Apply the class priors only to the selected likelihoods.
  for(FeatureIndex in FeatureOrderOrSubset){
    for(ClassIndex in seq_len(NumClasses)){
      Likelihoods[[FeatureIndex]][, ClassIndex] =
        Priors[ClassIndex] *
        Likelihoods[[FeatureIndex]][, ClassIndex]
    }
  }
  
  NumSelectedFeatures = length(FeatureOrderOrSubset)
  ListFigs            = vector("list", NumSelectedFeatures)
  
  # PlotPosition is the position within the requested subset/order.
  # FeatureIndex is the original feature index in the model.
  for(PlotPosition in seq_along(FeatureOrderOrSubset)){
    
    FeatureIndex = FeatureOrderOrSubset[PlotPosition]
    
    CurrKernel = Kernels[[FeatureIndex]]
    CurrLikeli = Likelihoods[[FeatureIndex]]
    
    if(nrow(CurrKernel) < 2L){
      stop("PlotNaiveBayes: Kernel for feature ", FeatureIndex,
        " must contain at least two rows."
      )
    }
    
    if(ncol(CurrKernel) != NumClasses || ncol(CurrLikeli) != NumClasses) {
      stop("PlotNaiveBayes: Kernel and likelihood matrices for feature ",
        FeatureIndex,
        " must have one column per class."
      )
    }
    
    if (nrow(CurrKernel) != nrow(CurrLikeli)) {
      stop(
        "Kernel and likelihood matrices for feature ",
        FeatureIndex,
        " have different numbers of rows."
      )
    }
    
    # Extend the curves so the filled regions return to zero.
    Diff = CurrKernel[2, 1] - CurrKernel[1, 1]
    
    CurrKernel = rbind(
      CurrKernel[1, ] - Diff,
      CurrKernel,
      CurrKernel[nrow(CurrKernel), ] + Diff
    )
    
    CurrLikeli = rbind(
      rep(0, NumClasses),
      CurrLikeli,
      rep(0, NumClasses)
    )
    
    # Show the legend on the final figure of each row.
    IsEndOfRow = PlotPosition %% NumFeaturesPerRow == 0L
    IsLastPlot = PlotPosition == NumSelectedFeatures
    ShowLegend = IsEndOfRow || IsLastPlot
    
    fig = plotly::plot_ly(fill = "tozeroy")
    
    for(ClassIndex in seq_len(NumClasses)){
      
      RgbaLine = paste0(
        "rgba(",
        paste0(grDevices::col2rgb(Colors[ClassIndex]), collapse = ", "),
        ", 1)"
      )
      
      RgbaFill = paste0(
        "rgba(",
        paste0(grDevices::col2rgb(Colors[ClassIndex]), collapse = ", "),
        ", 0.2)"
      )
      
      fig = plotly::add_trace(
        p           = fig,
        y           = CurrKernel[, ClassIndex],
        x           = CurrLikeli[, ClassIndex],
        type        = "scatter",
        mode        = "lines",
        name        = ClassNames[ClassIndex],
        legendgroup = ClassNames[ClassIndex],
        line        = list(color = RgbaLine),
        fillcolor   = RgbaFill,
        showlegend  = FALSE
      )
      
      fig = plotly::add_trace(
        p           = fig,
        y           = CurrKernel[, ClassIndex],
        x           = -CurrLikeli[, ClassIndex],
        type        = "scatter",
        mode        = "lines",
        name        = ClassNames[ClassIndex],
        legendgroup = ClassNames[ClassIndex],
        line        = list(color = RgbaLine),
        fillcolor   = RgbaFill,
        showlegend  = ShowLegend
      )
    }
    
    fig = plotly::layout(
      p = fig,
      xaxis = list(
        title     = FeatureNames[FeatureIndex],
        titlefont = list(size = 20, face = "bold"),
        tickfont  = list(size = 12)
      ),
      yaxis = list(
        title     = "Class-cond. PDE",
        titlefont = list(size = 18, face = "bold"),
        tickfont  = list(size = 12)
      )
    )
    
    # The list follows FeatureOrderOrSubset rather than the original order.
    ListFigs[[PlotPosition]] = fig
  }
  
  names(ListFigs) = FeatureNames[FeatureOrderOrSubset]
  
  if(isTRUE(IndividualFigures)){
    return(ListFigs)
  }
  
  CreateSubplot = function(Figures,SelectedFeatureIndices,DatasetName){
    
    NumFigures = length(Figures)
    
    if(NumFigures < 1L || NumFigures > NumFeaturesPerRow){
      stop("PlotNaiveBayes`CreateSubplot` received an unsupported number of figures: ",
        NumFigures, "."
      )
    }
    
    # Preserve the original quarter-width behavior for rows of up to four.
    # For larger NumFeaturesPerRow values, use equal widths.
    if(NumFeaturesPerRow == 4L){
      Widths = rep(1 / 4, NumFigures)
    }else{
      Widths = rep(1 / NumFigures, NumFigures)
    }
    
    TmpMargin = c(0.05, 0.05, 0, 0)
    
    TmpPlot = plotly::subplot(
      Figures,
      nrows  = 1,
      titleX = TRUE,
      titleY = TRUE,
      margin = TmpMargin,
      widths = Widths,
      heights = 1
    )
    
    PlotMargin = list(
      l   = 0,
      r   = 0,
      b   = 0,
      t   = 50,
      pad = 1
    )
    
    FeatureLabel = paste(SelectedFeatureIndices, collapse = ", ")
    
    if(NumFigures == 1L){
      PlotTitle = paste0(
        DatasetName,
        " Feature ",
        FeatureLabel
      )
    }else{
      PlotTitle = paste0(
        DatasetName,
        " Features ",
        FeatureLabel
      )
    }
    
    plotly::layout(
      p      = TmpPlot,
      title  = PlotTitle,
      font   = list(size = 18),
      margin = PlotMargin
    )
  }
  
  PlotMe = list()
  
  RowStarts = seq.int(
    from = 1L,
    to   = NumSelectedFeatures,
    by   = NumFeaturesPerRow
  )
  
  for(RowIndex in seq_along(RowStarts)){
    
    Start = RowStarts[RowIndex]
    End   = min(
      Start + NumFeaturesPerRow - 1L,
      NumSelectedFeatures
    )
    
    PlotMe[[RowIndex]] = CreateSubplot(
      Figures = ListFigs[Start:End],
      SelectedFeatureIndices =
        FeatureOrderOrSubset[Start:End],
      DatasetName = DatasetName
    )
  }
  
  return(PlotMe)
}

GetLikelihoodFunction=function(Kernels_list,ListOfLikelihoods){

  PDFs_funs = list()
  d = length(Kernels_list)
  class_len = ncol(Kernels_list[[1]])
  for (i in 1:d) {
    c_pdf = list()
    c_kernel = list()
    c_theta = list()
    funs = list()
    KernelsMat = Kernels_list[[i]]
    LL = ListOfLikelihoods[[i]]
    for (cc in 1:class_len) {
      Kernels = KernelsMat[, cc]
      pdf = LL[, cc]
      ind=is.finite(Kernels)&is.finite(pdf)
      Kernels=Kernels[ind]
      pdf=pdf[ind]
	   if(length(pdf)>2 &length(Kernels)>2){
		  smoothX  = seq(
			from = min(Kernels,na.rm=T),
			to = max(Kernels,na.rm=T),
			length.out = 1000
		  )
        Fun=stats::splinefun(x = Kernels, y = pdf,method="monoH.FC", ties = "ordered")
      }else{
        mm=min(Kernels,na.rm=T)
        nn=max(Kernels,na.rm = T)
        if(!is.finite(mm)) mm=0
        if(!is.finite(nn)) nn=1
        Fun=stats::approxfun(x = c(mm,nn), y = c(0,0),rule = 2, ties = "ordered")
      }
      funs[[cc]] = Fun
      #alles ist in reihenfolge der klassen anzuordnen!
    }
    PDFs_funs[[i]] = funs
  }#end for each class
  PDFs_funs[[i]] = funs
  return(PDFs_funs)
  
}#GetLikelihoodFunction

densityEstimation4smallNoCases=function(FeatureFull,ClassInd){
  
  FeatureClass=FeatureFull[ClassInd]
  FeatureClass = FeatureClass[is.finite(FeatureClass)]
  
  D = FeatureFull[is.finite(FeatureFull)]
  if(length(D)==0){
    warning("densityEstimation4smallNoCases: no finite values in current data column.")
    return(list("Kernels"       = c(0,1),
                "Density"      = c(0,0)))
  }
 
  MinD = min(D, na.rm = TRUE)
  MaxD = max(D, na.rm = TRUE)
  if(length(FeatureClass)==0){
    warning("densityEstimation4smallNoCases: no finite values in the current class.")
    if(MinD==MaxD)
      MaxD=MinD+1
    return(list("Kernels"       = c(MinD,MaxD),
                "Density"      = c(0,0)))
  }
  if(MinD==MaxD){
    warning("densityEstimation4smallNoCases: only one unique value in current data column. Please check input data")
    #one unique value in data
    #make sure that two kernels are given back
    if(is.finite(MinD)){
      if(MinD<0){
        return(list("Kernels"       = c(MinD,0),
                    "Density"      = c(1,0)))
      }else if(MinD>0){
        return(list("Kernels"       = c(0,MinD),
                    "Density"      = c(0,1)))
      }else{
        return(list("Kernels"       = c(MinD,1),
                    "Density"      = c(1,0)))
      }
    }else{
      return(list("Kernels"       = c(0,1),
                  "Density"      = c(0,0)))
    }
  }
  if(requireNamespace("DataVisualizations",quietly = TRUE))
    optNrOfBins=DataVisualizations::OptimalNoBins(D)
  else
    optNrOfBins=grDevices::nclass.scott(D)
  optNrOfBins = max(c(1,min(c(100,optNrOfBins)))) #
  
  edges = seq(MinD, MaxD, length.out = optNrOfBins+1L)
  bin_width <- diff(edges)[1L]
  
  bin_indices=findInterval(FeatureClass,vec = edges,all.inside = TRUE)
  
  bin_counts <- tabulate(bin_indices, nbins = optNrOfBins)
  
  Kernels=edges[-1]-diff(edges)/2

  Density=bin_counts/(sum(bin_counts) * bin_width)

  if(length(Kernels)==length(Density)){
    return(list("Kernels"       = Kernels,
                "Density"      = Density))
  }else{
    warning("densityEstimation4smallNoCases: unable to count values in current data column. Please check input data")
    Kernels=c(MinD,MaxD)
    return(list("Kernels"       = Kernels,
                "Density"      = c(0,0)))
  }

}

sampleByPDE <- function(Kernels, pdf, n) {
  #ind=which(!duplicated(Kernels))
  #Kernels=Kernels[ind]
  #pdf=pdf[ind]
  if(sum(is.finite(Kernels))>3&sum(is.finite(pdf))>3){
    
    if(length(unique(Kernels))>3&length(unique(pdf))>3){
    dx      <- diff(Kernels)
    mass    <- c(0, cumsum(0.5 * (pdf[-1] + pdf[-length(pdf)]) * dx))
    cdf     <- mass / tail(mass, 1)
    
    u       <- runif(n)
    #u=unique(u)
    feature=approx(cdf, Kernels, xout = u,rule = 2, ties = "ordered")$y #rule = 1, ties = "ordered"
    }else{
      feature=NULL
    }
  }else{
    feature=NULL
  }
  return(feature)
}

PlotLikelihoodFuns=function(LikelihoodFuns, Data, PlausibleLikelihoodFuns = NULL,
                            Epsilon = NULL, PlausibleCenters = NULL,
                            PlotCutOff = 4, xlim){
  # PlotLikelihoodFuns(Likelihoods,Data)
  # Input
  # Likelihoods            List of d numeric matrices, one per feature, each matrix with 1:k columns containing 
  #                             the distribution of class 1:k per feature, i.e., the Likelihood per class
  # Data                         numerical data matrix [1:n,1:k]
  # OPTIONAL
  # PlausibleLikelihoods     List of d numeric matrices of plausible Likelihood, one per feature, each matrix with [1:n,1:k] with k columns containing 
  #                             the distribution of class 1:k per feature 
  # Epsilon                      scalar defining epsiolon fo plausible likelihoods
  # PlausibleCenters             [d:k] plausible centers used to compute plausible likelihoods
  # PlotCutOff                   scalar defining the how many feature starting from 1 should ne plottet or numerical vector defining the index of features to be plottet
  #                             in second case should not be too many otherwise plot yields an error
  # OUTPUT
  # native plots in grid
  
  #author MCT 05/2025
  #reset par after plotting
  if(!is.matrix(Data)){
    Data=as.matrix(Data)
  }
  
  def.par <- par(no.readonly = TRUE)
  on.exit(par(def.par))
  
  Likelihoods=interpolatePDF4Data(Data,LikelihoodFuns)
  
  if(!is.null(PlausibleLikelihoodFuns))
    PlausibleLikelihoods=interpolatePDF4Data(Data,PlausibleLikelihoodFuns)
  else
    PlausibleLikelihoods=NULL
  
    V=dim(Likelihoods)
    N=V[1]
    class_l=V[2]
    dinit=V[3]
  

  
  if(!is.null(PlotCutOff)){
    if(is.numeric(PlotCutOff)){
      if(length(PlotCutOff)==1){
        d=min(c(dinit,PlotCutOff),na.rm = T)
        rangePlots=1:d
      }else{
        rangePlots=PlotCutOff
        rangePlots[rangePlots>dinit]=dinit
        rangePlots=rangePlots[rangePlots>0]
        d=max(rangePlots,na.rm = T)-min(rangePlots,na.rm = T)+1
      }
    }else{
      rangePlots=1:dinit
      d=dinit
    }
  }else{
    rangePlots=1:dinit
    d=dinit
  }
  
  if(is.null(PlausibleLikelihoods)){
    par(mfrow = c(d, 1))  # d rows, 1 columns
  }else{
    par(mfrow = c(d, 2))  # d rows, 2 columns
  }
  
  if(isTRUE(requireNamespace("DataVisualizations",quietly = T))){
    Colors=DataVisualizations::DefaultColorSequence[1:class_l]
  }else{
    warning("PlotLikelihoods: Please install DataVisualizations package for color setting.")
    Colors=rep(c("black","blue","green","gold","magenta","red","grey","orange"),class_l)
    Colors=Colors[1:class_l]
  }
  
  #plot all dimensions
  
  for(f in rangePlots){
    x=Data[,f]
    ind=order(x,decreasing = F)
    fnam=colnames(Data)[f]
    if(is.list(Likelihoods)){
      LL=Likelihoods[[f]]
    }else{
      LL=Likelihoods[,,f]
    }
  
    maxY1=apply(LL,2,function(x) max(x[is.finite(x)]))
    
    if(!is.null(PlausibleLikelihoods)){
      if(is.list(Likelihoods)){
        LL2=PlausibleLikelihoods[[f]]
      }else{
        LL2=PlausibleLikelihoods[,,f]
      }
      maxY2=apply(LL2,2,function(x) max(x[is.finite(x)]))
      ylim=c(0,max(c(maxY1,maxY2)))
    }else{
      ylim=c(0,c(max(maxY1)))
    }
    
    if(missing(xlim)){
      plot(x[ind],LL[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],main="Bayesian",
           ylim=ylim,lwd=2,ylab="Class Likelihoods")
    }else{
      plot(x[ind],LL[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],main="Bayesian",
           ylim=ylim,lwd=2,ylab="Class Likelihoods",xlim=xlim)
    }
    for(i in 2:class_l){
      points(x[ind],LL[ind,i],type="l",col=Colors[i],ylim=ylim,lwd=2)
    }
    if(!is.null(PlausibleLikelihoods)){
      if(missing(xlim)){
        plot(x[ind],LL2[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],ylab="Class Likelihoods",
             main=paste0("Plausible Bayes with approx. Limit ",round(Epsilon[f]/class_l,3)),ylim=ylim,lwd=2)
      }else{
        plot(x[ind],LL2[ind,1],type="l",xlab=paste("Feature",fnam),col=Colors[1],ylab="Class Likelihoods",
             main=paste0("Plausible Bayes with approx. Limit ",round(Epsilon[f]/class_l,3)),ylim=ylim,lwd=2,xlim=xlim)
      }
  
      if(!is.null(PlausibleCenters)){
        PlausibleCenters_cur=PlausibleCenters[f,]
        abline(v=PlausibleCenters_cur[1],col=Colors[1])
      }
      for(i in 2:class_l){
        points(x[ind],LL2[ind,i],type="l",col=Colors[i],ylim=ylim,lwd=2)
        if(!is.null(PlausibleCenters)){
          abline(v=PlausibleCenters_cur[i],col=Colors[i])
        }
      }
    }#end in case of plausible
  }#end for rangePlots
}

---
title: "An Introduction to PDE naive Bayes (PDENB)"
author: "Michael C. Thrun"
date: "`r format(Sys.time(), '%d %b %Y')`"
output: 
          html_document:
            theme: united
            highlight: tango 
            toc: true
            number_sections: true
            doc_depth: 2
            toc_float: true
            fig.width: 8
            fig.height: 8
vignette: >
  %\VignetteIndexEntry{An Introduction to PDE naive Bayes (PDENB)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r setup, include=FALSE}
# setupKnitr()
# knitr::opts_chunk$set(echo = TRUE,
#                       fig.align = "center",
#                       warning = FALSE,
#                       webgl = TRUE,
#                       fig.width = 8, 
#                       fig.height = 8,
#                       fig.keep = "all",
#                       fig.ext = "jpeg"
#                       )
```
# PDE Naive Bayes (PDENB)

In naive Bayes assumptions about the feature distributions are necessary. Classically the assumption is, that the features are independent and gaussian distributed. The assumption of a Gaussian feature distribution might not reflect the real feature distributions, potentially impacting the model quality of the naive Bayes classifier. In contrast the PDENB uses a nonparametric approach in estimating the feature densities using the Pareto Density Estimation (PDE). The PDE is a hyper parameter free density estimator, since the Pareto Radius, the analog to the kernel band width in Kernel Density Estimators (KDE) is derived from the data itself using a information theoretic approach [Ultsch, 2005]. This gives the PDENB classifier more flexibility on modeling the feature distributions and reduce the reliance on pre defined assumption on the data [Stier et al., 2026]. Additionally the PDENB allows for a plausible posterior corrections in regions with high uncertainty [Stier et al., 2026].


## Fitting a PDENB model

First load the Hepta dataset, which is a 3 dimensional data set of 7 well separable clusters:

```{r,fig.width=5, fig.height=5}
library(PDEnaiveBayes)

data(Hepta)
Data = Hepta$Data
Cls = Hepta$Cls
```

Fitting the PDENB classifier

```{r,fig.width=5, fig.height=5}

model = Train_naiveBayes(Data, 
                          Cls, 
                          Gaussian = FALSE, 
                          Plausible = FALSE)
table(Cls, model$ClsTrain)

```
For the `Train_naiveBayes()` function, further optional arguments can be specified — for example, `Gaussian = TRUE`, when, as in classic Naive Bayes, the assumption is of Gaussian-distributed features.  

If the `Gaussian` argument is not specified or is set to `FALSE` (the default), then the nonparametric PDE approach is used.  

Furthermore, a plausible correction of the posteriors is performed by default in regions of high uncertainty, but this can be deactivated by setting `Plausible = FALSE`.  

The class-conditional likelihoods can be plotted by setting `PlotIt = TRUE`.  

It is also possible to train the PDENB model in parallel through a multicore implementation.  
For this, the user must specify, in addition to the input data, a cluster object from the **parallel** package:

```{r,fig.width=5, fig.height=5, eval = FALSE}

if (requireNamespace("parallel")) {

  library(parallel)
  
  train_index = sample(1:nrow(Data), 0.8 * nrow(Data))
  
  train_data = Data[train_index, ]
  test_data = Data[-train_index, ]
  train_cls = Cls[train_index]
  test_cls = Cls[-train_index]
  
  num_cores = detectCores() - 1
  cl = makeCluster(num_cores)
  if (requireNamespace("FCPS")) {
    data(Hepta)
    Data = Hepta$Data
    Cls  = Hepta$Cls
    model_mc = Train_naiveBayes_multicore(
      cl       = cl,
      Data     = train_data,
      Cls      = train_cls,
      Predict  = TRUE
    )
    table(train_cls, model_mc$ClsTrain)
  }
}
```


Optionally the `Train_naiveBayes_multicore()` also allows the usage of the **memshare**-package, for reducing memory usage in parallel computing, through setting the argument `UseMemshare = TRUE`

Predictions with the PDENB model:

```{r,fig.width=5, fig.height=5, eval=FALSE}
preds = Predict_naiveBayes(test_data, model_mc)
table(test_cls, preds$Cls)
```

## Visualization and interpretation

The **PDEnaiveBayes** package offers several visualization methods, which can be helpful in interpreting the fitted model.

The `PlotLikelihoodFuns()` function can be used to plot the estimated class-conditional likelihoods $p(\mathbf{x^{(i)}} \mid c_k)$ per feature:

Exemplary here it can be seen, that around 0 one class has the highest likelihood in all feature dimensions, which corresponds to the middle cluster in the Hepta data set.

```{r,fig.width=5, fig.height=5}
PlotLikelihoodFuns(
  LikelihoodFuns = model$Model$PDFs_funs,
  Data        = Hepta$Data
)
```


Another option of visualizing the estimated class-conditional likelihoods per feature is through Mirror Density Plots (MD-Plots), which can be helpful in identifying discriminating features.

```{r,fig.width=5, fig.height=5, eval = FALSE}
# CRAN does not allow interactive plotly plots, hence the plot is given as a static image
PlotNaiveBayes(Model = model$Model, FeatureNames = colnames(Hepta$Data))
```
![Alt text](./images/ClassConditionalLikelihoodsPerFeatureMDPlot.png)


Visualizing the posteriors or more general the decision boundaries of the classifier in more than 2 or 3 dimensions is challenging. The function `PlotBayesianDecision2D()` allows to visualize an estimation of the decision boundary of the PDENB classifier, by considering 2D slices of the data and using the posteriors of the data points in combination with a Voronoi diagram to plot estimations of the decision boundary of the classifier. 

```{r,fig.width=5, fig.height=5}
PlotBayesianDecision2D(
  X          = Hepta$Data[, 1],
  Y          = Hepta$Data[, 2],
  Posteriors = model$Posteriors,
  Class      = 2,
  xlab = "X1",
  ylab = "X2"
)
```
The plot shows the Voronoi tessellations for each data point in the first and second feature dimensions, where each tessellation is the region which is closest to the data point in the tessellation. The red colored regions are then the regions, where the posteriori of class 2 are maximal, which allows an interpretation of the decision boundaries of the classifier.

The function `PlotPosteriors()`, for a given class plots the above plot for each possible 2D slice combination of the data set.

```{r,fig.width=5, fig.height=5}
PlotPosteriors(
  Data       = Data,
  Posteriors = model$Posteriors,
  Class      = 2)
```

## Plausible Bayes correction

As example consider a univariate Gaussian mixture of two distributions, where the class labels are sampled according to the densities of the components. In this example the component with a mode on the right side, due to a higher variance, has a minimal higher density on the left most side of the distributions. This can result in the naive Bayes classifier, classifying all data points far enough on the left as coming from the second mode, even if the second mode has a high distance to the data points. In real world classification problems this can lead to non-intuitive classifications. An examples for this is given in [Stier et al., 2025] where a naive Bayes classifier classifies the humans with highest height as woman instead of men, because the the women height class has a higher variance. If in the `Train_naiveBayes()` `Plausbile=TRUE` is set, a correction to the likelihoods and posteriors is done to potentially avoid such non intuitive classifications in regions of high uncertainty. Further examples are given in [Ultsch/Lötsch, 2022].


```{r,fig.width=5, fig.height=5}

x1 = rnorm(1000,0,2)
x2 = rnorm(1000,5,5.5)
x = c(x1, x2)

f1 = dnorm(x, 0, 2)
f2 = dnorm(x, 5, 5.5)
p1 = f1 / (f1 + f2)
comp = rbinom(2000, 1, p1) + 1

model_1d = Train_naiveBayes(as.matrix(x), 
                          comp, 
                          Gaussian = FALSE, 
                          Plausible = TRUE,
                          PlotIt = TRUE)

```


# References

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 

[Ultsch, 2005] Ultsch, A.: Pareto Density Estimation: A Density Estimation for Knowledge Discovery, in Baier, D.; Wernecke, K.-D. (Eds.), Innovations in Classification, Data Science, and Information Systems (Proc. 27th Annual GfKl Conference), Springer, Berlin, pp. 91–100, https://doi.org/10.1007/3-540-26981-9_12, 2005.

[John/Langley, 1995] John, G. H. / Langley, P.: Estimating Continuous Distributions in Bayesian Classifiers, in Proceedings of the Eleventh Conference on Uncertainty in Artificial Intelligence (UAI–95), Morgan Kaufmann, San Mateo, pp. 338–345, 1995.

[Ultsch/Lötsch, 2022] Ultsch, A., & Lötsch, J.: Robust classification using posterior probability threshold computation followed by Voronoi cell based class assignment circumventing pitfalls of Bayesian analysis of biomedical data, International Journal of Molecular Sciences, Vol. 23(22), pp. 14081, 2022b.

\name{ApplyBayesTheorem4Likelihoods}
\alias{ApplyBayesTheorem4Likelihoods}
\title{ApplyBayesTheorem4Likelihoods}
\description{
Calculates the posteriors, for given likelihoods and priors using the Bayes Theorem
}
\usage{
ApplyBayesTheorem4Likelihoods(Likelihoods,Priors,threshold=.Machine$double.eps*1000)
}
\arguments{
\item{Likelihoods}{List of d numeric matrices, one per feature, each matrix with 1:k columns containing the distribution of class 1:k.}
\item{Priors}{[1:k] Numeric vector with prior probability for each class.}
\item{threshold}{(Optional: Default=0.00001).}
}
\value{
\item{Posteriors}{[1:n, 1:d] Numeric matrix with posterior probability according to the bayes theorem.}
}
\author{
Michael Thrun
}
\examples{
if(requireNamespace("FCPS")){
  data(Hepta)
  Data=Hepta$Data
  Cls=Hepta$Cls
  #parametric
  #V=Train_naiveBayes(Data,Cls,Gaussian=TRUE)
  #ClsTrain=V$ClsTrain
  #table(Cls,ClsTrain)
  
  #non-parametric
  V=Train_naiveBayes(Data,Cls,Gaussian=FALSE)
  ClsTrain=V$ClsTrain
  table(Cls,ClsTrain)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{PlotBayesianDecision2D}
\alias{PlotBayesianDecision2D}
\title{PlotBayesianDecision2D}
\description{
Plots estimation of decision boundary in a 2D slice of the data using the posteriors}
\usage{
PlotBayesianDecision2D(X, Y, Posteriors, Class = 1, NoBins,
CellColorsOrPallette, Showpoints = TRUE, xlim, ylim, xlab, ylab, main,
PlotIt = TRUE)
}
\arguments{
\item{X}{Numeric vector with point coordinates of first dimension of data selection.}
\item{Y}{Numeric vector with point coordinates of second dimension of data selection.}
\item{Posteriors}{[1:n, 1:Class] matrix of posteriors.}
\item{Class}{Optional,Integer defining which class to look at.}
\item{NoBins}{Optional,Number of bins for class posteriori.}
\item{CellColorsOrPallette}{Optional, Either a function defining the color palette of a
character vector or character vector of length NoBins stating colors.}
\item{Showpoints}{Optional, TRUE, points are displayed.}
\item{xlim}{Optional,Numeric vector of length 2 stating limits of x axis.}
\item{ylim}{Optional,Numeric vector of length 2 stating limits of y axis.}
\item{xlab}{Optional,Character stating name of x axis.}
\item{ylab}{Optional,Character stating name of y axis.}
\item{main}{Optional, Character name of title}
\item{PlotIt}{Optional, TRUE: prints GGPLOT2 object, FALSE: not shown plot.}
}
\value{
List of:
\item{Mapping}{List containing a map for colors, kernels and bin number.}
\item{GGobj}{ggplot2 object containing 2D visualization of Posteriori.}
}
\details{
Boundaries are assumed to be zero for plotting.
}
\author{
Michael Thrun
}
\examples{

Data = as.matrix(iris[,1:4])
Cls = as.numeric(iris[,5])

TrainIdx = c(17, 73, 46, 29, 68, 35, 131, 62, 132, 127, 71, 72, 
144, 99, 93, 13, 38, 21, 102, 53, 36, 111, 114, 96, 57, 74, 145, 
86, 3, 16, 52, 59, 140, 40, 122, 109, 6, 91, 79, 15, 108, 139, 
37, 76, 20, 115, 66, 28, 100, 117, 44, 78, 80, 150, 146, 142, 
9, 90, 45, 58, 134, 11, 87, 125, 141, 118, 136, 48, 124, 47, 
8, 27, 33, 92, 130, 54, 65, 104, 23, 98, 129, 123, 34, 128, 135, 
51, 64, 5, 94, 83, 42, 116, 101, 43, 7, 12, 82, 1, 84, 138, 2, 
56, 4, 106, 120)

TestIdx = c(60, 10, 75, 70, 81, 18, 97, 95, 67, 22, 55, 143, 
88, 24, 105, 26, 119, 31, 107, 63, 41, 61, 32, 147, 89, 14, 121, 
19, 113, 49, 126, 112, 25, 77, 137, 103, 50, 30, 149, 110, 39, 
69, 148, 85, 133)

TrainX = Data[TrainIdx, ]
TestX  = Data[TestIdx, ]
TrainY = Cls[TrainIdx]
TestY  = Cls[TestIdx]

VPDENB = Train_naiveBayes(Data = TrainX, Cls = TrainY, Plausible = FALSE)

PlotBayesianDecision2D(X = TrainX[, 1], Y = TrainX[, 2],
Posteriors = VPDENB$Posteriors, Class = 1)

}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{predict.PDEbayes}
\alias{predict.PDEbayes}
\title{predict.PDEbayes}
\description{
Predict a classification with the Pareto Density estimated naive Bayes model [Stier et al., 2026]	.
(PDENB).
}
\usage{
predict.PDEbayes(object, newdata, type = c("class", "response","prob"), ...)
}
\arguments{
  \item{object}{Model obtained from training routine in PDEnaiveBayes package.}
  \item{newdata}{[1:n,1:d] matrix of test data. It consists of n cases of d-dimensional data points. Every case has d attributes, variables or features.}
  \item{type}{Optional parameter.}
  \item{\dots}{
    \code{Gaussian}: Optional: Default=TRUE). Assume gaussian distribution.
    \code{Plausible}: (Optional: TRUE: uses plausble bayesian theorem, FALSE non-plausible bayesian theorem
    \code{Type}: (Optional: default=1, 1 = original PDE, 2 = R native density estimation
    \code{Threshold}: Threshold for which the standard deviation cannot be smaller (default =1e-12)
    \code{PlotIt}: Optional: Default=FALSE, TRUE: Plots Likelihoods
    \code{PlotCutOff}: Optional: Scalar indicating how many features (starting from 1) should be plotted, or a numerical vector specifying the indices of the features to plot.
      Note: In the second case, avoid selecting too many features, as this may cause the plot to fail
    \code{ParetoRadiusPerFeauture}: Optional [1:d] numerical vector for pareto radii computed priorly, see  \code{\link[DataVisualizations]{ParetoRadius}} or  \code{{ParetoRadius_fast}}
    \code{cl}: Optional: a cluster object, created by parallel, if given and \code{ParetoRadiusPerFeauture} missing, then \code{ParetoRadiusPerFeauture} is compputed multicore otherwise single core
    \code{Robust}: Optional: Default=FALSE, TRUE: robust estimation of mean and std in case of \code{Gaussian=TRUE}
  }
}
\value{
\item{Cls}{Numeric vector with predicted class associated with newdata.}
}
\details{
The function is implemented in a way so that one can combine training and test data although it is intended to be applied on test data only.
}
\author{
Michael Thrun
}
\seealso{
\code{\link{Train_naiveBayes}}
}
\examples{
if(requireNamespace("FCPS")){
V=FCPS::ClusterChallenge("Hepta",1000)
Data=V$Hepta
Cls=V$Cls
ind=1:length(Cls)
indtrain=sample(ind,800)
indtest=setdiff(ind,indtrain)

model=Train_naiveBayes(Data[indtrain,],Cls[indtrain],Gaussian=FALSE)
ClsTrain=model$ClsTrain
table(Cls[indtrain],ClsTrain)

ClsTest=predict.PDEbayes(object = model, newdata = Data[indtest,])
table(Cls[indtest],ClsTest)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}
\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{fitParameters}
\alias{fitParameters}
\title{fitParameters}
\description{
Fit Gaussian parameters.
}
\usage{
fitParameters(Feature,ClassInd,Robust=FALSE,na.rm=TRUE,SD_Threshold=0.0001)
}
\arguments{
\item{Feature}{[1:n] Numeric Vector}
\item{ClassInd}{Integer Vector with class indices}
\item{Robust}{(Optional: Default=FALSE). Robust computation if set to TRUE.}
\item{na.rm}{(Optional: Default=TRUE). Remove na.}
\item{SD_Threshold}{(Optional: Default=0.00001).}
}
\value{
\item{Parameters}{[1:2] Numeric vector with Mean and Std.}
}
\author{
Michael Thrun
}
\examples{
if(requireNamespace("FCPS")){
data(Hepta)
Data=Hepta$Data
Cls=Hepta$Cls
Priors=getPriors(Cls)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\name{Train_naiveBayes}
\alias{Train_naiveBayes}
\title{Train_naiveBayes}
\description{
Trains a Pareto Density estimated naive Bayes model (PDENB) of [Stier et al., 2026].
}
\usage{
Train_naiveBayes(Data,Cls,Predict=TRUE,Priors,...)
}
\arguments{
\item{Data}{\code{[1:n,1:d]} matrix of training data. It consists of \code{n}
cases of \code{d}-dimensional data points. Every case has\code{d} attributes,
variables or features.
}
\item{Cls}{\code{[1:n]}  numerical vector with n numbers defining the
classification. It has \code{k} unique numbers representing the arbitrary labels
of the classification.
}
\item{Predict}{Optional, boolean to decide extent of output. In case of TRUE,
yields ClsTrain and Posteriors, else it yields only Model and Thetas.
Note: Only if Predict is set to TRUE, parameter EvalPlausible can be set true!
}
\item{Priors}{Optional, \code{[1:k]}  numerical vector defining the prior
probabilities of the \code{k} classes. If missing, estimated from \code{Cls}.
}
\item{\dots}{

\code{Gaussian}: Optional: Default=TRUE). Assume gaussian distribution.

\code{Plausible}: (Optional: TRUE: uses plausble bayesian theorem, FALSE non-plausible bayesian theorem.

\code{Type}: (Optional: default=1, 1 = original PDE, 2 = R native density estimation.

\code{Threshold}: Threshold for which the standard deviation cannot be smaller (default =1e-12).

\code{PlotIt}: Optional: Default=FALSE, TRUE: Plots Likelihoods.

\code{PlotCutOff}: Optional: Scalar indicating how many features (starting from 1) should be plotted, or a numerical vector specifying the indices of the features to plot.
Note: In the second case, avoid selecting too many features, as this may cause the plot to fail.

\code{ParetoRadiusPerFeauture}: Optional [1:d] numerical vector for pareto radii computed priorly, see  \code{\link[DataVisualizations]{ParetoRadius}} or  \code{{ParetoRadius_fast}}

\code{cl}: Optional: a cluster object, created by parallel, if given and \code{ParetoRadiusPerFeauture} missing, then \code{ParetoRadiusPerFeauture} is compputed multicore otherwise single core

\code{Robust}: Optional: Default=FALSE, TRUE: robust estimation of mean and std in case of \code{Gaussian=TRUE}.

\code{GlobalPR}: Optional: Default=TRUE, FALSE: estimation of pareto radius for each class individually.

}
}

\value{
\item{Model}{List of model parameters and results.}
\item{c_Kernels_list}{List of matrices, where each matrix represent the kernels of one feature for all classes.}
\item{ListOfLikelihoods}{List of matrices, where each matrix represent the likelihood of one feature for all classes.}
\item{PDFs_funs}{Nested list of depth 1, where the first index assigns the
feature index and the second index assigns the class.
The elements are functions for the density estimation for each feature and each
class.}
\item{ParetoRadiusPerFeauture}{Numeric vector which stores the pareto radius
for each feature.}
\item{Theta}{Parameters mean and standard deviation of the Gaussian
distributions per class and feaures.}
\item{Priors}{Numeric vector which stores the prior probability of each class to appear.}
\item{PlausibleCenters}{[1:k, 1:f] Numeric matrix which stores the centers for
each feature and each class, where the row index assigns features and the column
index assigns classes.}
\item{ClsTrain}{[1:n]  numerical vector with n numbers defining the
classification. It has k unique numbers representing the arbitrary labels of the
classification.}
\item{Posteriors}{[1:n, 1:k] Numeric matrices with posterior probabilities.}
}
\details{
Precomputation of \code{ParetoRadiusPerFeauture} can be usefull to make cross-validation faster although it should be only done on the training data.

If \code{Plausible} is not given, both options are evalauted using shannon information.

\code{c_Kernels_list} and \code{ListOfLikelihoods} have \code{d} elements each storing a matrix \code{[1:m,1:k]}, usually \code{m!=n}. In contrast to \code{DataLikelihoodsPerClass} in which by interpolation the matrix are of size \code{[1:n,1:k]}
}
\author{
Michael Thrun
}
\seealso{
\code{\link{Predict_naiveBayes}}
}
\examples{
if(requireNamespace("FCPS")){
data(Hepta)
Data=Hepta$Data
Cls=Hepta$Cls

#non-parametric
V=Train_naiveBayes(Data,Cls,Gaussian=FALSE)
ClsTrain=V$ClsTrain
table(Cls,ClsTrain)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{Predict_naiveBayes}
\alias{Predict_naiveBayes}
\title{Predict_naiveBayes}
\description{
Predict classification with naive Bayes model [Stier et al., 2026].
}
\usage{
Predict_naiveBayes(Data, Model, \dots)
}
\arguments{
\item{Data}{[1:n,1:d] matrix of test data. It consists of n cases of
d-dimensional data points. Every case has d attributes, variables or features.}

\item{Model}{Optional, list with elements \code{Priors},\code{c_2List_Train},\code{Thetas}, alternative set arguements seperatly }
\item{\dots}{
\code{Priors}: Optional, if \code{Model} missing, then [1:k] Numeric vector with prior probability for each class.

\code{c_2List_Train}: Optional, if \code{Model} missing, then c_2List_Train is the output of GetLikelihoods: a list of two
three  elements of Kernels,  Likelihoods per feature and class, optional Thetas or PlausibleCenters depending on parameter setting

\code{Thetas}: Optional, if \code{Model} missing, then If c_2List_Train is missing, alternatively the parameters mean and
standard deviation of the gaussian distributions per class and feaures.

\code{PlotIt}: Optional: Default=FALSE, TRUE: Plots Likelihoods

\code{PlotCutOff}: Optional: Scalar indicating how many features (starting from 1) should be plotted, or a numerical vector specifying the indices of the features to plot.
Note: In the second case, avoid selecting too many features, as this may cause the plot to fail

}
}
\value{
\item{Cls}{[1:n]  numerical vector with n numbers defining the
classification. It has k unique numbers representing the arbitrary labels of the
classification.}
\item{Posteriors}{[1:n, 1:l] Numeric matrices with posterior probabilities.}
\item{DataLikelihoodsPerClass}{list of length \code{d}, each element is a matrix \code{[1:n,1:k]} of interpolated class likelihoods per feature \code{d}}

}
\details{
The function is implemented in a way so that one can combine training and test data although it is intended to be applied on test data only.
}
\author{
Michael Thrun
}
\seealso{
\code{\link{Train_naiveBayes}}
}
\examples{
if(requireNamespace("FCPS")){
V=FCPS::ClusterChallenge("Hepta",1000)
Data=V$Hepta
Cls=V$Cls
ind=1:length(Cls)
indtrain=sample(ind,800)
indtest=setdiff(ind,indtrain)

#PDEbayes
model=Train_naiveBayes(Data[indtrain,],Cls[indtrain],Gaussian=FALSE)
ClsTrain=model$ClsTrain
table(Cls[indtrain],ClsTrain)

res=Predict_naiveBayes(Data[indtest,], Model = model)
table(Cls[indtest],res$ClsTest)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{Train_naiveBayes_multicore}
\alias{Train_naiveBayes_multicore}
\title{Train a Multicore Pareto Density Naive Bayes Classifier}
\description{
Trains a naive Bayes classifier using Gaussian likelihoods or nonparametric
Pareto density estimation (PDE), with an optional plausible correction for
low-evidence cases [Stier et al., 2026]. Model training can be distributed across a
parallel cluster. Shared-memory computation can be enabled to prevent memory
duplication among workers [Thrun and Märte, 2026].
}
\usage{
Train_naiveBayes_multicore(
  cl = NULL,
  Data,
  Cls,
  Plausible = TRUE,
  Predict = FALSE,
  Priors,
  UseMemshare = FALSE,
  ...
)
}
\arguments{
\item{cl}{
A cluster object, typically created with
\code{\link[parallel]{makeCluster}}. If \code{NULL}, the function runs in a
single R process; this mode is intended primarily for debugging.
}
\item{Data}{
A numeric \code{n} by \code{d} matrix containing \code{n} observations in
rows and \code{d} features in columns. Objects that are not matrices are
converted with \code{as.matrix()}.
}
\item{Cls}{
A numeric vector of length \code{n} containing the class label for each
observation. The vector must contain at least two distinct finite class labels.
If \code{Cls} is not numeric, the function attempts to convert it using
\pkg{FCPS}; otherwise it falls back to \code{as.numeric()}.
}
\item{Plausible}{
Logical. If \code{TRUE}, use the plausible naive Bayes extension described by
Stier et al. (2026) for low-evidence cases. If \code{FALSE}, use the standard
Bayes decision rule. The plausible adjustment is intended for non-Gaussian
likelihood models and requires \pkg{FCPS}.
}
\item{Predict}{
Logical. If \code{TRUE}, classify the training observations after fitting and
return their predicted classes and posterior probabilities. If \code{FALSE},
only the model is fitted; \code{ClsTrain} and \code{Posteriors} are returned as
\code{NULL}.
}
\item{Priors}{
Optional numeric vector of length \code{k} containing the prior probabilities
of the \code{k} classes. Names should correspond to the class labels. If
missing, priors are estimated from the filtered training labels in \code{Cls}.
}
\item{UseMemshare}{
Logical. If \code{TRUE} and \code{cl} is not \code{NULL}, use
\code{\link[memshare]{memApply}} so that memory can be shared within the
multicore computation. This can reduce duplication of large training objects
among workers and requires the \pkg{memshare} package. If \code{FALSE}, the
\pkg{parallel} backend is used.
}
\item{\dots}{
Additional named arguments passed to the underlying likelihood-estimation and
prediction functions. Common arguments include:

\describe{
\item{\code{Gaussian}}{
Logical; default \code{FALSE}. If \code{TRUE}, estimate a Gaussian likelihood
for every feature and class. If \code{FALSE}, estimate likelihoods
nonparametrically.
}
\item{\code{Type}}{
Integer selecting the nonparametric density estimator. \code{1} uses Pareto
density estimation, \code{2} uses \code{stats::density()}, and \code{3} uses
\code{stats::density()} with the Pareto radius as its bandwidth. The default is
\code{1}.
}
\item{\code{Threshold}}{
Numeric threshold used by the plausible-likelihood adjustment. The default in
\code{Train_naiveBayes()} is \code{1e-4}.
}
\item{\code{SD_Threshold}}{
Positive numeric lower bound used for the standard deviation during Gaussian
parameter estimation. The default is \code{0.001}.
}
\item{\code{ParetoRadiusPerFeauture}}{
Optional precomputed Pareto radius. The spelling of this argument is retained
for API compatibility. The current multicore wrapper supports a scalar value;
a feature-specific vector is not supported. A radius must be estimated using
training data only, for example with
\code{\link[DataVisualizations]{ParetoRadius}} or
\code{\link[DataVisualizations]{ParetoRadius_fast}}.
}
\item{\code{Robust}}{
Logical; default \code{FALSE}. If \code{TRUE} and \code{Gaussian = TRUE}, use
robust estimates of the class-wise means and standard deviations.
}
\item{\code{GlobalPR}}{
Logical; default \code{TRUE}. If \code{TRUE}, estimate one Pareto radius per
feature. If \code{FALSE}, estimate a separate Pareto radius for each class and
feature.
}
\item{\code{PlotIt}}{
Logical; default \code{FALSE}. If \code{TRUE}, plot the estimated likelihoods.
Plotting is intended for the single-process debugging mode
\code{cl = NULL}.
}
\item{\code{PlotCutOff}}{
A positive integer giving the number of leading features to plot, or an integer
vector containing the feature indices to plot. Selecting many features can
produce an excessively large plot.
}
}
}
}
\value{
A list with the following components:
\describe{
\item{\code{Model}}{
An object of class \code{PDEbayes} containing the fitted model. Depending on the
selected options, it contains:
\describe{
\item{\code{c_Kernels_list}}{
A list of length \code{d}. Each element is a matrix containing the density
kernels for one feature across the \code{k} classes.
}
\item{\code{ListOfLikelihoods}}{
A list of length \code{d}. Each element is a matrix containing the estimated
likelihood values for one feature across the \code{k} classes.
}
\item{\code{PDFs_funs}}{
A feature-indexed list of class-specific density functions.
}
\item{\code{ParetoRadiusPerFeauture}}{
For non-Gaussian models, either a numeric vector with one Pareto radius per
feature or, when \code{GlobalPR = FALSE}, a class-by-feature matrix. For
Gaussian models, this component is not used for likelihood estimation.
}
\item{\code{Thetas}}{
For Gaussian models, a list of length \code{d}. Each element is a \code{k} by
\code{2} matrix containing the class-wise means and standard deviations. It is
\code{NULL} for non-Gaussian models.
}
\item{\code{Priors}}{
A named numeric vector containing the class prior probabilities.
}
\item{\code{Plausible}}{
Logical value indicating whether plausible likelihood correction was requested.
}
\item{\code{PlausibleCenters}}{
When plausible correction is used, a \code{d} by \code{k} numeric matrix of
feature-wise class centers; otherwise \code{NULL}.
}
}
}
\item{\code{ClsTrain}}{
If \code{Predict = TRUE}, a numeric vector of length \code{n} containing the
predicted class labels for the training observations; otherwise \code{NULL}.
}
\item{\code{Posteriors}}{
If \code{Predict = TRUE}, an \code{n} by \code{k} matrix of posterior class
probabilities; otherwise \code{NULL}.
}
}
}
\details{
The function trains every column of \code{Data} independently
and combines the resulting components into one multivariate naive Bayes model.
With a non-\code{NULL} cluster, the feature-wise fits are evaluated in
parallel.

When \code{UseMemshare = FALSE}, the standard \pkg{parallel} backend is used
and worker processes can hold separate copies of the required objects. When
\code{UseMemshare = TRUE}, \pkg{memshare} provides shared-memory objects during
the multicore computation [Thrun and Märte, 2026]. The potential memory saving is greatest for large
training matrices and multiple workers. Memory sharing has no effect when
\code{cl = NULL}.

If the number of rows in \code{Data} differs from the length of \code{Cls}, the
inputs are shortened to their common length with a warning. Observations with
non-finite class labels are removed before default priors are estimated.

If \code{Data} has no column names, feature names of the form \code{X1},
\code{X2}, and so on are generated. Feature names are retained in the fitted
model components.

A precomputed Pareto radius can accelerate repeated fitting, such as during
cross-validation. To avoid information leakage, it must be estimated separately
within each training fold and never from the corresponding validation data.
}
\author{
Michael Thrun
}
\seealso{
\code{\link{Train_naiveBayes}},
\code{\link{Predict_naiveBayes}},
\code{\link{predict.PDEbayes}}
}
\examples{
data(Hepta)
Data <- Hepta$Data
Cls <- Hepta$Cls

## Single-process debugging mode
Vdebug <- Train_naiveBayes_multicore(
  cl = NULL,
  Data = Data,
  Cls = Cls,
  Plausible = FALSE,
  Gaussian = FALSE,
  Predict = TRUE
)

    if(requireNamespace("parallel")){
    cl = parallel::makeCluster(1)#set to number of cores >1
    #each core copies the memory
    V=Train_naiveBayes_multicore(cl=cl,Data=Data,Cls=Cls,
    Predict=TRUE,UseMemshare=FALSE)
    ClsTrain=V$ClsTrain
    table(Cls,ClsTrain)
    
    #each core shares the memory
    V=Train_naiveBayes_multicore(cl=cl,Data=Data,Cls=Cls,
    Predict=TRUE,UseMemshare=TRUE)
    ClsTrain=V$ClsTrain
    table(Cls,ClsTrain)
    on.exit(parallel::stopCluster(cl))
    }
}



\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 

[Thrun and Märte, 2026] Thrun, M.C., Märte, J.: Memshare: Memory Sharing for Multicore Computation in R with an Application to Feature Selection by Mutual Information using PDE, The R Journal, Vol. 17(4), pp. 306 - 322, doi 10.32614/RJ-2025-043, 2026.

}

\keyword{Classification}
\keyword{Bayes}
\keyword{classif}
\keyword{models}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}
\concept{Parallel Computing}
\concept{Shared Memory}

\name{PDEnaiveBayes-package}
\alias{PDEnaiveBayes-package}
\docType{package}
\title{
\packageTitle{PDEnaiveBayes}
}
\description{
\packageDescription{PDEnaiveBayes}
}
\details{
Pareto Density Estimated naive Bayes Classifier \packageIndices{PDEnaiveBayes} (PDENB) of [Stier et al., 2026].
}
\author{
Michal Thrun

Maintainer: Michael Thrun <mthrun@informatik.uni-marburg.de>
}
\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 

[Thrun et al., 2020]  Thrun, M. C., Gehlert, T., & Ultsch, A.: Analyzing the
Fine Structure of Distributions, PloS one, Vol. 15(10), pp. e0238835,
doi 10.1371/journal.pone.0238835 	 2020.

[Thrun/Ultsch, 2020]  Thrun, M. C., & Ultsch, A.: Clustering Benchmark Datasets
Exploiting the Fundamental Clustering Problems, Data in Brief, Vol. 30(C),
pp. 105501, doi 10.1016/j.dib.2020.105501, 2020.

[Ultsch et al., 2015]  Ultsch, A., Thrun, M. C., Hansen-Goos, O., & L?tsch, J.:
Identification of Molecular Fingerprints in Human Heat Pain Thresholds by Use of
an Interactive Mixture Model R Toolbox (AdaptGauss), International journal of
molecular sciences, Vol. 16(10), pp. 25897-25911, doi 10.3390/ijms161025897,
2015.
}
\examples{
if(requireNamespace("FCPS")){
V=FCPS::ClusterChallenge("Hepta",1000)
Data=V$Hepta
Cls=V$Cls
ind=1:length(Cls)
indtrain=sample(ind,800)
indtest=setdiff(ind,indtrain)
#parametric
#model=Train_naiveBayes(Data[indtrain,],Cls[indtrain],Gaussian=TRUE)
#ClsTrain=model$ClsTrain
#table(Cls[indtrain],ClsTrain)

#res=Predict_naiveBayes(Data[indtest,], Model = model)
#table(Cls[indtest],res$ClsTest)

#PDEbayes
model=Train_naiveBayes(Data[indtrain,],Cls[indtrain],Gaussian=FALSE)
ClsTrain=model$ClsTrain
table(Cls[indtrain],ClsTrain)

res=Predict_naiveBayes(Data[indtest,], Model = model)
table(Cls[indtest],res$ClsTest)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\name{getPriors}
\alias{getPriors}
\title{getPriors}
\description{
Get a prior via class proportions.
}
\usage{
getPriors(Cls)
}
\arguments{
\item{Cls}{[1:n]  numerical vector with n numbers defining the classification.
It has k unique numbers representing the arbitrary labels of the classification.
}
}
\value{
\item{Priors}{[1:k] Numeric vector with prior probability for each class.}
}
\author{
Michael Thrun
}
\examples{
if(requireNamespace("FCPS")){
data(Hepta)
Data=Hepta$Data
Cls=Hepta$Cls
Priors=getPriors(Cls)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\name{PlotLikelihoods}
\alias{PlotLikelihoods}
\title{PlotLikelihoods}
\description{
Plots the Likelihoods per feature.
}
\usage{
PlotLikelihoods(Likelihoods, Data, PlausibleLikelihoods=NULL,Epsilon=NULL,
PlausibleCenters=NULL,PlotCutOff=4,xlim)
}
\arguments{
\item{Likelihoods}{List with Likelihoods.}
\item{Data}{Numeric matrix with data.}
\item{PlausibleLikelihoods}{List with plausible Likelihoods.}
\item{Epsilon}{Numeric scalar defining epsilon fo plausible likelihoods.}
\item{PlausibleCenters}{Numeric vector [1:k] plausible centers used to compute plausible likelihoods.}
\item{PlotCutOff}{scalar defining the how many feature starting from 1 should be
plotted or numerical vector defining the index of features to be plotted in
second case should not be too many otherwise plot yields an error.}
\item{xlim}{Numeric vector of length 2 stating limits of x axis.}
}
\value{No return value.}
\details{
Boundaries are assumed to be zero for plotting.
}
\author{
Michael Thrun
}
\examples{

Data = as.matrix(iris[,1:4])
Cls = as.numeric(iris[,5])

TrainIdx = c(17, 73, 46, 29, 68, 35, 131, 62, 132, 127, 71, 72, 
144, 99, 93, 13, 38, 21, 102, 53, 36, 111, 114, 96, 57, 74, 145, 
86, 3, 16, 52, 59, 140, 40, 122, 109, 6, 91, 79, 15, 108, 139, 
37, 76, 20, 115, 66, 28, 100, 117, 44, 78, 80, 150, 146, 142, 
9, 90, 45, 58, 134, 11, 87, 125, 141, 118, 136, 48, 124, 47, 
8, 27, 33, 92, 130, 54, 65, 104, 23, 98, 129, 123, 34, 128, 135, 
51, 64, 5, 94, 83, 42, 116, 101, 43, 7, 12, 82, 1, 84, 138, 2, 
56, 4, 106, 120)

TestIdx = c(60, 10, 75, 70, 81, 18, 97, 95, 67, 22, 55, 143, 
88, 24, 105, 26, 119, 31, 107, 63, 41, 61, 32, 147, 89, 14, 121, 
19, 113, 49, 126, 112, 25, 77, 137, 103, 50, 30, 149, 110, 39, 
69, 148, 85, 133)

TrainX = Data[TrainIdx, ]
TestX  = Data[TestIdx, ]
TrainY = Cls[TrainIdx]
TestY  = Cls[TestIdx]

VPDENB = Train_naiveBayes(Data = TrainX, Cls = TrainY, Plausible = FALSE)

PlotLikelihoods(Likelihoods = VPDENB$Model$ListOfLikelihoods, Data = TrainX)
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}
\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{GetLikelihoods}
\alias{GetLikelihoods}
\title{GetLikelihoods}
\description{
Yields the likelihoods per feauture and class as values of distribution either
defined by Gaussian or estimated form the data using pareto density estimation.
}
\usage{
GetLikelihoods(Data,Cls,\dots)
}
\arguments{
\item{Data}{[1:n,1:d] matrix of training data. It consists of n cases of
d-dimensional data points. Every case has d attributes, variables or features.}
\item{Cls}{[1:n]  numerical vector with n numbers defining the classification.
It has k unique numbers representing the arbitrary labels of the
classification.}
\item{\dots}{Further arguements for \code{\link{defineOrEstimateDistribution}}
Robust=TRUE: robustly estimated gaussians
na.rm=TRUE: remove NaNs
Threshold: threshold for which the standard deviation cannot be smaller
(defaul 0.0001)}
}
\value{
\item{c_Kernels_list}{List of d numeric matrices, one per feature, each matrix
with 1:k columns containing 
  the kernels of class 1:k}
\item{ListOfLikelihoods}{
List of d numeric matrices, one per feature, each matrix with 1:k columns containing 
distribution values (likelihood) of class 1:k }
\item{Thetas}{If Gaussian=TRUE:  List of d numeric matrices, one per feauture,
each matrix with 1:k rows containing the mean in the first column and the
standard deviation in teh seconf columd of class 1:k Otherwise: NULL}
\item{ParetoRadiusPerFeauture}{Numeric vector with estimated pareto radius
per feature.}
}
\details{
Due to pareto density estimation per class and  feature, usually the number of rows in each element of
\code{c_Kernels_list} and \code{ListOfLikelihoods} varies and does not equal the number of rows of data \code{n}.
}
\author{
Michael Thrun
}
\examples{
if(requireNamespace("FCPS")){
data(Hepta)
Data=Hepta$Data
Cls=Hepta$Cls
Priors=getPriors(Cls)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{PlotNaiveBayes}
\alias{PlotNaiveBayes}
\title{PlotNaiveBayes}
\description{
Visualize the class-conditional distributions of the Pareto Density estimated
naive Bayes model (PDENB) [Stier et al., 2026].
}
\usage{
PlotNaiveBayes(Model, FeatureNames, ClassNames, DatasetName = "Data",
nrows = 1, FeatureOrderOrSubset, NumFeaturesPerRow = 4, Colors,
IndividualFigures = FALSE)
}
\arguments{
\item{Model}{List with elements \code{Priors},\code{c_2List_Train}.}
\item{FeatureNames}{Character vector of names with a name for each feature
contained in the data used to create the naive bayes model.}
\item{ClassNames}{Character vector of class names to present in the legend
of the plots.}
\item{DatasetName}{Character title for each plot.}
\item{nrows}{Number of rows inside one plot.}
\item{FeatureOrderOrSubset}{Numeric vector representing the order of the features
to be displayed or a subset as col indices om data column order.}
\item{NumFeaturesPerRow}{Maximum number of features to be displayed in one plot.}
\item{Colors}{Character vector of color names. The length of the vector must
be the same as the number of classes within the data modeled by the naive Bayes
classifier.}
\item{IndividualFigures}{Optional boolean: If set to TRUE, it returns a list
of the individual figures for customization.}
}
\value{
\item{Cls}{[1:n]  numerical vector with n numbers defining the
classification. It has k unique numbers representing the arbitrary labels of the
classification.}
\item{Posteriors}{[1:n, 1:l] Numeric matrices with posterior probabilities.}
\item{DataLikelihoodsPerClass}{list of length \code{d}, each element is a matrix \code{[1:n,1:k]} of interpolated class likelihoods per feature \code{d}}

}
\details{
Boundaries are assumed to be zero for plotting.
}
\author{
Quirin Stier
}
\examples{

Data = as.matrix(iris[,1:4])
Cls = as.numeric(iris[,5])
DatasetName = "Iris"

TrainIdx = c(17, 73, 46, 29, 68, 35, 131, 62, 132, 127, 71, 72, 
144, 99, 93, 13, 38, 21, 102, 53, 36, 111, 114, 96, 57, 74, 145, 
86, 3, 16, 52, 59, 140, 40, 122, 109, 6, 91, 79, 15, 108, 139, 
37, 76, 20, 115, 66, 28, 100, 117, 44, 78, 80, 150, 146, 142, 
9, 90, 45, 58, 134, 11, 87, 125, 141, 118, 136, 48, 124, 47, 
8, 27, 33, 92, 130, 54, 65, 104, 23, 98, 129, 123, 34, 128, 135, 
51, 64, 5, 94, 83, 42, 116, 101, 43, 7, 12, 82, 1, 84, 138, 2, 
56, 4, 106, 120)

TestIdx = c(60, 10, 75, 70, 81, 18, 97, 95, 67, 22, 55, 143, 
88, 24, 105, 26, 119, 31, 107, 63, 41, 61, 32, 147, 89, 14, 121, 
19, 113, 49, 126, 112, 25, 77, 137, 103, 50, 30, 149, 110, 39, 
69, 148, 85, 133)

TrainX = Data[TrainIdx, ]
TestX  = Data[TestIdx, ]
TrainY = Cls[TrainIdx]
TestY  = Cls[TestIdx]

VPDENB = Train_naiveBayes(Data = TrainX, Cls = TrainY, Plausible = FALSE)

FeatureNames = colnames(Data)

PlotNaiveBayes(Model = VPDENB$Model, FeatureNames = FeatureNames)
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}
\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}

\name{Hepta}
\alias{Hepta}

\docType{data}
\title{
Hepta introduced in [Ultsch, 2003]
}
\description{
Clearly defined clusters, different variances. Detailed description of dataset and its clustering challenge is provided in [Thrun/Ultsch, 2020].
}
\usage{data("Hepta")}

\details{
Size 212, Dimensions 3, stored in \code{Hepta$Data}

Classes 7, stored in \code{Hepta$Cls}

}

\references{
[Ultsch, 2003]  Ultsch, A.: Maps for the visualization of high-dimensional data spaces, Proc. Workshop on Self organizing Maps (WSOM), pp. 225-230, Kyushu, Japan, 2003.

[Thrun/Ultsch, 2020]  Thrun, M. C., & Ultsch, A.: Clustering Benchmark Datasets Exploiting the Fundamental Clustering Problems, Data in Brief, Vol. 30(C), pp. 105501, \doi{10.1016/j.dib.2020.105501}, 2020.
}
\examples{
data(Hepta)
str(Hepta)
}
\keyword{Hepta}
\keyword{datasets}
\keyword{FCPS}

\name{defineOrEstimateDistribution}
\alias{defineOrEstimateDistribution}
\title{defineOrEstimateDistribution}
\description{
The function estimates the distribution of values within a features that belong
to a specific class, i.e., the conditional probability of the likelihood
}
\usage{
defineOrEstimateDistribution(Feature,ClassInd,Gaussian=FALSE,ParetoRadius=NULL,
InternalPlotIt=FALSE,SD_Threshold=0.001,...)
}
\arguments{
\item{Feature}{[1:n] Numeric Vector}
\item{ClassInd}{Integer Vector with class indices}
\item{Gaussian}{(Optional: Default=TRUE). Assume gaussian distribution.}
\item{ParetoRadius}{ Optional [1:d] numerical vector for pareto radii computed
priorly, see \code{\link[DataVisualizations]{ParetoRadius}}}
\item{InternalPlotIt}{Optional: Default=FALSE). Create plot if set to TRUE.}
\item{SD_Threshold}{Optional: Default=0.001.}
\item{\dots}{

\code{Robust}: Optional: Default=FALSE, TRUE: robust estimation of mean and std in case of \code{Gaussian=TRUE}

\code{Type}: (Optional: Default=2, 1=original PDE, 2= improved PDE

\code{na.rm}: (Optional: Default=TRUE). Remove na.
}
}
\value{
\item{Kernels}{[1:m] Numeric vector with kernels (x-values) of a 1D pdf.}
\item{PDF}{[1:m] Numeric vector with the distribution values of a 1D pdf.}
\item{Theta}{Numeric vector with parameters of gaussian of mean and standard deviation -
NULL if no gaussian used.}
}

\author{
Michael Thrun
}
\examples{
if(requireNamespace("FCPS")){
data(Hepta)
Data=Hepta$Data
Cls=Hepta$Cls
Priors=getPriors(Cls)
}
}
\keyword{Classification}
\keyword{Bayes}
\concept{Pareto Density Estimation}
\concept{Pareto Law}
\concept{Kernel Density Estimation}
\concept{Bayesian Classifier}

\references{

[Stier et al., 2026]	Stier, Q.,Hoffmann, J. & Thrun, M. C.: Classifying with the Fine Structure of Distributions: Leveraging Distributional Information for Robust and Plausible Naïve Bayes, Machine Learning and Knowledge Extraction (MAKE), Vol. 8(1), 13, doi 10.3390/make8010013, MDPI, 2026. 
}
